@@ -1,7 +1,7 @@
 /**
  * This file is part of the CernVM File System.
  */
-#include "cvmfs_config.h"
+
 #include "cache_extern.h"
 
 #include <errno.h>
@@ -22,17 +22,18 @@
 #include <set>
 #include <string>
 
-#include "atomic.h"
 #include "cache.pb.h"
-#include "hash.h"
-#include "logging.h"
-#ifdef __APPLE__
-#include "smalloc.h"
-#endif
+#include "crypto/hash.h"
+#include "util/atomic.h"
+#include "util/concurrency.h"
+#include "util/exception.h"
+#include "util/logging.h"
 #include "util/pointer.h"
 #include "util/posix.h"
+#ifdef __APPLE__
+#include "util/smalloc.h"
+#endif
 #include "util/string.h"
-#include "util_concurrency.h"
 
 using namespace std;  // NOLINT
 
@@ -295,13 +296,13 @@ ExternalCacheManager::PluginHandle *ExternalCacheManager::CreatePlugin(
 
 
 void ExternalCacheManager::CtrlTxn(
-  const ObjectInfo &object_info,
+  const Label &label,
   const int flags,
   void *txn)
 {
   Transaction *transaction = reinterpret_cast<Transaction *>(txn);
-  transaction->object_info = object_info;
-  transaction->object_info_modified = true;
+  transaction->label = label;
+  transaction->label_modified = true;
 }
 
 
@@ -324,7 +325,7 @@ int ExternalCacheManager::DoOpen(const shash::Any &id) {
     WriteLockGuard guard(rwlock_fd_table_);
     fd = fd_table_.OpenFd(ReadOnlyHandle(id));
     if (fd < 0) {
-      LogCvmfs(kLogCache, kLogDebug, "error while creating new fd",
+      LogCvmfs(kLogCache, kLogDebug, "error while creating new fd: %s",
                strerror(-fd));
       return fd;
     }
@@ -444,11 +445,11 @@ int ExternalCacheManager::Flush(bool do_commit, Transaction *transaction) {
   msg_store.set_expected_size(transaction->expected_size);
   msg_store.set_last_part(do_commit);
 
-  if (transaction->object_info_modified) {
+  if (transaction->label_modified) {
     cvmfs::EnumObjectType object_type;
-    transport_.FillObjectType(transaction->object_info.type, &object_type);
+    transport_.FillObjectType(transaction->label.flags, &object_type);
     msg_store.set_object_type(object_type);
-    msg_store.set_description(transaction->object_info.description);
+    msg_store.set_description(transaction->label.GetDescription());
   }
 
   RpcJob rpc_job(&msg_store);
@@ -536,9 +537,8 @@ void *ExternalCacheManager::MainRead(void *data) {
       cache_mgr->quota_mgr_->BroadcastBackchannels("R");
       continue;
     } else {
-      LogCvmfs(kLogCache, kLogSyslogErr | kLogDebug, "unexpected message %s",
-               msg->GetTypeName().c_str());
-      abort();
+      PANIC(kLogSyslogErr | kLogDebug, "unexpected message %s",
+            msg->GetTypeName().c_str());
     }
 
     RpcInFlight rpc_inflight;
@@ -564,16 +564,15 @@ void *ExternalCacheManager::MainRead(void *data) {
   }
 
   if (!cache_mgr->terminated_) {
-    LogCvmfs(kLogCache, kLogSyslogErr | kLogDebug,
-             "connection to external cache manager broken (%d)", errno);
-    abort();
+    PANIC(kLogSyslogErr | kLogDebug,
+          "connection to external cache manager broken (%d)", errno);
   }
   LogCvmfs(kLogCache, kLogDebug, "stopping external cache reader thread");
   return NULL;
 }
 
 
-int ExternalCacheManager::Open(const BlessedObject &object) {
+int ExternalCacheManager::Open(const LabeledObject &object) {
   return DoOpen(object.id);
 }
 
@@ -591,7 +590,7 @@ int ExternalCacheManager::OpenFromTxn(void *txn) {
     WriteLockGuard guard(rwlock_fd_table_);
     fd = fd_table_.OpenFd(ReadOnlyHandle(transaction->id));
     if (fd < 0) {
-      LogCvmfs(kLogCache, kLogDebug, "error while creating new fd",
+      LogCvmfs(kLogCache, kLogDebug, "error while creating new fd: %s",
                strerror(-fd));
       return fd;
     }
@@ -658,7 +657,7 @@ int ExternalCacheManager::Reset(void *txn) {
   transaction->size = 0;
   transaction->open_fds = 0;
   transaction->committed = false;
-  transaction->object_info_modified = true;
+  transaction->label_modified = true;
 
   if (!transaction->flushed)
     return 0;
@@ -702,6 +701,11 @@ manifest::Breadcrumb ExternalCacheManager::LoadBreadcrumb(
     assert(rv);
     breadcrumb.catalog_hash.suffix = shash::kSuffixCatalog;
     breadcrumb.timestamp = msg_reply->breadcrumb().timestamp();
+    if (msg_reply->breadcrumb().has_revision()) {
+      breadcrumb.revision = msg_reply->breadcrumb().revision();
+    } else {
+      breadcrumb.revision = 0;
+    }
   }
   return breadcrumb;
 }
@@ -717,6 +721,7 @@ bool ExternalCacheManager::StoreBreadcrumb(const manifest::Manifest &manifest) {
   breadcrumb.set_fqrn(manifest.repository_name());
   breadcrumb.set_allocated_hash(&hash);
   breadcrumb.set_timestamp(manifest.publish_timestamp());
+  breadcrumb.set_revision(manifest.revision());
   cvmfs::MsgBreadcrumbStoreReq msg_breadcrumb_store;
   msg_breadcrumb_store.set_session_id(session_id_);
   msg_breadcrumb_store.set_req_id(NextRequestId());
@@ -767,6 +772,7 @@ bool ExternalCacheManager::SpawnPlugin(const vector<string> &cmd_line) {
                        preserve_filedes,
                        map_fildes,
                        false,  // drop_credentials
+                       false,  // clear_env
                        true,   // double fork
                        &child_pid);
   unsetenv(CacheTransport::kEnvReadyNotifyFd);
