@@ -1,7 +1,7 @@
 /**
  * This file is part of the CernVM File System.
  */
-#include "cvmfs_config.h"
+
 #include "channel.h"
 
 #include <errno.h>
@@ -15,13 +15,14 @@
 #include <cstring>
 #include <vector>
 
-#include "logging.h"
-#include "platform.h"
-#include "smalloc.h"
+#include "util/concurrency.h"
+#include "util/exception.h"
+#include "util/logging.h"
+#include "util/platform.h"
 #include "util/pointer.h"
 #include "util/posix.h"
+#include "util/smalloc.h"
 #include "util/string.h"
-#include "util_concurrency.h"
 
 using namespace std;  // NOLINT
 
@@ -214,6 +215,11 @@ void CachePlugin::HandleBreadcrumbStore(
     msg_reply.set_status(cvmfs::STATUS_MALFORMED);
   } else {
     breadcrumb.timestamp = msg_req->breadcrumb().timestamp();
+    if (msg_req->breadcrumb().has_revision()) {
+      breadcrumb.revision = msg_req->breadcrumb().revision();
+    } else {
+      breadcrumb.revision = 0;
+    }
     cvmfs::EnumStatus status =
       StoreBreadcrumb(msg_req->breadcrumb().fqrn(), breadcrumb);
     msg_reply.set_status(status);
@@ -243,6 +249,7 @@ void CachePlugin::HandleBreadcrumbLoad(
     msg_breadcrumb->set_fqrn(msg_req->fqrn());
     msg_breadcrumb->set_allocated_hash(msg_hash);
     msg_breadcrumb->set_timestamp(breadcrumb.timestamp);
+    msg_breadcrumb->set_revision(breadcrumb.revision);
     msg_reply.set_allocated_breadcrumb(msg_breadcrumb);
   }
   transport->SendFrame(&frame_send);
@@ -800,9 +807,8 @@ void *CachePlugin::MainProcessRequests(void *data) {
     if (retval < 0) {
       if (errno == EINTR)
         continue;
-      LogCvmfs(kLogCache, kLogSyslogErr | kLogDebug,
-               "cache plugin connection failure (%d)", errno);
-      abort();
+      PANIC(kLogSyslogErr | kLogDebug, "cache plugin connection failure (%d)",
+            errno);
     }
 
     // Termination or detach
