@@ -1,7 +1,7 @@
 /**
  * This file is part of the CernVM File System.
  */
-#include "cvmfs_config.h"
+
 #include "cache_transport.h"
 
 #include <alloca.h>
@@ -12,10 +12,11 @@
 #include <cstdlib>
 #include <cstring>
 
-#include "hash.h"
-#include "logging.h"
-#include "smalloc.h"
+#include "crypto/hash.h"
+#include "util/exception.h"
+#include "util/logging.h"
 #include "util/posix.h"
+#include "util/smalloc.h"
 
 // TODO(jblomer): Check for possible starvation of plugin by dying clients
 // (blocking read).  Probably only relevant for TCP sockets.
@@ -218,7 +219,7 @@ void CacheTransport::Frame::WrapMsg() {
     is_msg_out_of_band_ = true;
   } else {
     // Unexpected message type, should never happen
-    abort();
+    PANIC(NULL);
   }
   is_wrapped_ = true;
 }
@@ -274,7 +275,7 @@ void CacheTransport::Frame::UnwrapMsg() {
     is_msg_out_of_band_ = true;
   } else {
     // Unexpected message type, should never happen
-    abort();
+  PANIC(NULL);
   }
 }
 
@@ -313,31 +314,20 @@ void CacheTransport::FillMsgHash(
       msg_hash->set_algorithm(cvmfs::HASH_SHAKE128);
       break;
     default:
-      abort();
+      PANIC(NULL);
   }
   msg_hash->set_digest(hash.digest, shash::kDigestSizes[hash.algorithm]);
 }
 
 
 void CacheTransport::FillObjectType(
-  CacheManager::ObjectType object_type,
-  cvmfs::EnumObjectType *wire_type)
+  int object_flags, cvmfs::EnumObjectType *wire_type)
 {
-  switch (object_type) {
-    case CacheManager::kTypeRegular:
-    // TODO(jblomer): "pinned" should mean a permanently open fd
-    case CacheManager::kTypePinned:
-      *wire_type = cvmfs::OBJECT_REGULAR;
-      break;
-    case CacheManager::kTypeCatalog:
-      *wire_type = cvmfs::OBJECT_CATALOG;
-      break;
-    case CacheManager::kTypeVolatile:
-      *wire_type = cvmfs::OBJECT_VOLATILE;
-      break;
-    default:
-      abort();
-  }
+  *wire_type = cvmfs::OBJECT_REGULAR;
+  if (object_flags & CacheManager::kLabelCatalog)
+    *wire_type = cvmfs::OBJECT_CATALOG;
+  if (object_flags & CacheManager::kLabelVolatile)
+    *wire_type = cvmfs::OBJECT_VOLATILE;
 }
 
 
@@ -367,18 +357,17 @@ bool CacheTransport::ParseMsgHash(
 
 
 bool CacheTransport::ParseObjectType(
-  cvmfs::EnumObjectType wire_type,
-  CacheManager::ObjectType *object_type)
+  cvmfs::EnumObjectType wire_type, int *object_flags)
 {
+  *object_flags = 0;
   switch (wire_type) {
     case cvmfs::OBJECT_REGULAR:
-      *object_type = CacheManager::kTypeRegular;
       return true;
     case cvmfs::OBJECT_CATALOG:
-      *object_type = CacheManager::kTypeCatalog;
+      *object_flags |= CacheManager::kLabelCatalog;
       return true;
     case cvmfs::OBJECT_VOLATILE:
-      *object_type = CacheManager::kTypeVolatile;
+      *object_flags |= CacheManager::kLabelVolatile;
       return true;
     default:
       return false;
@@ -507,10 +496,8 @@ void CacheTransport::SendData(
   bool retval = SafeWriteV(fd_connection_, iov, (att_size == 0) ? 2 : 4);
 
   if (!retval && !(flags_ & kFlagSendIgnoreFailure)) {
-    LogCvmfs(kLogCache, kLogSyslogErr | kLogDebug,
-             "failed to write to external cache transport (%d), aborting",
-             errno);
-    abort();
+    PANIC(kLogSyslogErr | kLogDebug,
+          "failed to write to external cache transport (%d), aborting", errno);
   }
 }
 
@@ -531,10 +518,9 @@ void CacheTransport::SendNonBlocking(struct iovec *iov, unsigned iovcnt) {
   if (retval < 0) {
     assert(errno != EMSGSIZE);
     if (!(flags_ & kFlagSendIgnoreFailure)) {
-      LogCvmfs(kLogCache, kLogSyslogErr | kLogDebug,
-             "failed to write to external cache transport (%d), aborting",
-             errno);
-      abort();
+      PANIC(kLogSyslogErr | kLogDebug,
+            "failed to write to external cache transport (%d), aborting",
+            errno);
     }
   }
 }
