@@ -2,6 +2,9 @@
  * This file is part of the CernVM File System
  */
 
+// avoid clang-tidy false positives (at least starting with clang14)
+//NOLINTBEGIN
+
 #ifndef CVMFS_CATALOG_MGR_IMPL_H_
 #define CVMFS_CATALOG_MGR_IMPL_H_
 
@@ -9,15 +12,15 @@
 #define __STDC_FORMAT_MACROS
 #endif
 
-#include "cvmfs_config.h"
+
 
 #include <cassert>
 #include <string>
 #include <vector>
 
-#include "logging.h"
 #include "shortstring.h"
 #include "statistics.h"
+#include "util/logging.h"
 #include "xattr.h"
 
 using namespace std;  // NOLINT
@@ -31,6 +34,7 @@ AbstractCatalogManager<CatalogT>::AbstractCatalogManager(
   inode_watermark_status_ = 0;
   inode_gauge_ = AbstractCatalogManager<CatalogT>::kInodeOffset;
   revision_cache_ = 0;
+  timestamp_cache_ = 0;
   catalog_watermark_ = 0;
   volatile_flag_ = false;
   has_authz_cache_ = false;
@@ -115,28 +119,79 @@ bool AbstractCatalogManager<CatalogT>::Init() {
  * are detached)
  */
 template <class CatalogT>
-LoadError AbstractCatalogManager<CatalogT>::Remount(const bool dry_run) {
+LoadReturn AbstractCatalogManager<CatalogT>::RemountDryrun() {
   LogCvmfs(kLogCatalog, kLogDebug,
-           "remounting repositories (dry run %d)", dry_run);
-  if (dry_run)
-    return LoadCatalog(PathString("", 0), shash::Any(), NULL, NULL);
+           "dryrun remounting repositories");
+  CatalogContext ctlg_context;
+  return GetNewRootCatalogContext(&ctlg_context);
+}
+
+template <class CatalogT>
+LoadReturn AbstractCatalogManager<CatalogT>::Remount() {
+  LogCvmfs(kLogCatalog, kLogDebug, "remounting repositories");
+  CatalogContext ctlg_context;
+
+  if (GetNewRootCatalogContext(&ctlg_context) != kLoadNew
+      && GetNewRootCatalogContext(&ctlg_context) != kLoadUp2Date) {
+    LogCvmfs(kLogCatalog, kLogDebug, "remounting repositories: "
+                                "Did not find any valid root catalog to mount");
+    return kLoadFail;
+  }
 
   WriteLock();
 
-  string     catalog_path;
-  shash::Any catalog_hash;
-  const LoadError load_error = LoadCatalog(PathString("", 0),
-                                           shash::Any(),
-                                           &catalog_path,
-                                           &catalog_hash);
+  const LoadReturn load_error = LoadCatalogByHash(&ctlg_context);
+
   if (load_error == kLoadNew) {
     inode_t old_inode_gauge = inode_gauge_;
     DetachAll();
     inode_gauge_ = AbstractCatalogManager<CatalogT>::kInodeOffset;
 
-    CatalogT *new_root = CreateCatalog(PathString("", 0), catalog_hash, NULL);
+    CatalogT *new_root = CreateCatalog(ctlg_context.mountpoint(),
+                                       ctlg_context.hash(), NULL);
     assert(new_root);
-    bool retval = AttachCatalog(catalog_path, new_root);
+    bool retval = AttachCatalog(ctlg_context.sqlite_path(), new_root);
+    assert(retval);
+
+    if (inode_annotation_) {
+      inode_annotation_->IncGeneration(old_inode_gauge);
+    }
+  }
+  CheckInodeWatermark();
+  Unlock();
+
+  return load_error;
+}
+
+/**
+ * Remounts to the given hash
+ */
+template <class CatalogT>
+LoadReturn AbstractCatalogManager<CatalogT>::ChangeRoot(
+  const shash::Any &root_hash)
+{
+  assert(!root_hash.IsNull());
+  LogCvmfs(kLogCatalog, kLogDebug,
+           "switching to root hash %s", root_hash.ToString().c_str());
+
+  WriteLock();
+
+  CatalogContext ctlg_context(root_hash, PathString("", 0),
+                                                         kCtlgNoLocationNeeded);
+  // we do not need to set revision as LoadCatalogByHash
+  // needs only mountpoint, hash
+
+  const LoadReturn load_error = LoadCatalogByHash(&ctlg_context);
+
+  if (load_error == kLoadNew) {
+    inode_t old_inode_gauge = inode_gauge_;
+    DetachAll();
+    inode_gauge_ = AbstractCatalogManager<CatalogT>::kInodeOffset;
+
+    CatalogT *new_root =
+                    CreateCatalog(PathString("", 0), ctlg_context.hash(), NULL);
+    assert(new_root);
+    bool retval = AttachCatalog(ctlg_context.sqlite_path(), new_root);
     assert(retval);
 
     if (inode_annotation_) {
@@ -232,7 +287,7 @@ bool AbstractCatalogManager<CatalogT>::LookupPath(const PathString &path,
   if (!found && MountSubtree(path, best_fit, false /* is_listable */, NULL)) {
     LogCvmfs(kLogCatalog, kLogDebug, "looking up '%s' in a nested catalog",
              path.c_str());
-    Unlock();
+    StageNestedCatalogAndUnlock(path, best_fit, false /* is_listable */);
     WriteLock();
     // Check again to avoid race
     best_fit = FindCatalog(path);
@@ -333,7 +388,7 @@ bool AbstractCatalogManager<CatalogT>::LookupNested(
   CatalogT *best_fit = FindCatalog(catalog_path);
   CatalogT *catalog = best_fit;
   if (MountSubtree(catalog_path, best_fit, false /* is_listable */, NULL)) {
-    Unlock();
+    StageNestedCatalogAndUnlock(path, best_fit, false);
     WriteLock();
     // Check again to avoid race
     best_fit = FindCatalog(catalog_path);
@@ -396,7 +451,7 @@ bool AbstractCatalogManager<CatalogT>::ListCatalogSkein(
   CatalogT *catalog = best_fit;
   // True if there is an available nested catalog
   if (MountSubtree(test, best_fit, false /* is_listable */, NULL)) {
-    Unlock();
+    StageNestedCatalogAndUnlock(path, best_fit, false);
     WriteLock();
     // Check again to avoid race
     best_fit = FindCatalog(test);
@@ -452,7 +507,7 @@ bool AbstractCatalogManager<CatalogT>::LookupXattrs(
   CatalogT *best_fit = FindCatalog(path);
   CatalogT *catalog = best_fit;
   if (MountSubtree(path, best_fit, false /* is_listable */, NULL)) {
-    Unlock();
+    StageNestedCatalogAndUnlock(path, best_fit, false);
     WriteLock();
     // Check again to avoid race
     best_fit = FindCatalog(path);
@@ -479,7 +534,8 @@ bool AbstractCatalogManager<CatalogT>::LookupXattrs(
  */
 template <class CatalogT>
 bool AbstractCatalogManager<CatalogT>::Listing(const PathString &path,
-                                     DirectoryEntryList *listing)
+                                     DirectoryEntryList *listing,
+                                     const bool expand_symlink)
 {
   EnforceSqliteMemLimit();
   bool result;
@@ -489,7 +545,7 @@ bool AbstractCatalogManager<CatalogT>::Listing(const PathString &path,
   CatalogT *best_fit = FindCatalog(path);
   CatalogT *catalog = best_fit;
   if (MountSubtree(path, best_fit, true /* is_listable */, NULL)) {
-    Unlock();
+    StageNestedCatalogAndUnlock(path, best_fit, true /* is_listable */);
     WriteLock();
     // Check again to avoid race
     best_fit = FindCatalog(path);
@@ -501,7 +557,7 @@ bool AbstractCatalogManager<CatalogT>::Listing(const PathString &path,
   }
 
   perf::Inc(statistics_.n_listing);
-  result = catalog->ListingPath(path, listing);
+  result = catalog->ListingPath(path, listing, expand_symlink);
 
   Unlock();
   return result;
@@ -526,7 +582,7 @@ bool AbstractCatalogManager<CatalogT>::ListingStat(const PathString &path,
   CatalogT *best_fit = FindCatalog(path);
   CatalogT *catalog = best_fit;
   if (MountSubtree(path, best_fit, true /* is_listable */, NULL)) {
-    Unlock();
+    StageNestedCatalogAndUnlock(path, best_fit, true /* is_listable */);
     WriteLock();
     // Check again to avoid race
     best_fit = FindCatalog(path);
@@ -566,7 +622,7 @@ bool AbstractCatalogManager<CatalogT>::ListFileChunks(
   CatalogT *best_fit = FindCatalog(path);
   CatalogT *catalog = best_fit;
   if (MountSubtree(path, best_fit, false /* is_listable */, NULL)) {
-    Unlock();
+    StageNestedCatalogAndUnlock(path, best_fit, false);
     WriteLock();
     // Check again to avoid race
     best_fit = FindCatalog(path);
@@ -583,15 +639,85 @@ bool AbstractCatalogManager<CatalogT>::ListFileChunks(
   return result;
 }
 
+template <class CatalogT>
+catalog::Counters AbstractCatalogManager<CatalogT>::LookupCounters(
+  const PathString &path,
+  std::string *subcatalog_path,
+  shash::Any *hash
+  )
+{
+  EnforceSqliteMemLimit();
+  bool result;
+  ReadLock();
+
+  // Look past current path to mount up to intended location
+  PathString catalog_path(path);
+  catalog_path.Append("/.cvmfscatalog", 14);
+
+  // Find catalog, possibly load nested
+  CatalogT *best_fit = FindCatalog(catalog_path);
+  CatalogT *catalog = best_fit;
+  if (MountSubtree(catalog_path, best_fit, false /* is_listable */, NULL)) {
+    StageNestedCatalogAndUnlock(path, best_fit, false /* is_listable */);
+    WriteLock();
+    // Check again to avoid race
+    best_fit = FindCatalog(catalog_path);
+    result =
+      MountSubtree(catalog_path, best_fit, false /* is_listable */, &catalog);
+    // Result is false if an available catalog failed to load (error happened)
+    if (!result) {
+      Unlock();
+      *subcatalog_path = "error: failed to load catalog!";
+      *hash = shash::Any();
+      return catalog::Counters();
+    }
+  }
+
+  *hash = catalog->hash();
+  *subcatalog_path = catalog->mountpoint().ToString();
+  catalog::Counters counters = catalog->GetCounters();
+  Unlock();
+  return counters;
+}
+
 
 template <class CatalogT>
 uint64_t AbstractCatalogManager<CatalogT>::GetRevision() const {
   ReadLock();
-  const uint64_t revision = revision_cache_;
+  const uint64_t revision = GetRevisionNoLock();
   Unlock();
+
   return revision;
 }
 
+/**
+ * Like GetRevision() only without any locking mechanism.
+ * As such should only be used in conditions where a lock was already taken
+ * and calling GetRevision() would otherwise result in a deadlock.
+ */
+template <class CatalogT>
+uint64_t AbstractCatalogManager<CatalogT>::GetRevisionNoLock() const {
+  return revision_cache_;
+}
+
+template <class CatalogT>
+uint64_t AbstractCatalogManager<CatalogT>::GetTimestamp() const {
+  ReadLock();
+  const uint64_t timestamp = GetTimestampNoLock();
+  Unlock();
+
+  return timestamp;
+}
+
+/**
+ * Like GetTimestamp() only without any locking mechanism.
+ * As such should only be used in conditions where a lock was already taken
+ * and calling GetTimestamp() would otherwise result in a deadlock.
+ */
+template <class CatalogT>
+uint64_t AbstractCatalogManager<CatalogT>::GetTimestampNoLock() const {
+  return timestamp_cache_;
+}
 
 template <class CatalogT>
 bool AbstractCatalogManager<CatalogT>::GetVOMSAuthz(std::string *authz) const {
@@ -637,7 +763,7 @@ int AbstractCatalogManager<CatalogT>::GetNumCatalogs() const {
 template <class CatalogT>
 string AbstractCatalogManager<CatalogT>::PrintHierarchy() const {
   ReadLock();
-  const string output = PrintHierarchyRecursively(GetRootCatalog(), 0);
+  string output = PrintHierarchyRecursively(GetRootCatalog(), 0);
   Unlock();
   return output;
 }
@@ -653,7 +779,7 @@ InodeRange AbstractCatalogManager<CatalogT>::AcquireInodes(uint64_t size) {
   result.size = size;
 
   inode_gauge_ += size;
-  LogCvmfs(kLogCatalog, kLogDebug, "allocating inodes from %d to %d.",
+  LogCvmfs(kLogCatalog, kLogDebug, "allocating inodes from %lu to %lu.",
            result.offset + 1, inode_gauge_);
 
   return result;
@@ -718,6 +844,45 @@ bool AbstractCatalogManager<CatalogT>::IsAttached(const PathString &root_path,
 }
 
 
+template <class CatalogT>
+void AbstractCatalogManager<CatalogT>::StageNestedCatalogAndUnlock(
+  const PathString &path,
+  const CatalogT *parent,
+  bool is_listable)
+{
+  assert(parent);
+  const unsigned path_len = path.GetLength();
+
+  perf::Inc(statistics_.n_nested_listing);
+  typedef typename CatalogT::NestedCatalogList NestedCatalogList;
+  const NestedCatalogList& nested_catalogs = parent->ListNestedCatalogs();
+
+  for (typename NestedCatalogList::const_iterator i = nested_catalogs.begin(),
+       iEnd = nested_catalogs.end(); i != iEnd; ++i)
+  {
+    if (!path.StartsWith(i->mountpoint))
+      continue;
+
+    // in this case the path doesn't start with
+    // the mountpoint in a file path sense
+    // (e.g. path is /a/bc and mountpoint is /a/b), and will be ignored
+    const unsigned mountpoint_len = i->mountpoint.GetLength();
+    if (path_len > mountpoint_len && path.GetChars()[mountpoint_len] != '/')
+      continue;
+
+    // Found a nested catalog transition point
+    if (!is_listable && (path_len == mountpoint_len))
+      break;
+
+    Unlock();
+    LogCvmfs(kLogCatalog, kLogDebug, "staging nested catalog at %s (%s)",
+             i->mountpoint.c_str(), i->hash.ToString().c_str());
+    StageNestedCatalogByHash(i->hash, i->mountpoint);
+    return;
+  }
+  Unlock();
+}
+
 /**
  * Recursively mounts all nested catalogs required to serve a path.
  * If leaf_catalog is NULL, just indicate if it is necessary to load a
@@ -725,7 +890,7 @@ bool AbstractCatalogManager<CatalogT>::IsAttached(const PathString &root_path,
  * The final leaf nested catalog is returned.
  * The is_listable parameter is relevant if path is a nested catalog.  Only
  * if is_listable is true, the nested catalog will be used; otherwise the parent
- * with the transation point is sufficient.
+ * with the transaction point is sufficient.
  */
 template <class CatalogT>
 bool AbstractCatalogManager<CatalogT>::MountSubtree(
@@ -739,9 +904,9 @@ bool AbstractCatalogManager<CatalogT>::MountSubtree(
                      GetRootCatalog() : const_cast<CatalogT *>(entry_point);
   assert(path.StartsWith(parent->mountpoint()));
 
+  unsigned path_len = path.GetLength();
+
   // Try to find path as a super string of nested catalog mount points
-  PathString path_slash(path);
-  path_slash.Append("/", 1);
   perf::Inc(statistics_.n_nested_listing);
   typedef typename CatalogT::NestedCatalogList NestedCatalogList;
   const NestedCatalogList& nested_catalogs =
@@ -750,11 +915,16 @@ bool AbstractCatalogManager<CatalogT>::MountSubtree(
        iEnd = nested_catalogs.end(); i != iEnd; ++i)
   {
     // Next nesting level
-    PathString nested_path_slash(i->mountpoint);
-    nested_path_slash.Append("/", 1);
-    if (path_slash.StartsWith(nested_path_slash)) {
+    if (path.StartsWith(i->mountpoint)) {
+      // in this case the path doesn't start with
+      // the mountpoint in a file path sense
+      // (e.g. path is /a/bc and mountpoint is /a/b), and will be ignored
+      unsigned mountpoint_len = i->mountpoint.GetLength();
+      if (path_len > mountpoint_len && path.GetChars()[mountpoint_len] != '/')
+        continue;
+
       // Found a nested catalog transition point
-      if (!is_listable && (path_slash == nested_path_slash))
+      if (!is_listable && (path_len == mountpoint_len))
         break;
 
       if (leaf_catalog == NULL)
@@ -793,23 +963,34 @@ CatalogT *AbstractCatalogManager<CatalogT>::MountCatalog(
                                               CatalogT *parent_catalog)
 {
   CatalogT *attached_catalog = NULL;
-  if (IsAttached(mountpoint, &attached_catalog))
+  if (IsAttached(mountpoint, &attached_catalog)) {
     return attached_catalog;
+  }
 
-  string     catalog_path;
-  shash::Any catalog_hash;
-  const LoadError retval =
-    LoadCatalog(mountpoint, hash, &catalog_path, &catalog_hash);
+  CatalogContext ctlg_context(hash, mountpoint, kCtlgLocationMounted);
+
+  if (ctlg_context.IsRootCatalog() && hash.IsNull()) {
+    if (GetNewRootCatalogContext(&ctlg_context) == kLoadFail) {
+      LogCvmfs(kLogCatalog, kLogDebug,
+                                   "failed to retrieve valid root catalog '%s'",
+                                   mountpoint.c_str());
+      return NULL;
+    }
+  }
+
+  const LoadReturn retval = LoadCatalogByHash(&ctlg_context);
   if ((retval == kLoadFail) || (retval == kLoadNoSpace)) {
     LogCvmfs(kLogCatalog, kLogDebug, "failed to load catalog '%s' (%d - %s)",
              mountpoint.c_str(), retval, Code2Ascii(retval));
     return NULL;
   }
 
-  attached_catalog = CreateCatalog(mountpoint, catalog_hash, parent_catalog);
+  attached_catalog = CreateCatalog(ctlg_context.mountpoint(),
+                                   ctlg_context.hash(),
+                                   parent_catalog);
 
   // Attach loaded catalog
-  if (!AttachCatalog(catalog_path, attached_catalog)) {
+  if (!AttachCatalog(ctlg_context.sqlite_path(), attached_catalog)) {
     LogCvmfs(kLogCatalog, kLogDebug, "failed to attach catalog '%s'",
              mountpoint.c_str());
     UnloadCatalog(attached_catalog);
@@ -824,6 +1005,32 @@ CatalogT *AbstractCatalogManager<CatalogT>::MountCatalog(
 }
 
 
+/**
+ * Load a catalog file as a freestanding Catalog object.
+ * Loading of catalogs is implemented by derived classes.
+ */
+template <class CatalogT>
+CatalogT *AbstractCatalogManager<CatalogT>::LoadFreeCatalog(
+                                            const PathString     &mountpoint,
+                                            const shash::Any     &hash)
+{
+  assert(!hash.IsNull());
+  CatalogContext ctlg_context(hash, mountpoint, kCtlgNoLocationNeeded);
+
+  const LoadReturn load_ret = LoadCatalogByHash(&ctlg_context);
+
+  if (load_ret != kLoadNew) {
+    return NULL;
+  }
+
+  CatalogT *catalog = CatalogT::AttachFreely(mountpoint.ToString(),
+                                             ctlg_context.sqlite_path(),
+                                             ctlg_context.hash());
+  catalog->TakeDatabaseFileOwnership();
+  return catalog;
+}
+
+
 /**
  * Attaches a newly created catalog.
  * @param db_path the file on a local file system containing the database
@@ -863,6 +1070,8 @@ bool AbstractCatalogManager<CatalogT>::AttachCatalog(const string &db_path,
   // The revision of the catalog tree is given by the root catalog revision
   if (catalogs_.empty()) {
     revision_cache_ = new_catalog->GetRevision();
+    timestamp_cache_ = new_catalog->GetLastModified();
+    statistics_.catalog_revision->Set(revision_cache_);
     has_authz_cache_ = new_catalog->GetVOMSAuthz(&authz_cache_);
     volatile_flag_ = new_catalog->volatile_flag();
   }
@@ -1017,7 +1226,7 @@ void AbstractCatalogManager<CatalogT>::EnforceSqliteMemLimit() {
     static_cast<char *>(pthread_getspecific(pkey_sqlitemem_));
   if (mem_enforced == NULL) {
     sqlite3_soft_heap_limit(kSqliteMemPerThread);
-    pthread_setspecific(pkey_sqlitemem_, reinterpret_cast<char *>(1));
+    pthread_setspecific(pkey_sqlitemem_, this);
   }
 }
 
@@ -1027,3 +1236,4 @@ void AbstractCatalogManager<CatalogT>::EnforceSqliteMemLimit() {
 
 
 #endif  // CVMFS_CATALOG_MGR_IMPL_H_
+//NOLINTEND
