@@ -16,12 +16,13 @@
 
 #include "catalog_balancer.h"
 #include "catalog_rw.h"
-#include "logging.h"
 #include "manifest.h"
-#include "smalloc.h"
 #include "statistics.h"
 #include "upload.h"
+#include "util/exception.h"
+#include "util/logging.h"
 #include "util/posix.h"
+#include "util/smalloc.h"
 
 using namespace std;  // NOLINT
 
@@ -40,9 +41,10 @@ WritableCatalogManager::WritableCatalogManager(
   perf::Statistics          *statistics,
   bool                       is_balanceable,
   unsigned                   max_weight,
-  unsigned                   min_weight)
+  unsigned                   min_weight,
+  const                      std::string &dir_cache)
   : SimpleCatalogManager(base_hash, stratum0, dir_temp, download_manager,
-      statistics)
+      statistics, false, dir_cache, true /* copy to tmpdir */)
   , spooler_(spooler)
   , enforce_limits_(enforce_limits)
   , nested_kcatalog_limit_(nested_kcatalog_limit)
@@ -127,7 +129,7 @@ manifest::Manifest *WritableCatalogManager::CreateRepository(
   root_entry.linkcount_         = 2;
   string root_path = "";
 
-  // Create the database schema and the inital root entry
+  // Create the database schema and the initial root entry
   {
     UniquePtr<CatalogDatabase> new_clg_db(CatalogDatabase::Create(file_path));
     if (!new_clg_db.IsValid() ||
@@ -243,9 +245,8 @@ void WritableCatalogManager::RemoveFile(const std::string &path) {
   SyncLock();
   WritableCatalog *catalog;
   if (!FindCatalog(parent_path, &catalog)) {
-    LogCvmfs(kLogCatalog, kLogStderr, "catalog for file '%s' cannot be found",
-             file_path.c_str());
-    assert(false);
+    PANIC(kLogStderr, "catalog for file '%s' cannot be found",
+          file_path.c_str());
   }
 
   catalog->RemoveEntry(file_path);
@@ -266,10 +267,8 @@ void WritableCatalogManager::RemoveDirectory(const std::string &path) {
   WritableCatalog *catalog;
   DirectoryEntry parent_entry;
   if (!FindCatalog(parent_path, &catalog, &parent_entry)) {
-    LogCvmfs(kLogCatalog, kLogStderr,
-             "catalog for directory '%s' cannot be found",
-             directory_path.c_str());
-    assert(false);
+    PANIC(kLogStderr, "catalog for directory '%s' cannot be found",
+          directory_path.c_str());
   }
 
   parent_entry.set_linkcount(parent_entry.linkcount() - 1);
@@ -301,21 +300,19 @@ void WritableCatalogManager::Clone(const std::string destination,
   const std::string relative_source = MakeRelativePath(source);
 
   DirectoryEntry source_dirent;
-  if (!LookupPath(relative_source, kLookupSole, &source_dirent)) {
-    LogCvmfs(kLogCatalog, kLogStderr,
-             "catalog for file '%s' cannot be found aborting", source.c_str());
-    assert(false);
+  if (!LookupPath(relative_source, kLookupDefault, &source_dirent)) {
+    PANIC(kLogStderr, "catalog for file '%s' cannot be found, aborting",
+          source.c_str());
   }
   if (source_dirent.IsDirectory()) {
-    LogCvmfs(kLogCatalog, kLogStderr,
-             "Trying to clone a directory: '%s' aborting", source.c_str());
-    assert(false);
+    PANIC(kLogStderr, "Trying to clone a directory: '%s', aborting",
+          source.c_str());
   }
 
   // if the file is already there we remove it and we add it back
   DirectoryEntry check_dirent;
   bool destination_already_present =
-      LookupPath(MakeRelativePath(destination), kLookupSole, &check_dirent);
+      LookupPath(MakeRelativePath(destination), kLookupDefault, &check_dirent);
   if (destination_already_present) {
     this->RemoveFile(destination);
   }
@@ -333,6 +330,138 @@ void WritableCatalogManager::Clone(const std::string destination,
   this->AddFile(destination_dirent, empty_xattrs, destination_dirname);
 }
 
+
+/**
+ * Copies an entire directory tree from the existing from_dir to the
+ * non-existing to_dir. The destination's parent directory must exist. On the
+ * catalog level, the new entries will be identical to the old ones except
+ * for their path hash fields.
+ */
+void WritableCatalogManager::CloneTree(const std::string &from_dir,
+                                       const std::string &to_dir)
+{
+  // Sanitize input paths
+  if (from_dir.empty() || to_dir.empty())
+    PANIC(kLogStderr, "clone tree from or to root impossible");
+
+  const std::string relative_source = MakeRelativePath(from_dir);
+  const std::string relative_dest = MakeRelativePath(to_dir);
+
+  if (relative_source == relative_dest) {
+    PANIC(kLogStderr, "cannot clone tree into itself ('%s')", to_dir.c_str());
+  }
+  if (HasPrefix(relative_dest, relative_source + "/", false /*ignore_case*/)) {
+    PANIC(kLogStderr,
+          "cannot clone tree into sub directory of source '%s' --> '%s'",
+          from_dir.c_str(), to_dir.c_str());
+  }
+
+  DirectoryEntry source_dirent;
+  if (!LookupPath(relative_source, kLookupDefault, &source_dirent)) {
+    PANIC(kLogStderr, "path '%s' cannot be found, aborting", from_dir.c_str());
+  }
+  if (!source_dirent.IsDirectory()) {
+    PANIC(kLogStderr, "CloneTree: source '%s' not a directory, aborting",
+          from_dir.c_str());
+  }
+
+  DirectoryEntry dest_dirent;
+  if (LookupPath(relative_dest, kLookupDefault, &dest_dirent)) {
+    PANIC(kLogStderr, "destination '%s' exists, aborting", to_dir.c_str());
+  }
+
+  const std::string dest_parent = GetParentPath(relative_dest);
+  DirectoryEntry dest_parent_dirent;
+  if (!LookupPath(dest_parent, kLookupDefault, &dest_parent_dirent)) {
+    PANIC(kLogStderr, "destination '%s' not on a known path, aborting",
+          to_dir.c_str());
+  }
+
+  CloneTreeImpl(PathString(from_dir),
+                GetParentPath(to_dir),
+                NameString(GetFileName(to_dir)));
+}
+
+
+/**
+ * Called from CloneTree(), assumes that from_dir and to_dir are sufficiently
+ * sanitized
+ */
+void WritableCatalogManager::CloneTreeImpl(
+  const PathString &source_dir,
+  const std::string &dest_parent_dir,
+  const NameString &dest_name)
+{
+  LogCvmfs(kLogCatalog, kLogDebug, "cloning %s --> %s/%s", source_dir.c_str(),
+           dest_parent_dir.c_str(), dest_name.ToString().c_str());
+  PathString relative_source(MakeRelativePath(source_dir.ToString()));
+
+  DirectoryEntry source_dirent;
+  bool retval = LookupPath(relative_source, kLookupDefault, &source_dirent);
+  assert(retval);
+  assert(!source_dirent.IsBindMountpoint());
+
+  DirectoryEntry dest_dirent(source_dirent);
+  dest_dirent.name_.Assign(dest_name);
+  // Just in case, reset the nested catalog markers
+  dest_dirent.set_is_nested_catalog_mountpoint(false);
+  dest_dirent.set_is_nested_catalog_root(false);
+
+  XattrList xattrs;
+  if (source_dirent.HasXattrs()) {
+    retval = LookupXattrs(relative_source, &xattrs);
+    assert(retval);
+  }
+  AddDirectory(dest_dirent, xattrs, dest_parent_dir);
+
+  std::string dest_dir = dest_parent_dir;
+  if (!dest_dir.empty())
+    dest_dir.push_back('/');
+  dest_dir += dest_name.ToString();
+  if (source_dirent.IsNestedCatalogRoot() ||
+      source_dirent.IsNestedCatalogMountpoint())
+  {
+    CreateNestedCatalog(dest_dir);
+  }
+
+  DirectoryEntryList ls;
+  retval = Listing(relative_source, &ls, false /* expand_symlink */);
+  assert(retval);
+  for (unsigned i = 0; i < ls.size(); ++i) {
+    PathString sub_path(source_dir);
+    assert(!sub_path.IsEmpty());
+    sub_path.Append("/", 1);
+    sub_path.Append(ls[i].name().GetChars(), ls[i].name().GetLength());
+
+    if (ls[i].IsDirectory()) {
+      CloneTreeImpl(sub_path, dest_dir, ls[i].name());
+      continue;
+    }
+
+    // We break hard-links during cloning
+    ls[i].set_hardlink_group(0);
+    ls[i].set_linkcount(1);
+
+    xattrs.Clear();
+    if (ls[i].HasXattrs()) {
+      retval = LookupXattrs(sub_path, &xattrs);
+      assert(retval);
+    }
+
+    if (ls[i].IsChunkedFile()) {
+      FileChunkList chunks;
+      std::string relative_sub_path = MakeRelativePath(sub_path.ToString());
+      retval = ListFileChunks(
+        PathString(relative_sub_path), ls[i].hash_algorithm(), &chunks);
+      assert(retval);
+      AddChunkedFile(ls[i], xattrs, dest_dir, chunks);
+    } else {
+      AddFile(ls[i], xattrs, dest_dir);
+    }
+  }
+}
+
+
 /**
  * Add a new directory to the catalogs.
  * @param entry a DirectoryEntry structure describing the new directory
@@ -352,15 +481,12 @@ void WritableCatalogManager::AddDirectory(const DirectoryEntryBase &entry,
   WritableCatalog *catalog;
   DirectoryEntry parent_entry;
   if (!FindCatalog(parent_path, &catalog, &parent_entry)) {
-    LogCvmfs(kLogCatalog, kLogStderr,
-             "catalog for directory '%s' cannot be found",
-             directory_path.c_str());
-    assert(false);
+    PANIC(kLogStderr, "catalog for directory '%s' cannot be found",
+          directory_path.c_str());
   }
 
   DirectoryEntry fixed_hardlink_count(entry);
   fixed_hardlink_count.set_linkcount(2);
-  // No support for extended attributes on directories yet
   catalog->AddEntry(fixed_hardlink_count, xattrs,
                     directory_path, parent_path);
 
@@ -396,9 +522,8 @@ void WritableCatalogManager::AddFile(
   SyncLock();
   WritableCatalog *catalog;
   if (!FindCatalog(parent_path, &catalog)) {
-    LogCvmfs(kLogCatalog, kLogStderr, "catalog for file '%s' cannot be found",
-             file_path.c_str());
-    assert(false);
+    PANIC(kLogStderr, "catalog for file '%s' cannot be found",
+          file_path.c_str());
   }
 
   assert(!entry.IsRegular() || entry.IsChunkedFile() ||
@@ -412,11 +537,11 @@ void WritableCatalogManager::AddFile(
              "%s: file at %s is larger than %u megabytes (%u). "
              "CernVM-FS works best with small files. "
              "Please remove the file or increase the limit.",
-             enforce_limits_ ? "FATAL" : "WARNING",
-             file_path.c_str(),
-             file_mbyte_limit_,
-             mbytes);
-    assert(!enforce_limits_);
+             enforce_limits_ ? "FATAL" : "WARNING", file_path.c_str(),
+             file_mbyte_limit_, mbytes);
+    if (enforce_limits_)
+      PANIC(kLogStderr, "file at %s is larger than %u megabytes (%u).",
+            file_path.c_str(), file_mbyte_limit_, mbytes);
   }
 
   catalog->AddEntry(entry, xattrs, file_path, parent_path);
@@ -443,9 +568,8 @@ void WritableCatalogManager::AddChunkedFile(
   SyncLock();
   WritableCatalog *catalog;
   if (!FindCatalog(parent_path, &catalog)) {
-    LogCvmfs(kLogCatalog, kLogStderr, "catalog for file '%s' cannot be found",
-             file_path.c_str());
-    assert(false);
+    PANIC(kLogStderr, "catalog for file '%s' cannot be found",
+          file_path.c_str());
   }
 
   for (unsigned i = 0; i < file_chunks.size(); ++i) {
@@ -496,20 +620,22 @@ void WritableCatalogManager::AddHardlinkGroup(
              (parent_path + entries[0].name().ToString()).c_str(),
              file_mbyte_limit_,
              mbytes);
-    assert(!enforce_limits_);
+    if (enforce_limits_)
+      PANIC(kLogStderr, "hard link at %s is larger than %u megabytes (%u)",
+            (parent_path + entries[0].name().ToString()).c_str(),
+            file_mbyte_limit_, mbytes);
   }
 
   SyncLock();
   WritableCatalog *catalog;
   if (!FindCatalog(parent_path, &catalog)) {
-    LogCvmfs(kLogCatalog, kLogStderr,
-             "catalog for hardlink group containing '%s' cannot be found",
-             parent_path.c_str());
-    assert(false);
+    PANIC(kLogStderr,
+          "catalog for hardlink group containing '%s' cannot be found",
+          parent_path.c_str());
   }
 
   // Get a valid hardlink group id for the catalog the group will end up in
-  // TODO(unkown): Compaction
+  // TODO(unknown): Compaction
   uint32_t new_group_id = catalog->GetMaxLinkId() + 1;
   LogCvmfs(kLogCatalog, kLogVerboseMsg, "hardlink group id %u issued",
            new_group_id);
@@ -546,10 +672,9 @@ void WritableCatalogManager::ShrinkHardlinkGroup(const string &remove_path) {
   SyncLock();
   WritableCatalog *catalog;
   if (!FindCatalog(relative_path, &catalog)) {
-    LogCvmfs(kLogCatalog, kLogStderr,
-             "catalog for hardlink group containing '%s' cannot be found",
-             remove_path.c_str());
-    assert(false);
+    PANIC(kLogStderr,
+          "catalog for hardlink group containing '%s' cannot be found",
+          remove_path.c_str());
   }
 
   catalog->IncLinkcount(relative_path, -1);
@@ -578,9 +703,8 @@ void WritableCatalogManager::TouchDirectory(const DirectoryEntryBase &entry,
   // find the catalog to be updated
   WritableCatalog *catalog;
   if (!FindCatalog(parent_path, &catalog)) {
-    LogCvmfs(kLogCatalog, kLogStderr, "catalog for entry '%s' cannot be found",
-             entry_path.c_str());
-    assert(false);
+    PANIC(kLogStderr, "catalog for entry '%s' cannot be found",
+          entry_path.c_str());
   }
 
   catalog->TouchEntry(entry, xattrs, entry_path);
@@ -601,7 +725,7 @@ void WritableCatalogManager::TouchDirectory(const DirectoryEntryBase &entry,
     LogCvmfs(kLogCatalog, kLogVerboseMsg,
              "updating transition point at %s", entry_path.c_str());
 
-    // find and mount nested catalog assciated to this transition point
+    // find and mount nested catalog associated to this transition point
     shash::Any nested_hash;
     uint64_t nested_size;
     retval = catalog->FindNested(transition_path, &nested_hash, &nested_size);
@@ -637,13 +761,13 @@ void WritableCatalogManager::CreateNestedCatalog(const std::string &mountpoint)
   WritableCatalog *old_catalog = NULL;
   DirectoryEntry new_root_entry;
   if (!FindCatalog(nested_root_path, &old_catalog, &new_root_entry)) {
-    LogCvmfs(kLogCatalog, kLogStderr, "failed to create nested catalog '%s': "
-             "mountpoint was not found in current catalog structure",
-             nested_root_path.c_str());
-    assert(false);
+    PANIC(kLogStderr,
+          "failed to create nested catalog '%s': "
+          "mountpoint was not found in current catalog structure",
+          nested_root_path.c_str());
   }
 
-  // Create the database schema and the inital root entry
+  // Create the database schema and the initial root entry
   // for the new nested catalog
   const string database_file_path = CreateTempPath(dir_temp() + "/catalog",
                                                    0666);
@@ -658,7 +782,7 @@ void WritableCatalogManager::CreateNestedCatalog(const std::string &mountpoint)
                                                     // catalog gets VOMS authz
                                                new_root_entry);
   assert(retval);
-  // TODO(rmeusel): we need a way to attach a catalog directy from an open
+  // TODO(rmeusel): we need a way to attach a catalog directly from an open
   // database to remove this indirection
   delete new_catalog_db;
   new_catalog_db = NULL;
@@ -732,10 +856,10 @@ void WritableCatalogManager::RemoveNestedCatalog(const string &mountpoint,
   // Find the catalog which should be removed
   WritableCatalog *nested_catalog = NULL;
   if (!FindCatalog(nested_root_path, &nested_catalog)) {
-    LogCvmfs(kLogCatalog, kLogStderr, "failed to remove nested catalog '%s': "
-             "mountpoint was not found in current catalog structure",
-             nested_root_path.c_str());
-    assert(false);
+    PANIC(kLogStderr,
+          "failed to remove nested catalog '%s': "
+          "mountpoint was not found in current catalog structure",
+          nested_root_path.c_str());
   }
 
   // Check if the found catalog is really the nested catalog to be deleted
@@ -751,10 +875,9 @@ void WritableCatalogManager::RemoveNestedCatalog(const string &mountpoint,
 
   // Delete the catalog database file from the working copy
   if (unlink(nested_catalog->database_path().c_str()) != 0) {
-    LogCvmfs(kLogCatalog, kLogStderr,
-             "unable to delete the removed nested catalog database file '%s'",
-             nested_catalog->database_path().c_str());
-    assert(false);
+    PANIC(kLogStderr,
+          "unable to delete the removed nested catalog database file '%s'",
+          nested_catalog->database_path().c_str());
   }
 
   // Remove the catalog from internal data structures
@@ -763,6 +886,125 @@ void WritableCatalogManager::RemoveNestedCatalog(const string &mountpoint,
 }
 
 
+/**
+ * Swap in a new nested catalog
+ *
+ * The old nested catalog must not have been already attached to the
+ * catalog tree.  This method will not attach the new nested catalog
+ * to the catalog tree.
+ *
+ * @param mountpoint - the path of the nested catalog to be removed
+ * @param new_hash - the hash of the new nested catalog
+ * @param new_size - the size of the new nested catalog
+ */
+void WritableCatalogManager::SwapNestedCatalog(const string &mountpoint,
+                                               const shash::Any &new_hash,
+                                               const uint64_t new_size) {
+  const string nested_root_path = MakeRelativePath(mountpoint);
+  const string parent_path = GetParentPath(nested_root_path);
+  const PathString nested_root_ps = PathString(nested_root_path);
+
+  SyncLock();
+
+  // Find the immediate parent catalog
+  WritableCatalog *parent = NULL;
+  if (!FindCatalog(parent_path, &parent)) {
+    SyncUnlock();  // this is needed for the unittest. otherwise they get stuck
+    PANIC(kLogStderr,
+          "failed to swap nested catalog '%s': could not find parent '%s'",
+          nested_root_path.c_str(), parent_path.c_str());
+  }
+
+  // Get old nested catalog counters
+  Catalog *old_attached_catalog = parent->FindChild(nested_root_ps);
+  Counters old_counters;
+  if (old_attached_catalog) {
+    // Old catalog was already attached (e.g. as a child catalog
+    // attached by a prior call to CreateNestedCatalog()).  Ensure
+    // that it has not been modified, get counters, and detach it.
+    WritableCatalogList list;
+    if (GetModifiedCatalogLeafsRecursively(old_attached_catalog, &list)) {
+      SyncUnlock();
+      PANIC(kLogStderr,
+            "failed to swap nested catalog '%s': already modified",
+            nested_root_path.c_str());
+    }
+    old_counters = old_attached_catalog->GetCounters();
+    DetachSubtree(old_attached_catalog);
+
+  } else {
+    // Old catalog was not attached.  Download a freely attached
+    // version and get counters.
+    shash::Any old_hash;
+    uint64_t old_size;
+    const bool old_found = parent->FindNested(nested_root_ps, &old_hash,
+                                              &old_size);
+    if (!old_found) {
+      SyncUnlock();
+      PANIC(kLogStderr,
+            "failed to swap nested catalog '%s': not found in parent",
+            nested_root_path.c_str());
+    }
+    UniquePtr<Catalog> old_free_catalog(
+      LoadFreeCatalog(nested_root_ps, old_hash));
+    if (!old_free_catalog.IsValid()) {
+      SyncUnlock();
+      PANIC(kLogStderr,
+            "failed to swap nested catalog '%s': failed to load old catalog",
+            nested_root_path.c_str());
+    }
+    old_counters = old_free_catalog->GetCounters();
+  }
+
+  // Load freely attached new catalog
+  UniquePtr<Catalog> new_catalog(LoadFreeCatalog(nested_root_ps, new_hash));
+  if (!new_catalog.IsValid()) {
+    SyncUnlock();
+    PANIC(kLogStderr,
+          "failed to swap nested catalog '%s': failed to load new catalog",
+          nested_root_path.c_str());
+  }
+
+  // Get new catalog root directory entry
+  DirectoryEntry dirent;
+  XattrList xattrs;
+  const bool dirent_found = new_catalog->LookupPath(nested_root_ps, &dirent);
+  if (!dirent_found) {
+    SyncUnlock();
+    PANIC(kLogStderr,
+          "failed to swap nested catalog '%s': missing dirent in new catalog",
+          nested_root_path.c_str());
+  }
+  if (dirent.HasXattrs()) {
+    const bool xattrs_found = new_catalog->LookupXattrsPath(nested_root_ps,
+                                                            &xattrs);
+    if (!xattrs_found) {
+      SyncUnlock();
+      PANIC(kLogStderr,
+            "failed to swap nested catalog '%s': missing xattrs in new catalog",
+            nested_root_path.c_str());
+    }
+  }
+
+  // Swap catalogs
+  parent->RemoveNestedCatalog(nested_root_path, NULL);
+  parent->InsertNestedCatalog(nested_root_path, NULL, new_hash, new_size);
+
+  // Update parent directory entry
+  dirent.set_is_nested_catalog_mountpoint(true);
+  dirent.set_is_nested_catalog_root(false);
+  parent->UpdateEntry(dirent, nested_root_path);
+  parent->TouchEntry(dirent, xattrs, nested_root_path);
+
+  // Update counters
+  DeltaCounters delta = Counters::Diff(old_counters,
+                                       new_catalog->GetCounters());
+  delta.PopulateToParent(&parent->delta_counters_);
+
+  SyncUnlock();
+}
+
+
 /**
  * Checks if a nested catalog starts at this path.  The path must be valid.
  */
@@ -773,9 +1015,8 @@ bool WritableCatalogManager::IsTransitionPoint(const string &mountpoint) {
   WritableCatalog *catalog;
   DirectoryEntry entry;
   if (!FindCatalog(path, &catalog, &entry)) {
-    LogCvmfs(kLogCatalog, kLogStderr,
-             "catalog for directory '%s' cannot be found", path.c_str());
-    assert(false);
+    PANIC(kLogStderr, "catalog for directory '%s' cannot be found",
+          path.c_str());
   }
   const bool result = entry.IsNestedCatalogRoot();
   SyncUnlock();
@@ -816,10 +1057,10 @@ bool WritableCatalogManager::Commit(const bool           stop_for_tweaks,
   if (manual_revision > 0) {
     const uint64_t revision = root_catalog->GetRevision();
     if (revision >= manual_revision) {
-      LogCvmfs(kLogCatalog, kLogStderr, "Manual revision (%d) must not be "
-                                        "smaller than the current root "
-                                        "catalog's (%d). Skipped!",
-                                        manual_revision, revision);
+      LogCvmfs(kLogCatalog, kLogStderr,
+               "Manual revision (%" PRIu64 ") must not be "
+               "smaller than the current root catalog's (%" PRIu64
+               "). Skipped!", manual_revision, revision);
     } else {
       // Gets incremented by FinalizeCatalog() afterwards!
       root_catalog->SetRevision(manual_revision - 1);
@@ -853,12 +1094,12 @@ bool WritableCatalogManager::Commit(const bool           stop_for_tweaks,
 
 /**
  * Handles the snapshotting of dirty (i.e. modified) catalogs while trying to
- * parallize the compression and upload as much as possible. We use a parallel
+ * parallelize the compression and upload as much as possible. We use a parallel
  * depth first post order tree traversal based on 'continuations'.
  *
  * The idea is as follows:
  *  1. find all leaf-catalogs (i.e. dirty catalogs with no dirty children)
- *     --> these can be processed and uploaded immedately and independently
+ *     --> these can be processed and uploaded immediately and independently
  *         see WritableCatalogManager::GetModifiedCatalogLeafs()
  *  2. annotate non-leaf catalogs with their number of dirty children
  *     --> a finished child will notify it's parent and decrement this number
@@ -875,7 +1116,7 @@ bool WritableCatalogManager::Commit(const bool           stop_for_tweaks,
  *       catalogs.
  *
  * TODO(rmeusel): since all leaf catalogs are finalized in the main thread, we
- *                sacrafice some potential concurrency for simplicity.
+ *                sacrifice some potential concurrency for simplicity.
  */
 WritableCatalogManager::CatalogInfo WritableCatalogManager::SnapshotCatalogs(
                                                    const bool stop_for_tweaks) {
@@ -953,14 +1194,16 @@ void WritableCatalogManager::FinalizeCatalog(WritableCatalog *catalog,
   if ((catalog_limit > 0) &&
       (catalog->GetCounters().GetSelfEntries() > catalog_limit)) {
     LogCvmfs(kLogCatalog, kLogStderr,
-             "%s: catalog at %s has more than %u entries (%u). "
+             "%s: catalog at %s has more than %lu entries (%lu). "
              "Large catalogs stress the CernVM-FS transport infrastructure. "
              "Please split it into nested catalogs or increase the limit.",
              enforce_limits_ ? "FATAL" : "WARNING",
              (catalog->IsRoot() ? "/" : catalog->mountpoint().c_str()),
-             catalog_limit,
-             catalog->GetCounters().GetSelfEntries());
-    assert(!enforce_limits_);
+             catalog_limit, catalog->GetCounters().GetSelfEntries());
+    if (enforce_limits_)
+      PANIC(kLogStderr, "catalog at %s has more than %u entries (%u). ",
+            (catalog->IsRoot() ? "/" : catalog->mountpoint().c_str()),
+            catalog_limit, catalog->GetCounters().GetSelfEntries());
   }
 
   // allow for manual adjustments in the catalog
@@ -987,14 +1230,43 @@ void WritableCatalogManager::ScheduleCatalogProcessing(
   spooler_->ProcessCatalog(catalog->database_path());
 }
 
+/**
+ * Copy catalog to local cache.server
+ * Must be an atomic write into the cache_dir
+ * As such: create a temporary copy in cache_dir/txn and then do a 
+ * `rename` (which is atomic) to the actual cache path
+ * 
+ * @returns true on success, otherwise false
+ */
+bool WritableCatalogManager::CopyCatalogToLocalCache(
+                                          const upload::SpoolerResult &result) {
+  std::string tmp_catalog_path;
+  const std::string cache_catalog_path = dir_cache_ + "/"
+                                  + result.content_hash.MakePathWithoutSuffix();
+  FILE *fcatalog = CreateTempFile(dir_cache_ + "/txn/catalog", 0666,
+                                                        "w", &tmp_catalog_path);
+  if (!fcatalog) {
+    PANIC(kLogDebug | kLogStderr,
+                               "Creating file for temporary catalog failed: %s",
+                               tmp_catalog_path.c_str());
+  }
+  CopyPath2File(result.local_path.c_str(), fcatalog);
+  (void) fclose(fcatalog);
+
+  if (rename(tmp_catalog_path.c_str(), cache_catalog_path.c_str()) != 0) {
+    PANIC(kLogDebug | kLogStderr,
+                         "Failed to copy catalog from %s to cache %s",
+                         result.local_path.c_str(), cache_catalog_path.c_str());
+  }
+  return true;
+}
 
 void WritableCatalogManager::CatalogUploadCallback(
                           const upload::SpoolerResult &result,
                           const CatalogUploadContext   catalog_upload_context) {
   if (result.return_code != 0) {
-    LogCvmfs(kLogCatalog, kLogStderr, "failed to upload '%s' (retval: %d)",
-             result.local_path.c_str(), result.return_code);
-    assert(false);
+    PANIC(kLogStderr, "failed to upload '%s' (retval: %d)",
+          result.local_path.c_str(), result.return_code);
   }
 
   // retrieve the catalog object based on the callback information
@@ -1011,6 +1283,10 @@ void WritableCatalogManager::CatalogUploadCallback(
   uint64_t catalog_size = GetFileSize(result.local_path);
   assert(catalog_size > 0);
 
+  if (UseLocalCache()) {
+    CopyCatalogToLocalCache(result);
+  }
+
   SyncLock();
   if (catalog->HasParent()) {
     // finalized nested catalogs will update their parent's pointer and schedule
@@ -1047,14 +1323,14 @@ void WritableCatalogManager::CatalogUploadCallback(
     catalog_upload_context.root_catalog_info->Set(root_catalog_info);
     SyncUnlock();
   } else {
-    assert(false && "inconsistent state detected");
+    PANIC(kLogStderr, "inconsistent state detected");
   }
 }
 
 
 /**
  * Finds dirty catalogs that can be snapshot right away and annotates all the
- * other catalogs with their number of dirty decendants.
+ * other catalogs with their number of dirty descendants.
  * Note that there is a convenience wrapper to start the recursion:
  *   WritableCatalogManager::GetModifiedCatalogLeafs()
  *
@@ -1160,10 +1436,14 @@ void WritableCatalogManager::CatalogUploadSerializedCallback(
   const CatalogUploadContext unused)
 {
   if (result.return_code != 0) {
-    LogCvmfs(kLogCatalog, kLogStderr, "failed to upload '%s' (retval: %d)",
-             result.local_path.c_str(), result.return_code);
-    assert(false);
+    PANIC(kLogStderr, "failed to upload '%s' (retval: %d)",
+          result.local_path.c_str(), result.return_code);
   }
+
+  if (UseLocalCache()) {
+    CopyCatalogToLocalCache(result);
+  }
+
   unlink(result.local_path.c_str());
 }
 
@@ -1194,8 +1474,8 @@ WritableCatalogManager::SnapshotCatalogsSerialized(
     if (!zlib::CompressPath2Null((*i)->database_path(),
                                  &hash_catalog))
     {
-      PrintError("could not compress catalog " + (*i)->mountpoint().ToString());
-      assert(false);
+      PANIC(kLogStderr, "could not compress catalog %s",
+            (*i)->mountpoint().ToString().c_str());
     }
 
     int64_t catalog_size = GetFileSize((*i)->database_path());
@@ -1213,7 +1493,7 @@ WritableCatalogManager::SnapshotCatalogsSerialized(
       root_catalog_info.content_hash = hash_catalog;
       root_catalog_info.revision = (*i)->GetRevision();
     } else {
-      assert(false && "inconsistent state detected");
+      PANIC(kLogStderr, "inconsistent state detected");
     }
 
     spooler_->ProcessCatalog((*i)->database_path());
