@@ -10,8 +10,9 @@
 #include <cstdio>
 #include <cstdlib>
 
-#include "logging.h"
-#include "util_concurrency.h"
+#include "util/concurrency.h"
+#include "util/exception.h"
+#include "util/logging.h"
 #include "xattr.h"
 
 using namespace std;  // NOLINT
@@ -266,8 +267,8 @@ void WritableCatalog::AddFileChunk(const std::string &entry_path,
 
   shash::Md5 path_hash((shash::AsciiPtr(entry_path)));
 
-  LogCvmfs(kLogCatalog, kLogVerboseMsg, "adding chunk for %s from offset %d "
-                                        "and chunk size: %d bytes",
+  LogCvmfs(kLogCatalog, kLogVerboseMsg, "adding chunk for %s from offset %ld "
+                                        "and chunk size: %ld bytes",
            entry_path.c_str(),
            chunk.offset(),
            chunk.offset() + chunk.size());
@@ -527,7 +528,7 @@ void WritableCatalog::InsertNestedCatalog(const string &mountpoint,
 
 
 /**
- * Registeres a snapshot in /.cvmfs/snapshots. Note that bind mountpoints are
+ * Registers a snapshot in /.cvmfs/snapshots. Note that bind mountpoints are
  * not universally handled: in Partition and MergeIntoParent, bind mountpoint
  * handling is missing!
  */
@@ -589,7 +590,7 @@ void WritableCatalog::RemoveNestedCatalog(const string &mountpoint,
 
 
 /**
- * Unregisteres a snapshot from /.cvmfs/snapshots. Note that bind mountpoints
+ * Unregisters a snapshot from /.cvmfs/snapshots. Note that bind mountpoints
  * are not universally handled: in Partition and MergeIntoParent, bind
  * mountpoint handling is missing!
  */
@@ -669,16 +670,9 @@ void WritableCatalog::RemoveFromParent() {
 
   // Remove the nested catalog reference for this nested catalog.
   // From now on this catalog will be dangling!
-  Catalog* child_catalog;
-  parent->RemoveNestedCatalog(this->mountpoint().ToString(), &child_catalog);
-
-  const Counters& child_counters = child_catalog->GetCounters();
-
-  parent->delta_counters_.subtree.directories -= 1;
-  parent->delta_counters_.subtree.file_size -= child_counters.self.file_size;
-  parent->delta_counters_.subtree.regular_files -=
-      child_counters.self.regular_files;
-  parent->delta_counters_.subtree.symlinks -= child_counters.self.symlinks;
+  parent->RemoveNestedCatalog(this->mountpoint().ToString(), NULL);
+  parent->delta_counters_.RemoveFromSubtree(
+    Counters::Diff(Counters(), GetCounters()));
 }
 
 
@@ -704,7 +698,7 @@ void WritableCatalog::CopyCatalogsToParent() {
 void WritableCatalog::CopyToParent() {
   // We could simply copy all entries from this database to the 'other' database
   // BUT: 1. this would create collisions in hardlink group IDs.
-  //         therefor we first update all hardlink group IDs to fit behind the
+  //         therefore we first update all hardlink group IDs to fit behind the
   //         ones in the 'other' database
   //      2. the root entry of the nested catalog is present twice:
   //         1. in the parent directory (as mount point) and
@@ -800,9 +794,7 @@ void WritableCatalog::VacuumDatabaseIfNecessary() {
              ratio * 100.0,
              reason.c_str());
     if (!db.Vacuum()) {
-      LogCvmfs(kLogCatalog, kLogStderr, "failed (SQLite: %s)",
-               db.GetLastErrorMsg().c_str());
-      assert(false);
+      PANIC(kLogStderr, "failed (SQLite: %s)", db.GetLastErrorMsg().c_str());
     }
     LogCvmfs(kLogCatalog, kLogStdout, "done");
   }
