@@ -15,14 +15,14 @@
 #include <cassert>
 #include <string>
 
-#include "atomic.h"
 #include "bigvector.h"
 #include "catalog_mgr.h"
+#include "crypto/hash.h"
 #include "file_chunk.h"
 #include "glue_buffer.h"
-#include "hash.h"
 #include "shortstring.h"
 #include "util/algorithm.h"
+#include "util/atomic.h"
 
 namespace compat {
 
