@@ -9,14 +9,13 @@
 #define ENOATTR ENODATA  /**< instead of including attr/xattr.h */
 
 #include <sys/xattr.h>
-#include "cvmfs_config.h"
+
 #include "libcvmfs_int.h"
 
 #include <dirent.h>
 #include <errno.h>
 #include <fcntl.h>
 #include <google/dense_hash_map>
-#include <openssl/crypto.h>
 #include <pthread.h>
 #include <stddef.h>
 #include <stdint.h>
@@ -46,30 +45,31 @@
 #include <string>
 #include <vector>
 
-#include "atomic.h"
 #include "cache_posix.h"
 #include "catalog.h"
 #include "catalog_mgr_client.h"
 #include "clientctx.h"
-#include "compression.h"
+#include "compression/compression.h"
+#include "crypto/crypto_util.h"
+#include "crypto/hash.h"
+#include "crypto/signature.h"
 #include "directory_entry.h"
-#include "download.h"
 #include "duplex_sqlite3.h"
 #include "fetch.h"
 #include "globals.h"
-#include "hash.h"
+#include "interrupt.h"
 #include "libcvmfs.h"
-#include "logging.h"
 #include "lru_md.h"
-#include "murmur.h"
-#include "platform.h"
+#include "network/download.h"
 #include "quota.h"
 #include "shortstring.h"
-#include "signature.h"
-#include "smalloc.h"
 #include "sqlitemem.h"
 #include "sqlitevfs.h"
+#include "util/atomic.h"
+#include "util/logging.h"
+#include "util/murmur.hxx"
 #include "util/posix.h"
+#include "util/smalloc.h"
 #include "util/string.h"
 #include "xattr.h"
 
@@ -101,14 +101,7 @@ loader::Failures LibGlobals::Initialize(OptionsManager *options_mgr) {
   assert(instance_ != NULL);
 
   // Multi-threaded libcrypto (otherwise done by the loader)
-  instance_->libcrypto_locks_ = static_cast<pthread_mutex_t *>(
-    OPENSSL_malloc(CRYPTO_num_locks() * sizeof(pthread_mutex_t)));
-  for (int i = 0; i < CRYPTO_num_locks(); ++i) {
-    int retval = pthread_mutex_init(&(instance_->libcrypto_locks_[i]), NULL);
-    assert(retval == 0);
-  }
-  CRYPTO_set_id_callback(LibGlobals::CallbackLibcryptoThreadId);
-  CRYPTO_set_locking_callback(LibGlobals::CallbackLibcryptoLock);
+  crypto::SetupLibcryptoMt();
 
   FileSystem::FileSystemInfo fs_info;
   fs_info.name = "libcvmfs";
@@ -146,7 +139,6 @@ void LibGlobals::CleanupInstance() {
 LibGlobals::LibGlobals()
   : options_mgr_(NULL)
   , file_system_(NULL)
-  , libcrypto_locks_(NULL)
 { }
 
 
@@ -154,41 +146,7 @@ LibGlobals::~LibGlobals() {
   delete file_system_;
   delete options_mgr_;
 
-  if (libcrypto_locks_) {
-    CRYPTO_set_locking_callback(NULL);
-    for (int i = 0; i < CRYPTO_num_locks(); ++i)
-      pthread_mutex_destroy(&(libcrypto_locks_[i]));
-    OPENSSL_free(libcrypto_locks_);
-  }
-}
-
-
-void LibGlobals::CallbackLibcryptoLock(
-  int mode,
-  int type,
-  const char *file,
-  int line)
-{
-  (void)file;
-  (void)line;
-
-  int retval;
-  LibGlobals *globals = LibGlobals::GetInstance();
-  pthread_mutex_t *locks = globals->libcrypto_locks_;
-  pthread_mutex_t *lock = &(locks[type]);
-
-  if (mode & CRYPTO_LOCK) {
-    retval = pthread_mutex_lock(lock);
-  } else {
-    retval = pthread_mutex_unlock(lock);
-  }
-  assert(retval == 0);
-}
-
-
-// Type unsigned long required by libcrypto (openssl)
-unsigned long LibGlobals::CallbackLibcryptoThreadId() {  // NOLINT
-  return platform_gettid();
+  crypto::CleanupLibcryptoMt();
 }
 
 
@@ -237,7 +195,7 @@ bool LibContext::GetDirentForPath(const PathString         &path,
     return dirent->GetSpecial() != catalog::kDirentNegative;
 
   // TODO(jblomer): not twice md5 calculation
-  if (mount_point_->catalog_mgr()->LookupPath(path, catalog::kLookupSole,
+  if (mount_point_->catalog_mgr()->LookupPath(path, catalog::kLookupDefault,
                                               dirent))
   {
     mount_point_->md5path_cache()->Insert(md5path, *dirent);
@@ -276,9 +234,26 @@ void LibContext::AppendStringToList(char const   *str,
 }
 
 
+void LibContext::AppendStatToList(const cvmfs_stat_t   st,
+                                  cvmfs_stat_t       **buf,
+                                  size_t              *listlen,
+                                  size_t              *buflen)
+{
+  if (*listlen + 1 >= *buflen) {
+       size_t newbuflen = (*listlen)*2 + 5;
+       *buf = reinterpret_cast<cvmfs_stat_t *>(
+         realloc(*buf, sizeof(cvmfs_stat_t) * newbuflen));
+       assert(*buf);
+       *buflen = newbuflen;
+       assert(*listlen < *buflen);
+  }
+  (*buf)[(*listlen)].info = st.info;
+  (*buf)[(*listlen)++].name = st.name;
+}
+
 int LibContext::GetAttr(const char *c_path, struct stat *info) {
   perf::Inc(file_system()->n_fs_stat());
-  ClientCtxGuard ctxg(geteuid(), getegid(), getpid());
+  ClientCtxGuard ctxg(geteuid(), getegid(), getpid(), &default_interrupt_cue_);
 
   LogCvmfs(kLogCvmfs, kLogDebug, "cvmfs_getattr (stat) for path: %s", c_path);
 
@@ -316,7 +291,7 @@ void LibContext::CvmfsAttrFromDirent(
 
 
 int LibContext::GetExtAttr(const char *c_path, struct cvmfs_attr *info) {
-  ClientCtxGuard ctxg(geteuid(), getegid(), getpid());
+  ClientCtxGuard ctxg(geteuid(), getegid(), getpid(), &default_interrupt_cue_);
 
   LogCvmfs(kLogCvmfs, kLogDebug, "cvmfs_getattr (stat) for path: %s", c_path);
 
@@ -367,7 +342,7 @@ int LibContext::GetExtAttr(const char *c_path, struct cvmfs_attr *info) {
 int LibContext::Readlink(const char *c_path, char *buf, size_t size) {
   perf::Inc(file_system()->n_fs_readlink());
   LogCvmfs(kLogCvmfs, kLogDebug, "cvmfs_readlink on path: %s", c_path);
-  ClientCtxGuard ctxg(geteuid(), getegid(), getpid());
+  ClientCtxGuard ctxg(geteuid(), getegid(), getpid(), &default_interrupt_cue_);
 
   PathString p;
   p.Assign(c_path, strlen(c_path));
@@ -399,7 +374,7 @@ int LibContext::ListDirectory(
   bool self_reference
 ) {
   LogCvmfs(kLogCvmfs, kLogDebug, "cvmfs_listdir on path: %s", c_path);
-  ClientCtxGuard ctxg(geteuid(), getegid(), getpid());
+  ClientCtxGuard ctxg(geteuid(), getegid(), getpid(), &default_interrupt_cue_);
 
   if (c_path[0] == '/' && c_path[1] == '\0') {
     // root path is expected to be "", not "/"
@@ -448,11 +423,53 @@ int LibContext::ListDirectory(
   return 0;
 }
 
+int LibContext::ListDirectoryStat(
+  const char *c_path,
+  cvmfs_stat_t **buf,
+  size_t *listlen,
+  size_t *buflen) {
+  LogCvmfs(kLogCvmfs, kLogDebug, "cvmfs_listdir_stat on path: %s", c_path);
+  ClientCtxGuard ctxg(geteuid(), getegid(), getpid(), &default_interrupt_cue_);
+
+  if (c_path[0] == '/' && c_path[1] == '\0') {
+    // root path is expected to be "", not "/"
+    c_path = "";
+  }
+
+  PathString path;
+  path.Assign(c_path, strlen(c_path));
+
+  catalog::DirectoryEntry d;
+  const bool found = GetDirentForPath(path, &d);
+
+  if (!found) {
+    return -ENOENT;
+  }
+
+  if (!d.IsDirectory()) {
+    return -ENOTDIR;
+  }
+
+  // Build listing
+  catalog::StatEntryList listing_from_catalog;
+  if (!mount_point_->catalog_mgr()->ListingStat(path, &listing_from_catalog)) {
+    return -EIO;
+  }
+  for (unsigned i = 0; i < listing_from_catalog.size(); ++i) {
+    cvmfs_stat_t st;
+    st.info = listing_from_catalog.AtPtr(i)->info;
+    st.name = strdup(listing_from_catalog.AtPtr(i)->name.c_str());
+    AppendStatToList(st, buf, listlen, buflen);
+  }
+
+  return 0;
+}
+
 int LibContext::GetNestedCatalogAttr(
   const char *c_path,
   struct cvmfs_nc_attr *nc_attr
 ) {
-  ClientCtxGuard ctxg(geteuid(), getegid(), getpid());
+  ClientCtxGuard ctxg(geteuid(), getegid(), getpid(), &default_interrupt_cue_);
   LogCvmfs(kLogCvmfs, kLogDebug,
     "cvmfs_stat_nc (cvmfs_nc_attr) : %s", c_path);
 
@@ -470,10 +487,29 @@ int LibContext::GetNestedCatalogAttr(
     return -ENOENT;
   }
 
+  std::string subcat_path;
+  shash::Any tmp_hash;
+  std::map<std::string, uint64_t> counters =
+    mount_point_->catalog_mgr()->
+      LookupCounters(p, &subcat_path, &tmp_hash).GetValues();
+
   // Set values of the passed structure
   nc_attr->mountpoint = strdup(mountpoint.ToString().c_str());
   nc_attr->hash = strdup(hash.ToString().c_str());
   nc_attr->size = size;
+
+  nc_attr->ctr_regular = counters["regular"];
+  nc_attr->ctr_symlink = counters["symlink"];
+  nc_attr->ctr_special = counters["special"];
+  nc_attr->ctr_dir = counters["dir"];
+  nc_attr->ctr_nested = counters["nested"];
+  nc_attr->ctr_chunked = counters["chunked"];
+  nc_attr->ctr_chunks = counters["chunks"];
+  nc_attr->ctr_file_size = counters["file_size"];
+  nc_attr->ctr_chunked_size = counters["chunked_size"];
+  nc_attr->ctr_xattr = counters["xattr"];
+  nc_attr->ctr_external = counters["external"];
+  nc_attr->ctr_external_file_size = counters["external_file_size"];
   return 0;
 }
 
@@ -483,7 +519,7 @@ int LibContext::ListNestedCatalogs(
   char ***buf,
   size_t *buflen
 ) {
-  ClientCtxGuard ctxg(geteuid(), getegid(), getpid());
+  ClientCtxGuard ctxg(geteuid(), getegid(), getpid(), &default_interrupt_cue_);
   LogCvmfs(kLogCvmfs, kLogDebug,
     "cvmfs_list_nc on path: %s", c_path);
 
@@ -516,7 +552,7 @@ int LibContext::ListNestedCatalogs(
 
 int LibContext::Open(const char *c_path) {
   LogCvmfs(kLogCvmfs, kLogDebug, "cvmfs_open on path: %s", c_path);
-  ClientCtxGuard ctxg(geteuid(), getegid(), getpid());
+  ClientCtxGuard ctxg(geteuid(), getegid(), getpid(), &default_interrupt_cue_);
 
   int fd = -1;
   catalog::DirectoryEntry dirent;
@@ -541,7 +577,7 @@ int LibContext::Open(const char *c_path) {
     {
       LogCvmfs(kLogCvmfs, kLogDebug | kLogSyslogErr, "file %s is marked as "
                "'chunked', but no chunks found.", path.c_str());
-      perf::Inc(file_system()->n_io_error());
+      file_system()->io_error_info()->AddIoError();
       delete chunks;
       return -EIO;
     }
@@ -555,12 +591,14 @@ int LibContext::Open(const char *c_path) {
   cvmfs::Fetcher *this_fetcher = dirent.IsExternalFile()
     ? mount_point_->external_fetcher()
     : mount_point_->fetcher();
-  fd = this_fetcher->Fetch(
-    dirent.checksum(),
-    dirent.size(),
-    string(path.GetChars(), path.GetLength()),
-    dirent.compression_algorithm(),
-    CacheManager::kTypeRegular);
+  CacheManager::Label label;
+  label.path = std::string(path.GetChars(), path.GetLength());
+  label.size = dirent.size();
+  label.zip_algorithm = dirent.compression_algorithm();
+  if (dirent.IsExternalFile())
+    label.flags |= CacheManager::kLabelExternal;
+  fd =
+    this_fetcher->Fetch(CacheManager::LabeledObject(dirent.checksum(), label));
   perf::Inc(file_system()->n_fs_open());
 
   if (fd >= 0) {
@@ -576,7 +614,7 @@ int LibContext::Open(const char *c_path) {
     }
   }
 
-  perf::Inc(file_system()->n_io_error());
+  file_system()->io_error_info()->AddIoError();
   return fd;
 }
 
@@ -588,7 +626,8 @@ int64_t LibContext::Pread(
   uint64_t off)
 {
   if (fd & kFdChunked) {
-    ClientCtxGuard ctxg(geteuid(), getegid(), getpid());
+    ClientCtxGuard ctxg(geteuid(), getegid(), getpid(),
+                        &default_interrupt_cue_);
     const int chunk_handle = fd & ~kFdChunked;
     SimpleChunkTables::OpenChunks open_chunks =
       mount_point_->simple_chunk_tables()->Get(chunk_handle);
@@ -607,23 +646,21 @@ int64_t LibContext::Pread(
       ChunkFd *chunk_fd = open_chunks.chunk_fd;
       if ((chunk_fd->fd == -1) || (chunk_fd->chunk_idx != chunk_idx)) {
         if (chunk_fd->fd != -1) file_system()->cache_mgr()->Close(chunk_fd->fd);
+        cvmfs::Fetcher *this_fetcher = open_chunks.chunk_reflist.external_data
+          ? mount_point_->external_fetcher()
+          : mount_point_->fetcher();
+        CacheManager::Label label;
+        label.path = std::string(open_chunks.chunk_reflist.path.GetChars(),
+                                 open_chunks.chunk_reflist.path.GetLength());
+        label.size = chunk_list->AtPtr(chunk_idx)->size();
+        label.zip_algorithm = compression_alg;
+        label.flags |= CacheManager::kLabelChunked;
         if (open_chunks.chunk_reflist.external_data) {
-          chunk_fd->fd = mount_point_->external_fetcher()->Fetch(
-            chunk_list->AtPtr(chunk_idx)->content_hash(),
-            chunk_list->AtPtr(chunk_idx)->size(),
-            "no path info",
-            compression_alg,
-            CacheManager::kTypeRegular,
-            open_chunks.chunk_reflist.path.ToString(),
-            chunk_list->AtPtr(chunk_idx)->offset());
-        } else {
-          chunk_fd->fd = mount_point_->fetcher()->Fetch(
-            chunk_list->AtPtr(chunk_idx)->content_hash(),
-            chunk_list->AtPtr(chunk_idx)->size(),
-            "no path info",
-            compression_alg,
-            CacheManager::kTypeRegular);
+          label.flags |= CacheManager::kLabelExternal;
+          label.range_offset = chunk_list->AtPtr(chunk_idx)->offset();
         }
+        chunk_fd->fd = this_fetcher->Fetch(CacheManager::LabeledObject(
+          chunk_list->AtPtr(chunk_idx)->content_hash(), label));
         if (chunk_fd->fd < 0) {
           chunk_fd->fd = -1;
           return -EIO;
@@ -646,7 +683,7 @@ int64_t LibContext::Pread(
         offset_in_chunk);
 
       if (bytes_fetched < 0) {
-        LogCvmfs(kLogCvmfs, kLogSyslogErr, "read err no %d (%s)",
+        LogCvmfs(kLogCvmfs, kLogSyslogErr, "read err no %ld (%s)",
                  bytes_fetched,
                  open_chunks.chunk_reflist.path.ToString().c_str());
         return -bytes_fetched;
@@ -681,3 +718,36 @@ int LibContext::Close(int fd) {
   }
   return 0;
 }
+
+int LibContext::Remount() {
+  LogCvmfs(kLogCvmfs, kLogDebug, "remounting root catalog");
+  catalog::LoadReturn retval = mount_point_->catalog_mgr()->RemountDryrun();
+
+  switch (retval) {
+    case catalog::kLoadUp2Date:
+      LogCvmfs(kLogCvmfs, kLogDebug, "catalog up to date");
+      return 0;
+
+    case catalog::kLoadNew:
+      retval = mount_point_->catalog_mgr()->Remount();
+
+      if (retval != catalog::kLoadUp2Date && retval != catalog::kLoadNew) {
+        LogCvmfs(kLogCvmfs, kLogDebug,
+                              "Remount requested to switch catalog but failed");
+        return -1;
+      }
+
+      mount_point_->ReEvaluateAuthz();
+      LogCvmfs(kLogCvmfs, kLogDebug, "switched to catalog revision %" PRIu64,
+               mount_point_->catalog_mgr()->GetRevision());
+      return 0;
+
+    default:
+      return -1;
+  }
+}
+
+
+uint64_t LibContext::GetRevision() {
+  return mount_point_->catalog_mgr()->GetRevision();
+}
