@@ -12,13 +12,12 @@
 #define ENOATTR ENODATA  /**< instead of including attr/xattr.h */
 #define _FILE_OFFSET_BITS 64
 
-#include "cvmfs_config.h"
+
 #include "loader.h"
 
 #include <dlfcn.h>
 #include <errno.h>
 #include <fcntl.h>
-#include <openssl/crypto.h>
 #include <sched.h>
 #include <signal.h>
 #include <stddef.h>
@@ -37,16 +36,16 @@
 #include <string>
 #include <vector>
 
-#include "atomic.h"
 #include "duplex_fuse.h"
-#include "duplex_ssl.h"
 #include "fence.h"
 #include "fuse_main.h"
 #include "loader_talk.h"
-#include "logging.h"
 #include "options.h"
-#include "platform.h"
 #include "sanitizer.h"
+#include "util/atomic.h"
+#include "util/exception.h"
+#include "util/logging.h"
+#include "util/platform.h"
 #include "util/posix.h"
 #include "util/string.h"
 
@@ -128,6 +127,8 @@ string *mount_point_ = NULL;
 string *config_files_ = NULL;
 string *socket_path_ = NULL;
 string *usyslog_path_ = NULL;
+int fuse3_max_threads_ = 0;
+int fuse3_idle_threads_ = 0;
 uid_t uid_ = 0;
 gid_t gid_ = 0;
 bool single_threaded_ = false;
@@ -177,7 +178,7 @@ static void Usage(const string &exename) {
     "  -o allow_other       allow access to other users\n"
     "  -o allow_root        allow access to root\n"
     "  -o nonempty          allow mounts over non-empty directory\n",
-    PACKAGE_VERSION, exename.c_str());
+    CVMFS_VERSION, exename.c_str());
 }
 
 /**
@@ -324,6 +325,18 @@ static void stub_forget(
 }
 
 
+#if (FUSE_VERSION >= 29)
+static void stub_forget_multi(
+  fuse_req_t req,
+  size_t count,
+  struct fuse_forget_data *forgets
+) {
+  FenceGuard fence_guard(fence_reload_);
+  cvmfs_exports_->cvmfs_operations.forget_multi(req, count, forgets);
+}
+#endif
+
+
 /**
  * The callback used when fuse is parsing all the options
  * We separate CVMFS options from FUSE options here.
@@ -368,7 +381,7 @@ static int ParseFuseOptions(void *data __attribute__((unused)), const char *arg,
       exit(0);
     case KEY_VERSION:
       LogCvmfs(kLogCvmfs, kLogStdout, "CernVM-FS version %s\n",
-               PACKAGE_VERSION);
+               CVMFS_VERSION);
       exit(0);
     case KEY_FOREGROUND:
       foreground_ = true;
@@ -385,12 +398,10 @@ static int ParseFuseOptions(void *data __attribute__((unused)), const char *arg,
       parse_options_only_ = true;
       return 0;
     default:
-      LogCvmfs(kLogCvmfs, kLogStderr, "internal option parsing error");
-      abort();
+      PANIC(kLogStderr, "internal option parsing error");
   }
 }
 
-
 static fuse_args *ParseCmdLine(int argc, char *argv[]) {
   struct fuse_args *mount_options = new fuse_args();
   CvmfsOptions cvmfs_options;
@@ -487,7 +498,7 @@ static CvmfsExports *LoadLibrary(const bool debug_mode,
   library_name = platform_libname(library_name);
   string error_messages;
 
-  static vector<string> library_paths;  // TODO(rmeusel): C++11 initializer
+  vector<string> library_paths;  // TODO(rmeusel): C++11 initializer
   if (library_paths.empty()) {
     library_paths.push_back(local_lib_path + library_name);
     library_paths.push_back("/usr/lib/"   + library_name);
@@ -532,10 +543,17 @@ static CvmfsExports *LoadLibrary(const bool debug_mode,
   return *exports_ptr;
 }
 
-
-Failures Reload(const int fd_progress, const bool stop_and_go) {
+Failures Reload(const int fd_progress, const bool stop_and_go,
+                const ReloadMode reload_mode) {
   int retval;
 
+  // for legacy call we take the current state of debug_mode_
+  if (reload_mode == kReloadDebug) {
+    debug_mode_ = true;
+  } else if (reload_mode == kReloadNoDebug) {
+    debug_mode_ = false;
+  }
+
   retval = cvmfs_exports_->fnMaintenanceMode(fd_progress);
   if (!retval)
     return kFailMaintenanceMode;
@@ -597,57 +615,6 @@ Failures Reload(const int fd_progress, const bool stop_and_go) {
 
 using namespace loader;  // NOLINT(build/namespaces)
 
-// Making OpenSSL (libcrypto) thread-safe
-#ifndef OPENSSL_API_INTERFACE_V11
-
-pthread_mutex_t *gLibcryptoLocks;
-
-static void CallbackLibcryptoLock(int mode, int type,
-                                  const char *file, int line) {
-  (void)file;
-  (void)line;
-
-  int retval;
-
-  if (mode & CRYPTO_LOCK) {
-    retval = pthread_mutex_lock(&(gLibcryptoLocks[type]));
-  } else {
-    retval = pthread_mutex_unlock(&(gLibcryptoLocks[type]));
-  }
-  assert(retval == 0);
-}
-
-static unsigned long CallbackLibcryptoThreadId() {  // NOLINT(runtime/int)
-  return platform_gettid();
-}
-
-#endif
-
-static void SetupLibcryptoMt() {
-#ifndef OPENSSL_API_INTERFACE_V11
-  gLibcryptoLocks = static_cast<pthread_mutex_t *>(OPENSSL_malloc(
-    CRYPTO_num_locks() * sizeof(pthread_mutex_t)));
-  for (int i = 0; i < CRYPTO_num_locks(); ++i) {
-    int retval = pthread_mutex_init(&(gLibcryptoLocks[i]), NULL);
-    assert(retval == 0);
-  }
-
-  CRYPTO_set_id_callback(CallbackLibcryptoThreadId);
-  CRYPTO_set_locking_callback(CallbackLibcryptoLock);
-#endif
-}
-
-static void CleanupLibcryptoMt(void) {
-#ifndef OPENSSL_API_INTERFACE_V11
-  CRYPTO_set_locking_callback(NULL);
-  for (int i = 0; i < CRYPTO_num_locks(); ++i)
-    pthread_mutex_destroy(&(gLibcryptoLocks[i]));
-
-  OPENSSL_free(gLibcryptoLocks);
-#endif
-}
-
-
 int FuseMain(int argc, char *argv[]) {
   // Set a decent umask for new files (no write access to group/everyone).
   // We want to allow group write access for the talk-socket.
@@ -667,7 +634,23 @@ int FuseMain(int argc, char *argv[]) {
       bool stop_and_go = false;
       if ((argc > 3) && (string(argv[3]) == "stop_and_go"))
         stop_and_go = true;
-      retval = loader_talk::MainReload(argv[2], stop_and_go);
+
+      // always last param of the cvmfs2 __RELOAD__ command
+      // check if debug mode is requested
+      // NOTE:
+      // debug mode is decided based on CVMFS_DEBUGLOG being set or not
+      // this means: reloading is now always based on CVMFS_DEBUGLOG, and
+      // reload ignores the current state
+      //
+      // if you mount with debug but do not set CVMFS_DEBUGLOG and reload,
+      // you will reload with
+      if (std::string(argv[argc - 1]) == std::string("--debug")) {
+        debug_mode_ = true;
+      } else {
+        debug_mode_ = false;
+      }
+      retval = loader_talk::MainReload(argv[2], stop_and_go, debug_mode_);
+
       if ((retval != 0) && (stop_and_go)) {
         CreateFile(string(argv[2]) + ".paused.crashed", 0600);
       }
@@ -718,8 +701,6 @@ int FuseMain(int argc, char *argv[]) {
     return cvmfs_exports_->fnAltProcessFlavor(argc, argv);
   }
 
-  SetupLibcryptoMt();
-
   // Option parsing
   struct fuse_args *mount_options;
   mount_options = ParseCmdLine(argc, argv);
@@ -770,20 +751,41 @@ int FuseMain(int argc, char *argv[]) {
   }
   if (suid_mode_) {
     if (getuid() != 0) {
-      LogCvmfs(kLogCvmfs, kLogStderr | kLogSyslogErr,
-               "must be root to mount with suid option");
-      abort();
+      PANIC(kLogStderr | kLogSyslogErr,
+            "must be root to mount with suid option");
     }
     fuse_opt_add_arg(mount_options, "-osuid");
     LogCvmfs(kLogCvmfs, kLogStdout, "CernVM-FS: running with suid support");
   }
+
+  if (options_manager->GetValue("CVMFS_CPU_AFFINITY", &parameter)) {
+#ifndef __APPLE__
+     cpu_set_t mask;
+     vector<string> cpus = SplitString(parameter, ',');
+     CPU_ZERO(&mask);
+     for (vector<string>::iterator i = cpus.begin(); i != cpus.end(); i++) {
+        CPU_SET(String2Uint64(Trim(*i)), &mask);
+     }
+     LogCvmfs(kLogCvmfs, kLogStdout,
+              "CernVM-FS: setting CPU Affinity to %s", parameter.c_str());
+     int err = sched_setaffinity(0, sizeof(mask), &mask);
+     if (err != 0) {
+        LogCvmfs(kLogCvmfs, kLogStdout | kLogSyslogErr,
+                 "Setting CPU Affinity failed with error %d", errno);
+     }
+#else
+     LogCvmfs(kLogCvmfs, kLogStdout | kLogSyslogErr,
+              "CPU affinity setting not supported on macOS");
+#endif
+  }
   loader_exports_ = new LoaderExports();
-  loader_exports_->loader_version = PACKAGE_VERSION;
+  loader_exports_->loader_version = CVMFS_VERSION;
   loader_exports_->boot_time = time(NULL);
   loader_exports_->program_name = argv[0];
   loader_exports_->foreground = foreground_;
   loader_exports_->repository_name = *repository_name_;
   loader_exports_->mount_point = *mount_point_;
+  loader_exports_->device_id = "0:0";  // initially unknown, set after mount
   loader_exports_->disable_watchdog = disable_watchdog_;
   loader_exports_->simple_options_parsing = simple_options_parsing_;
   if (config_files_)
@@ -816,7 +818,7 @@ int FuseMain(int argc, char *argv[]) {
 
   if (!premounted_ && !DirectoryExists(*mount_point_)) {
     LogCvmfs(kLogCvmfs, kLogStderr | kLogSyslogErr,
-             "Moint point %s does not exist", mount_point_->c_str());
+             "Mount point %s does not exist", mount_point_->c_str());
     return kFailPermission;
   }
 
@@ -890,6 +892,10 @@ int FuseMain(int argc, char *argv[]) {
       return kFailPermission;
     }
   }
+  if (disable_watchdog_) {
+    LogCvmfs(kLogCvmfs, kLogDebug, "No watchdog, enabling core files");
+    prctl(PR_SET_DUMPABLE, 1, 0, 0, 0);
+  }
 
   // Only set usyslog now, otherwise file permissions are wrong
   usyslog_path_ = new string();
@@ -929,6 +935,30 @@ int FuseMain(int argc, char *argv[]) {
     return kFailLoaderTalk;
   }
 
+  // TODO(jblomer): we probably want to apply a default setting related to the
+  // number of cores.
+  if (options_manager->GetValue("CVMFS_FUSE3_MAX_THREADS", &parameter)) {
+    fuse3_max_threads_ = String2Int64(parameter);
+  }
+  if (options_manager->GetValue("CVMFS_FUSE3_IDLE_THREADS", &parameter)) {
+    fuse3_idle_threads_ = String2Int64(parameter);
+  }
+#ifdef CVMFS_ENABLE_FUSE3_LOOP_CONFIG
+  if (fuse3_max_threads_) {
+    LogCvmfs(kLogCvmfs, kLogStdout,
+             "CernVM-FS: Fuse3 max_threads=%d", fuse3_max_threads_);
+  }
+  if (fuse3_idle_threads_) {
+    LogCvmfs(kLogCvmfs, kLogStdout,
+             "CernVM-FS: Fuse3 min_idle_threads=%d", fuse3_idle_threads_);
+  }
+#else
+  if (fuse3_max_threads_ || fuse3_idle_threads_) {
+    LogCvmfs(kLogCvmfs, kLogStdout,
+             "CernVM-FS: ignoring fuse3 thread settings (libfuse too old)");
+  }
+#endif
+
   // Options are not needed anymore
   delete options_manager;
   options_manager = NULL;
@@ -983,6 +1013,10 @@ int FuseMain(int argc, char *argv[]) {
 
   struct fuse_lowlevel_ops loader_operations;
   SetFuseOperations(&loader_operations);
+#if (FUSE_VERSION >= 29)
+  if (cvmfs_exports_->cvmfs_operations.forget_multi)
+    loader_operations.forget_multi = stub_forget_multi;
+#endif
 
 #if CVMFS_USE_LIBFUSE == 2
   channel = fuse_mount(mount_point_->c_str(), mount_options);
@@ -1032,6 +1066,26 @@ int FuseMain(int argc, char *argv[]) {
     }
   }
 
+  // Determine device id
+  int fd_mountinfo = open("/proc/self/mountinfo", O_RDONLY);
+  if (fd_mountinfo > 0) {
+    std::string line;
+    while (GetLineFd(fd_mountinfo, &line)) {
+      std::vector<std::string> tokens = SplitString(line, ' ');
+      if (tokens.size() < 5) continue;
+      if (tokens[4] != loader_exports_->mount_point) continue;
+      unsigned i = 5;
+      for (; i < tokens.size(); ++i) {
+        if (tokens[i] == "-") break;
+      }
+      if (tokens.size() < i + 3) continue;
+      if (tokens[i + 2] != "cvmfs2") continue;
+      loader_exports_->device_id = tokens[2];
+      break;
+    }
+    close(fd_mountinfo);
+  }
+
   if (!premounted_) {
     LogCvmfs(kLogCvmfs, kLogStdout, "CernVM-FS: mounted cvmfs on %s",
              mount_point_->c_str());
@@ -1056,9 +1110,25 @@ int FuseMain(int argc, char *argv[]) {
   } else {
 #if CVMFS_USE_LIBFUSE == 2
     retval = fuse_session_loop_mt(session);
+#else
+#ifdef CVMFS_ENABLE_FUSE3_LOOP_CONFIG
+    struct fuse_loop_config *fuse_loop_cfg = fuse_loop_cfg_create();
+
+    fuse_loop_cfg_set_clone_fd(fuse_loop_cfg, 1);
+
+    if (fuse3_max_threads_ > 0) {
+      fuse_loop_cfg_set_max_threads(fuse_loop_cfg, fuse3_max_threads_);
+    }
+    if (fuse3_idle_threads_ > 0) {
+      fuse_loop_cfg_set_idle_threads(fuse_loop_cfg, fuse3_idle_threads_);
+    }
+
+    retval = fuse_session_loop_mt(session, fuse_loop_cfg);
+    fuse_loop_cfg_destroy(fuse_loop_cfg);
 #else
     retval = fuse_session_loop_mt(session, 1 /* use fd per thread */);
-#endif
+#endif  // CVMFS_ENABLE_FUSE3_LOOP_CONFIG
+#endif  // fuse2/3
   }
   SetLogMicroSyslog(*usyslog_path_);
 
@@ -1088,8 +1158,6 @@ int FuseMain(int argc, char *argv[]) {
   LogCvmfs(kLogCvmfs, kLogSyslog, "CernVM-FS: unmounted %s (%s)",
            mount_point_->c_str(), repository_name_->c_str());
 
-  CleanupLibcryptoMt();
-
   delete fence_reload_;
   delete loader_exports_;
   delete config_files_;
@@ -1121,4 +1189,3 @@ static void __attribute__((destructor)) LibraryExit() {
   delete g_cvmfs_stub_exports;
   g_cvmfs_stub_exports = NULL;
 }
-
