@@ -8,7 +8,7 @@
  * Also, it handles getting and setting the maximum number of file descriptors.
  */
 
-#include "cvmfs_config.h"
+
 #include "monitor.h"
 
 #include <errno.h>
@@ -25,6 +25,7 @@
 #endif
 #include <sys/uio.h>
 #include <sys/wait.h>
+#include <syslog.h>
 #include <time.h>
 #include <unistd.h>
 
@@ -33,20 +34,22 @@
 #include <cstdlib>
 #include <cstring>
 #include <map>
+#include <set>
 #include <string>
 #include <vector>
 
-#ifndef CVMFS_LIBCVMFS
+#if defined(CVMFS_FUSE_MODULE)
 #include "cvmfs.h"
 #endif
-#include "logging.h"
-#include "platform.h"
-#include "smalloc.h"
+#include "util/exception.h"
+#include "util/logging.h"
+#include "util/platform.h"
 #include "util/posix.h"
+#include "util/smalloc.h"
 #include "util/string.h"
 
 // Used for address offset calculation
-#ifndef CVMFS_LIBCVMFS
+#if defined(CVMFS_FUSE_MODULE)
 extern loader::CvmfsExports *g_cvmfs_exports;
 #endif
 
@@ -54,10 +57,20 @@ using namespace std;  // NOLINT
 
 Watchdog *Watchdog::instance_ = NULL;
 
+int Watchdog::g_suppressed_signals[] = { SIGHUP, SIGINT, SIGQUIT,
+                                SIGILL, SIGABRT, SIGBUS,
+                                SIGFPE, SIGUSR1, SIGSEGV,
+                                SIGUSR2, SIGTERM, SIGXCPU,
+                                SIGXFSZ};
 
-Watchdog *Watchdog::Create(const string &crash_dump_path) {
+int Watchdog::g_crash_signals[] = { SIGQUIT, SIGILL, SIGABRT,
+                                  SIGFPE, SIGSEGV, SIGBUS,
+                                  SIGPIPE, SIGXFSZ };
+
+Watchdog *Watchdog::Create(FnOnCrash on_crash) {
   assert(instance_ == NULL);
-  instance_ = new Watchdog(crash_dump_path);
+  instance_ = new Watchdog(on_crash);
+  instance_->Fork();
   return instance_;
 }
 
@@ -66,7 +79,7 @@ Watchdog *Watchdog::Create(const string &crash_dump_path) {
  * Uses an external shell and gdb to create a full stack trace of the dying
  * process. The same shell is used to force-quit the client afterwards.
  */
-string Watchdog::GenerateStackTrace(const string &exe_path, pid_t pid) {
+string Watchdog::GenerateStackTrace(pid_t pid) {
   int retval;
   string result = "";
 
@@ -81,13 +94,7 @@ string Watchdog::GenerateStackTrace(const string &exe_path, pid_t pid) {
   int fd_stdout;
   int fd_stderr;
   vector<string> argv;
-#ifndef __APPLE__
-  argv.push_back("-q");
-  argv.push_back("-n");
-  argv.push_back(exe_path);
-#else
   argv.push_back("-p");
-#endif
   argv.push_back(StringifyInt(pid));
   pid_t gdb_pid = 0;
   const bool double_fork = false;
@@ -130,6 +137,15 @@ string Watchdog::GenerateStackTrace(const string &exe_path, pid_t pid) {
 #endif
   result += ReadUntilGdbPrompt(fd_stdout) + "\n\n";
 
+  // Check for output on stderr
+  string result_err;
+  Block2Nonblock(fd_stderr);
+  char cbuf;
+  while (read(fd_stderr, &cbuf, 1) == 1)
+    result_err.push_back(cbuf);
+  if (!result_err.empty())
+    result += "\nError output:\n" + result_err + "\n";
+
   // Close the connection to the terminated gdb process
   close(fd_stderr);
   close(fd_stdout);
@@ -155,15 +171,11 @@ string Watchdog::GenerateStackTrace(const string &exe_path, pid_t pid) {
 
 pid_t Watchdog::GetPid() {
   if (instance_ != NULL) {
-    if (!instance_->spawned_)
-      return getpid();
-    else
-      return instance_->watchdog_pid_;
+    return instance_->watchdog_pid_;
   }
   return getpid();
 }
 
-
 /**
  * Log a string to syslog and into the crash dump file.
  * We expect ideally nothing to be logged, so that file is created on demand.
@@ -176,9 +188,12 @@ void Watchdog::LogEmergency(string msg) {
     if (fp) {
       time_t now = time(NULL);
       msg += "\nTimestamp: " + string(ctime_r(&now, ctime_buffer));
-      if (fwrite(&msg[0], 1, msg.length(), fp) != msg.length())
-        msg += " (failed to report into crash dump file "
-               + crash_dump_path_ + ")";
+      if (fwrite(&msg[0], 1, msg.length(), fp) != msg.length()) {
+        msg +=
+            " (failed to report into crash dump file " + crash_dump_path_ + ")";
+      } else {
+        msg += "\n Crash logged also on file: " + crash_dump_path_ + "\n";
+      }
       fclose(fp);
     } else {
       msg += " (failed to open crash dump file " + crash_dump_path_ + ")";
@@ -187,7 +202,6 @@ void Watchdog::LogEmergency(string msg) {
   LogCvmfs(kLogMonitor, kLogSyslogErr, "%s", msg.c_str());
 }
 
-
 /**
  * Reads from the file descriptor until the specific gdb prompt is reached or
  * the pipe gets broken.
@@ -207,7 +221,7 @@ string Watchdog::ReadUntilGdbPrompt(int fd_pipe) {
   int           chars_io;
   unsigned int  ring_buffer_pos = 0;
 
-  // read from stdout of gdb until gdb prompt occures --> (gdb)
+  // read from stdout of gdb until gdb prompt occurs --> (gdb)
   while (1) {
     chars_io = read(fd_pipe, &mini_buffer, 1);
 
@@ -231,31 +245,23 @@ string Watchdog::ReadUntilGdbPrompt(int fd_pipe) {
 }
 
 
-void Watchdog::RegisterOnCrash(void (*CleanupOnCrash)(void)) {
-  on_crash_ = CleanupOnCrash;
-}
-
-
 /**
  * Generates useful information from the backtrace log in the pipe.
  */
 string Watchdog::ReportStacktrace() {
-  // Re-activate µSyslog, if necessary
-  SetLogMicroSyslog(GetLogMicroSyslog());
-
   CrashData crash_data;
-  if (!pipe_watchdog_->Read(&crash_data)) {
+  if (!pipe_watchdog_->TryRead<CrashData>(&crash_data)) {
     return "failed to read crash data (" + StringifyInt(errno) + ")";
   }
 
   string debug = "--\n";
   debug += "Signal: "    + StringifyInt(crash_data.signal);
   debug += ", errno: "   + StringifyInt(crash_data.sys_errno);
-  debug += ", version: " + string(VERSION);
+  debug += ", version: " + string(CVMFS_VERSION);
   debug += ", PID: "     + StringifyInt(crash_data.pid) + "\n";
   debug += "Executable path: " + exe_path_ + "\n";
 
-  debug += GenerateStackTrace(exe_path_, crash_data.pid);
+  debug += GenerateStackTrace(crash_data.pid);
 
   // Give the dying process the finishing stroke
   if (kill(crash_data.pid, SIGKILL) != 0) {
@@ -280,6 +286,16 @@ string Watchdog::ReportStacktrace() {
 }
 
 
+void Watchdog::ReportSignalAndTerminate(
+  int sig, siginfo_t *siginfo, void * /* context */)
+{
+  LogCvmfs(kLogMonitor, kLogSyslogErr,
+           "watchdog: received unexpected signal %d from PID %d / UID %d",
+           sig, siginfo->si_pid, siginfo->si_uid);
+  _exit(1);
+}
+
+
 void Watchdog::SendTrace(int sig, siginfo_t *siginfo, void *context) {
   int send_errno = errno;
   if (platform_spinlock_trylock(&Me()->lock_handler_) != 0) {
@@ -301,7 +317,7 @@ void Watchdog::SendTrace(int sig, siginfo_t *siginfo, void *context) {
   crash_data.signal     = sig;
   crash_data.sys_errno  = send_errno;
   crash_data.pid        = getpid();
-  if (!Me()->pipe_watchdog_->Write(crash_data)) {
+  if (!Me()->pipe_watchdog_->Write<CrashData>(crash_data)) {
     _exit(1);
   }
 
@@ -314,7 +330,7 @@ void Watchdog::SendTrace(int sig, siginfo_t *siginfo, void *context) {
     if (++counter == 300) {
       LogCvmfs(kLogCvmfs, kLogSyslogErr, "stack trace generation failed");
       // Last attempt to log something useful
-#ifndef CVMFS_LIBCVMFS
+#if defined(CVMFS_FUSE_MODULE)
       LogCvmfs(kLogCvmfs, kLogSyslogErr, "Signal %d, errno %d",
                sig, send_errno);
       void *addr[kMaxBacktrace];
@@ -356,7 +372,7 @@ Watchdog::SigactionMap Watchdog::SetSignalHandlers(
   for (; i != iend; ++i) {
     struct sigaction old_signal_handler;
     if (sigaction(i->first, &i->second, &old_signal_handler) != 0) {
-      abort();
+      PANIC(NULL);
     }
     old_signal_handlers[i->first] = old_signal_handler;
   }
@@ -366,63 +382,132 @@ Watchdog::SigactionMap Watchdog::SetSignalHandlers(
 
 
 /**
- * Fork the watchdog process.
+ * Fork the watchdog process and put it on hold until Spawn() is called.
  */
-void Watchdog::Spawn() {
-  Pipe pipe_pid;
-  pipe_watchdog_ = new Pipe();
-  pipe_listener_ = new Pipe();
+void Watchdog::Fork() {
+  Pipe<kPipeWatchdogPid> pipe_pid;
+  pipe_watchdog_ = new Pipe<kPipeWatchdog>();
+  pipe_listener_ = new Pipe<kPipeWatchdogSupervisor>();
 
   pid_t pid;
   int statloc;
-  int max_fd = sysconf(_SC_OPEN_MAX);
-  assert(max_fd >= 0);
   switch (pid = fork()) {
-    case -1: abort();
+    case -1: PANIC(NULL);
     case 0:
       // Double fork to avoid zombie
       switch (fork()) {
-        case -1: exit(1);
+        case -1: _exit(1);
         case 0: {
-          close(pipe_watchdog_->write_end);
+          pipe_watchdog_->CloseWriteFd();
           Daemonize();
-          // send the watchdog PID to cvmfs
+          // send the watchdog PID to the supervisee
           pid_t watchdog_pid = getpid();
           pipe_pid.Write(watchdog_pid);
-          close(pipe_pid.write_end);
+          pipe_pid.CloseWriteFd();
           // Close all unused file descriptors
           // close also usyslog, only get it back if necessary
           // string usyslog_save = GetLogMicroSyslog();
           string debuglog_save = GetLogDebugFile();
-          // SetLogMicroSyslog("");
           SetLogDebugFile("");
-          for (int fd = 0; fd < max_fd; fd++) {
-            if (fd == pipe_watchdog_->read_end)
-              continue;
-            if (fd == pipe_listener_->write_end)
-              continue;
-            close(fd);
-          }
-          // SetLogMicroSyslog(usyslog_save);  // no-op if usyslog not used
+          string usyslog_save = GetLogMicroSyslog();
+          SetLogMicroSyslog("");
+          // Gracefully close the syslog before closing all fds. The next call
+          // to syslog will reopen it.
+          closelog();
+          // Let's keep stdin, stdout, stderr open at /dev/null (daemonized)
+          // in order to prevent accidental outputs from messing with another
+          // file descriptor
+          std::set<int> preserve_fds;
+          preserve_fds.insert(0);
+          preserve_fds.insert(1);
+          preserve_fds.insert(2);
+          preserve_fds.insert(pipe_watchdog_->GetReadFd());
+          preserve_fds.insert(pipe_listener_->GetWriteFd());
+          CloseAllFildes(preserve_fds);
+          SetLogMicroSyslog(usyslog_save);  // no-op if usyslog not used
           SetLogDebugFile(debuglog_save);  // no-op if debug log not used
-          Supervise();
+
+          if (WaitForSupervisee())
+            Supervise();
+
+          pipe_watchdog_->CloseReadFd();
+          pipe_listener_->CloseWriteFd();
           exit(0);
         }
         default:
-          exit(0);
+          _exit(0);
       }
     default:
-      close(pipe_watchdog_->read_end);
-      close(pipe_listener_->write_end);
-      if (waitpid(pid, &statloc, 0) != pid) abort();
-      if (!WIFEXITED(statloc) || WEXITSTATUS(statloc)) abort();
+      pipe_watchdog_->CloseReadFd();
+      pipe_listener_->CloseWriteFd();
+      pipe_pid.CloseWriteFd();
+      if (waitpid(pid, &statloc, 0) != pid) PANIC(NULL);
+      if (!WIFEXITED(statloc) || WEXITSTATUS(statloc)) PANIC(NULL);
   }
 
   // retrieve the watchdog PID from the pipe
-  close(pipe_pid.write_end);
   pipe_pid.Read(&watchdog_pid_);
-  close(pipe_pid.read_end);
+  pipe_pid.CloseReadFd();
+}
 
+
+bool Watchdog::WaitForSupervisee() {
+  // We want broken pipes not to raise a signal but handle the error in the
+  // read/write code
+  platform_sighandler_t rv_sig = signal(SIGPIPE, SIG_IGN);
+  assert(rv_sig != SIG_ERR);
+
+  // The watchdog is not supposed to receive signals. If it does, report it.
+  struct sigaction sa;
+  memset(&sa, 0, sizeof(sa));
+  sa.sa_sigaction = ReportSignalAndTerminate;
+  sa.sa_flags = SA_SIGINFO;
+  sigfillset(&sa.sa_mask);
+
+  SigactionMap signal_handlers;
+  for (size_t i = 0; i < sizeof(g_suppressed_signals)/sizeof(int); i++) {
+    signal_handlers[g_suppressed_signals[i]] = sa;
+  }
+  SetSignalHandlers(signal_handlers);
+
+  ControlFlow::Flags control_flow = ControlFlow::kUnknown;
+
+  if (!pipe_watchdog_->TryRead(&control_flow)) {
+    LogCvmfs(kLogMonitor, kLogDebug, "supervisee canceled watchdog");
+    return false;
+  }
+
+  switch (control_flow) {
+    case ControlFlow::kQuit:
+      return false;
+    case ControlFlow::kSupervise:
+      break;
+    default:
+      LogEmergency("Internal error: invalid control flow");
+      return false;
+  }
+
+  size_t size;
+  pipe_watchdog_->Read(&size);
+  crash_dump_path_.resize(size);
+  if (size > 0) {
+    pipe_watchdog_->Read(&crash_dump_path_[0], size);
+
+    int retval = chdir(GetParentPath(crash_dump_path_).c_str());
+    if (retval != 0) {
+      LogEmergency(std::string("Cannot change to crash dump directory: ") +
+                   crash_dump_path_);
+      return false;
+    }
+    crash_dump_path_ = GetFileName(crash_dump_path_);
+  }
+  return true;
+}
+
+/**
+ * Set up the signal handling and kick off the supervision.
+ */
+void Watchdog::Spawn(const std::string &crash_dump_path) {
   // lower restrictions for ptrace
   if (!platform_allow_ptrace(watchdog_pid_)) {
     LogCvmfs(kLogMonitor, kLogSyslogWarn,
@@ -437,7 +522,7 @@ void Watchdog::Spawn() {
   sighandler_stack_.ss_size = stack_size;
   sighandler_stack_.ss_flags = 0;
   if (sigaltstack(&sighandler_stack_, NULL) != 0)
-    abort();
+    PANIC(NULL);
 
   // define our crash signal handler
   struct sigaction sa;
@@ -447,21 +532,23 @@ void Watchdog::Spawn() {
   sigfillset(&sa.sa_mask);
 
   SigactionMap signal_handlers;
-  signal_handlers[SIGQUIT] = sa;
-  signal_handlers[SIGILL]  = sa;
-  signal_handlers[SIGABRT] = sa;
-  signal_handlers[SIGFPE]  = sa;
-  signal_handlers[SIGSEGV] = sa;
-  signal_handlers[SIGBUS]  = sa;
-  signal_handlers[SIGPIPE] = sa;
-  signal_handlers[SIGXFSZ] = sa;
+  for (size_t i = 0; i < sizeof(g_crash_signals)/sizeof(int); i++) {
+    signal_handlers[g_crash_signals[i]] = sa;
+  }
   old_signal_handlers_ = SetSignalHandlers(signal_handlers);
 
-  pipe_terminate_ = new Pipe();
+  pipe_terminate_ = new Pipe<kPipeThreadTerminator>();
   int retval =
     pthread_create(&thread_listener_, NULL, MainWatchdogListener, this);
   assert(retval == 0);
 
+  pipe_watchdog_->Write(ControlFlow::kSupervise);
+  size_t path_size = crash_dump_path.size();
+  pipe_watchdog_->Write(path_size);
+  if (path_size > 0) {
+    pipe_watchdog_->Write(crash_dump_path.data(), path_size);
+  }
+
   spawned_ = true;
 }
 
@@ -471,10 +558,10 @@ void *Watchdog::MainWatchdogListener(void *data) {
   LogCvmfs(kLogMonitor, kLogDebug, "starting watchdog listener");
 
   struct pollfd watch_fds[2];
-  watch_fds[0].fd = watchdog->pipe_listener_->read_end;
-  watch_fds[0].events = POLLIN | POLLPRI;
+  watch_fds[0].fd = watchdog->pipe_listener_->GetReadFd();
+  watch_fds[0].events = 0;  // Only check for POLL[ERR,HUP,NVAL] in revents
   watch_fds[0].revents = 0;
-  watch_fds[1].fd = watchdog->pipe_terminate_->read_end;
+  watch_fds[1].fd = watchdog->pipe_terminate_->GetReadFd();
   watch_fds[1].events = POLLIN | POLLPRI;
   watch_fds[1].revents = 0;
   while (true) {
@@ -493,13 +580,15 @@ void *Watchdog::MainWatchdogListener(void *data) {
           (watch_fds[0].revents & POLLNVAL))
       {
         LogCvmfs(kLogMonitor, kLogDebug | kLogSyslogErr,
-                 "watchdog disappeared, aborting");
-        abort();
+                 "watchdog disappeared, disabling stack trace reporting "
+                 "(revents: %d / %d|%d|%d)",
+                 watch_fds[0].revents, POLLERR, POLLHUP, POLLNVAL);
+        watchdog->SetSignalHandlers(watchdog->old_signal_handlers_);
+        PANIC(kLogDebug | kLogSyslogErr, "watchdog disappeared, aborting");
       }
-      assert(false);
+      PANIC(NULL);
     }
   }
-  close(watchdog->pipe_listener_->read_end);
 
   LogCvmfs(kLogMonitor, kLogDebug, "stopping watchdog listener");
   return NULL;
@@ -507,13 +596,11 @@ void *Watchdog::MainWatchdogListener(void *data) {
 
 
 void Watchdog::Supervise() {
-  signal(SIGPIPE, SIG_IGN);
-  ControlFlow::Flags control_flow;
+  ControlFlow::Flags control_flow = ControlFlow::kUnknown;
 
-  if (!pipe_watchdog_->Read(&control_flow)) {
-    // Re-activate µSyslog, if necessary
-    SetLogMicroSyslog(GetLogMicroSyslog());
-    LogEmergency("unexpected termination (" + StringifyInt(control_flow) + ")");
+  if (!pipe_watchdog_->TryRead<ControlFlow::Flags>(&control_flow)) {
+    LogEmergency("watchdog: unexpected termination (" +
+                 StringifyInt(control_flow) + ")");
     if (on_crash_) on_crash_();
   } else {
     switch (control_flow) {
@@ -526,27 +613,18 @@ void Watchdog::Supervise() {
         break;
 
       default:
-        // Re-activate µSyslog, if necessary
-        SetLogMicroSyslog(GetLogMicroSyslog());
-        LogEmergency("unexpected error");
+        LogEmergency("watchdog: unexpected error");
         break;
     }
   }
-
-  close(pipe_watchdog_->read_end);
-  close(pipe_listener_->write_end);
 }
 
 
-Watchdog::Watchdog(const string &crash_dump_path)
+Watchdog::Watchdog(FnOnCrash on_crash)
   : spawned_(false)
-  , crash_dump_path_(crash_dump_path)
   , exe_path_(string(platform_getexepath()))
   , watchdog_pid_(0)
-  , pipe_watchdog_(NULL)
-  , pipe_listener_(NULL)
-  , pipe_terminate_(NULL)
-  , on_crash_(NULL)
+  , on_crash_(on_crash)
 {
   int retval = platform_spinlock_init(&lock_handler_, 0);
   assert(retval == 0);
@@ -571,49 +649,13 @@ Watchdog::~Watchdog() {
     pipe_terminate_->Write(ControlFlow::kQuit);
     pthread_join(thread_listener_, NULL);
     pipe_terminate_->Close();
-
-    pipe_watchdog_->Write(ControlFlow::kQuit);
-    close(pipe_watchdog_->write_end);
   }
 
-  delete pipe_watchdog_;
-  delete pipe_listener_;
-  delete pipe_terminate_;
+  pipe_watchdog_->Write(ControlFlow::kQuit);
+  pipe_watchdog_->CloseWriteFd();
+  pipe_listener_->CloseReadFd();
 
   platform_spinlock_destroy(&lock_handler_);
   LogCvmfs(kLogMonitor, kLogDebug, "monitor stopped");
   instance_ = NULL;
 }
-
-
-
-namespace monitor {
-
-const unsigned kMinOpenFiles = 8192;
-
-unsigned GetMaxOpenFiles() {
-  static unsigned max_open_files;
-  static bool     already_done = false;
-
-  /* check number of open files (lazy evaluation) */
-  if (!already_done) {
-    unsigned soft_limit = 0;
-    unsigned hard_limit = 0;
-    GetLimitNoFile(&soft_limit, &hard_limit);
-
-    if (soft_limit < kMinOpenFiles) {
-      LogCvmfs(kLogMonitor, kLogSyslogWarn | kLogDebug,
-               "Warning: current limits for number of open files are "
-               "(%lu/%lu)\n"
-               "CernVM-FS is likely to run out of file descriptors, "
-               "set ulimit -n to at least %lu",
-               soft_limit, hard_limit, kMinOpenFiles);
-    }
-    max_open_files = soft_limit;
-    already_done   = true;
-  }
-
-  return max_open_files;
-}
-
-}  // namespace monitor
