@@ -1,7 +1,7 @@
 /**
  * This file is part of the CernVM File System.
  */
-#include "cvmfs_config.h"
+
 #include "mountpoint.h"
 
 #include <errno.h>
@@ -34,11 +34,12 @@
 #include "cache_extern.h"
 #include "cache_posix.h"
 #include "cache_ram.h"
+#include "cache_stream.h"
 #include "cache_tiered.h"
 #include "catalog.h"
 #include "catalog_mgr_client.h"
 #include "clientctx.h"
-#include "download.h"
+#include "crypto/signature.h"
 #include "duplex_sqlite3.h"
 #include "fetch.h"
 #include "file_chunk.h"
@@ -47,29 +48,30 @@
 #include "google/protobuf/stubs/common.h"
 #include "history.h"
 #include "history_sqlite.h"
-#include "logging.h"
 #include "lru_md.h"
 #include "manifest.h"
 #include "manifest_fetch.h"
+#include "network/download.h"
 #include "nfs_maps.h"
 #ifdef CVMFS_NFS_SUPPORT
 #include "nfs_maps_leveldb.h"
 #include "nfs_maps_sqlite.h"
 #endif
 #include "options.h"
-#include "platform.h"
 #include "quota_posix.h"
 #include "resolv_conf_event_handler.h"
-#include "signature.h"
 #include "sqlitemem.h"
 #include "sqlitevfs.h"
 #include "statistics.h"
+#include "telemetry_aggregator.h"
 #include "tracer.h"
+#include "util/concurrency.h"
+#include "util/logging.h"
+#include "util/platform.h"
 #include "util/pointer.h"
 #include "util/posix.h"
 #include "util/string.h"
-#include "util_concurrency.h"
-#include "uuid.h"
+#include "util/uuid.h"
 #include "wpad.h"
 
 using namespace std;  // NOLINT
@@ -80,6 +82,26 @@ const char *FileSystem::kDefaultCacheBase = "/var/lib/cvmfs";
 const char *FileSystem::kDefaultCacheMgrInstance = "default";
 
 
+FileSystem::IoErrorInfo::IoErrorInfo() : counter_(NULL), timestamp_last_(0) { }
+
+void FileSystem::IoErrorInfo::Reset() {
+  counter_->Set(0);
+  timestamp_last_ = 0;
+}
+
+void FileSystem::IoErrorInfo::AddIoError() {
+  perf::Inc(counter_);
+  timestamp_last_ = time(NULL);
+}
+
+void FileSystem::IoErrorInfo::SetCounter(perf::Counter *c) { counter_ = c; }
+
+int64_t FileSystem::IoErrorInfo::count() { return counter_->Get(); }
+
+time_t FileSystem::IoErrorInfo::timestamp_last() { return timestamp_last_; }
+
+
+
 /**
  * A cache instance name is part of a bash parameter and can only contain
  * certain characters.
@@ -145,6 +167,8 @@ FileSystem *FileSystem::Create(const FileSystem::FileSystemInfo &fs_info) {
   UniquePtr<FileSystem>
     file_system(new FileSystem(fs_info));
 
+  file_system->SetupGlobalEnvironmentParams();
+
   file_system->SetupLogging();
   LogCvmfs(kLogCvmfs, kLogDebug, "Options:\n%s",
            file_system->options_mgr()->Dump().c_str());
@@ -204,17 +228,45 @@ void FileSystem::CreateStatistics() {
   n_fs_lookup_negative_ = statistics_->Register("cvmfs.n_fs_lookup_negative",
                                                 "Number of negative lookups");
   n_fs_stat_ = statistics_->Register("cvmfs.n_fs_stat", "Number of stats");
+  n_fs_stat_stale_ = statistics_->Register("cvmfs.n_fs_stat_stale",
+    "Number of stats for stale (open, meanwhile changed) regular files");
+  n_fs_statfs_ = statistics_->Register("cvmfs.n_fs_statfs",
+                                       "Overall number of statsfs calls");
+  n_fs_statfs_cached_ = statistics_->Register("cvmfs.n_fs_statfs_cached",
+                "Number of statsfs calls that accessed the cached statfs info");
   n_fs_read_ = statistics_->Register("cvmfs.n_fs_read", "Number of files read");
   n_fs_readlink_ = statistics_->Register("cvmfs.n_fs_readlink",
                                          "Number of links read");
   n_fs_forget_ = statistics_->Register("cvmfs.n_fs_forget",
                                        "Number of inode forgets");
-  n_io_error_ = statistics_->Register("cvmfs.n_io_error",
-                                      "Number of I/O errors");
+  n_fs_inode_replace_ = statistics_->Register("cvmfs.n_fs_inode_replace",
+    "Number of stale inodes that got replaced by an up-to-date version");
   no_open_files_ = statistics_->Register("cvmfs.no_open_files",
                                          "Number of currently opened files");
   no_open_dirs_ = statistics_->Register("cvmfs.no_open_dirs",
                   "Number of currently opened directories");
+  io_error_info_.SetCounter(statistics_->Register("cvmfs.n_io_error",
+                                                  "Number of I/O errors"));
+  n_eio_total_ =  statistics_->Register("eio.total",
+     "EIO returned to calling process. Sum of individual eio counters");
+  n_eio_01_ =  statistics_->Register("eio.01",
+     "EIO returned to calling process. cvmfs.cc:cvmfs_lookup()");
+  n_eio_02_ =  statistics_->Register("eio.02",
+     "EIO returned to calling process. cvmfs.cc:ReplyNegative()");
+  n_eio_03_ =  statistics_->Register("eio.03",
+     "EIO returned to calling process. cvmfs.cc:cvmfs_opendir()");
+  n_eio_04_ =  statistics_->Register("eio.04",
+     "EIO returned to calling process. cvmfs.cc:cvmfs_open()");
+  n_eio_05_ =  statistics_->Register("eio.05",
+     "EIO returned to calling process. cvmfs.cc:cvmfs_read()");
+  n_eio_06_ =  statistics_->Register("eio.06",
+     "EIO returned to calling process. cvmfs.cc:cvmfs_open()");
+  n_eio_07_ =  statistics_->Register("eio.07",
+     "EIO returned to calling process. cvmfs.cc:cvmfs_read()");
+  n_eio_08_ =  statistics_->Register("eio.08",
+     "EIO returned to calling process. cvmfs.cc:cvmfs_read()");
+  n_emfile_ =  statistics_->Register("eio.emfile",
+     "EMFILE returned to calling process. cvmfs.cc:cvmfs_read()");
 
   string optarg;
   if (options_mgr_->GetValue("CVMFS_INSTRUMENT_FUSE", &optarg) &&
@@ -225,6 +277,7 @@ void FileSystem::CreateStatistics() {
 
   hist_fs_lookup_ = new Log2Histogram(30);
   hist_fs_forget_ = new Log2Histogram(30);
+  hist_fs_forget_multi_ = new Log2Histogram(30);
   hist_fs_getattr_ = new Log2Histogram(30);
   hist_fs_readlink_ = new Log2Histogram(30);
   hist_fs_opendir_ = new Log2Histogram(30);
@@ -246,6 +299,13 @@ FileSystem::PosixCacheSettings FileSystem::DeterminePosixCacheSettings(
   string optarg;
   PosixCacheSettings settings;
 
+  if (options_mgr_->GetValue(MkCacheParm("CVMFS_CACHE_REFCOUNT", instance),
+                             &optarg)
+      && options_mgr_->IsOff(optarg))
+  {
+    settings.do_refcount = false;
+  }
+
   if (options_mgr_->GetValue(MkCacheParm("CVMFS_CACHE_SHARED", instance),
                              &optarg)
       && options_mgr_->IsOn(optarg))
@@ -351,12 +411,25 @@ FileSystem::FileSystem(const FileSystem::FileSystemInfo &fs_info)
   , n_fs_lookup_(NULL)
   , n_fs_lookup_negative_(NULL)
   , n_fs_stat_(NULL)
+  , n_fs_stat_stale_(NULL)
+  , n_fs_statfs_(NULL)
+  , n_fs_statfs_cached_(NULL)
   , n_fs_read_(NULL)
   , n_fs_readlink_(NULL)
   , n_fs_forget_(NULL)
-  , n_io_error_(NULL)
+  , n_fs_inode_replace_(NULL)
   , no_open_files_(NULL)
   , no_open_dirs_(NULL)
+  , n_eio_total_(NULL)
+  , n_eio_01_(NULL)
+  , n_eio_02_(NULL)
+  , n_eio_03_(NULL)
+  , n_eio_04_(NULL)
+  , n_eio_05_(NULL)
+  , n_eio_06_(NULL)
+  , n_eio_07_(NULL)
+  , n_eio_08_(NULL)
+  , n_emfile_(NULL)
   , statistics_(NULL)
   , fd_workspace_lock_(-1)
   , found_previous_crash_(false)
@@ -408,6 +481,7 @@ FileSystem::~FileSystem() {
   SqliteMemoryManager::CleanupInstance();
 
   delete hist_fs_lookup_;
+  delete hist_fs_forget_multi_;
   delete hist_fs_forget_;
   delete hist_fs_getattr_;
   delete hist_fs_readlink_;
@@ -529,7 +603,17 @@ string FileSystem::MkCacheParm(
 
 
 void FileSystem::ResetErrorCounters() {
-  n_io_error_->Set(0);
+  io_error_info_.Reset();
+  n_eio_total_->Set(0);
+  n_eio_01_->Set(0);
+  n_eio_02_->Set(0);
+  n_eio_03_->Set(0);
+  n_eio_04_->Set(0);
+  n_eio_05_->Set(0);
+  n_eio_06_->Set(0);
+  n_eio_07_->Set(0);
+  n_eio_08_->Set(0);
+  n_emfile_->Set(0);
 }
 
 
@@ -613,12 +697,12 @@ CacheManager *FileSystem::SetupPosixCacheMgr(const string &instance) {
   PosixCacheSettings settings = DeterminePosixCacheSettings(instance);
   if (!CheckPosixCacheSettings(settings))
     return NULL;
-
   UniquePtr<PosixCacheManager> cache_mgr(PosixCacheManager::Create(
     settings.cache_path,
     settings.is_alien,
     settings.avoid_rename ? PosixCacheManager::kRenameLink
-                          : PosixCacheManager::kRenameNormal));
+                          : PosixCacheManager::kRenameNormal,
+    settings.do_refcount));
   if (!cache_mgr.IsValid()) {
     boot_error_ = "Failed to setup posix cache '" + instance + "' in " +
                   settings.cache_path + ": " + strerror(errno);
@@ -632,7 +716,7 @@ CacheManager *FileSystem::SetupPosixCacheMgr(const string &instance) {
   CreateFile(settings.cache_path + "/.cvmfscache", 0600, ignore_failure);
 
   if (settings.is_managed) {
-    if (!SetupPosixQuotaMgr(settings, cache_mgr))
+    if (!SetupPosixQuotaMgr(settings, cache_mgr.weak_ref()))
       return NULL;
   }
   return cache_mgr.Release();
@@ -778,24 +862,41 @@ bool FileSystem::SetupCwd() {
 }
 
 
-void FileSystem::SetupLogging() {
+/**
+ * Environment variables useful, e.g., for variant symlinks
+ */
+void FileSystem::SetupGlobalEnvironmentParams() {
+  setenv("CVMFS_ARCH", GetArch().c_str(), 1 /* overwrite */);
+}
+
+
+void FileSystem::SetupLoggingStandalone(
+  const OptionsManager &options_mgr, const std::string &prefix)
+{
+  SetupGlobalEnvironmentParams();
+
   string optarg;
-  if (options_mgr_->GetValue("CVMFS_SYSLOG_LEVEL", &optarg))
+  if (options_mgr.GetValue("CVMFS_SYSLOG_LEVEL", &optarg))
     SetLogSyslogLevel(String2Uint64(optarg));
-  if (options_mgr_->GetValue("CVMFS_SYSLOG_FACILITY", &optarg))
+  if (options_mgr.GetValue("CVMFS_SYSLOG_FACILITY", &optarg))
     SetLogSyslogFacility(String2Int64(optarg));
-  if (options_mgr_->GetValue("CVMFS_USYSLOG", &optarg))
+  if (options_mgr.GetValue("CVMFS_USYSLOG", &optarg))
     SetLogMicroSyslog(optarg);
-  if (options_mgr_->GetValue("CVMFS_DEBUGLOG", &optarg))
+  if (options_mgr.GetValue("CVMFS_DEBUGLOG", &optarg))
     SetLogDebugFile(optarg);
-  if (options_mgr_->GetValue("CVMFS_SYSLOG_PREFIX", &optarg)) {
+  if (options_mgr.GetValue("CVMFS_SYSLOG_PREFIX", &optarg)) {
     SetLogSyslogPrefix(optarg);
   } else {
-    SetLogSyslogPrefix(name_);
+    SetLogSyslogPrefix(prefix);
   }
 }
 
 
+void FileSystem::SetupLogging() {
+  SetupLoggingStandalone(*options_mgr_, name_);
+}
+
+
 bool FileSystem::SetupNfsMaps() {
 #ifdef CVMFS_NFS_SUPPORT
   if (!IsHaNfsSource())
@@ -1033,7 +1134,8 @@ void FileSystem::SetupUuid() {
  * cache in order to properly unravel the file system stack on shutdown.
  */
 void FileSystem::TearDown2ReadOnly() {
-  if ((cache_mgr_ != NULL) && (cache_mgr_->id() == kPosixCacheManager)) {
+  if ((cache_mgr_ != NULL) &&
+      (cache_mgr_->id() == kPosixCacheManager)) {
     PosixCacheManager *posix_cache_mgr =
       reinterpret_cast<PosixCacheManager *>(cache_mgr_);
     posix_cache_mgr->TearDown2ReadOnly();
@@ -1049,6 +1151,11 @@ void FileSystem::RemapCatalogFd(int from, int to) {
   sqlite::RegisterFdMapping(from, to);
 }
 
+void FileSystem::ReplaceCacheManager(CacheManager *new_cache_mgr) {
+  cache_mgr_ = new_cache_mgr;
+  sqlite::ReplaceCacheManager(new_cache_mgr);
+}
+
 
 bool FileSystem::TriageCacheMgr() {
   cache_mgr_instance_ = kDefaultCacheMgrInstance;
@@ -1062,7 +1169,24 @@ bool FileSystem::TriageCacheMgr() {
   }
 
   cache_mgr_ = SetupCacheMgr(cache_mgr_instance_);
-  return cache_mgr_ != NULL;
+  if (cache_mgr_ == NULL)
+    return false;
+
+  std::string optarg;
+  if (options_mgr_->GetValue("CVMFS_STREAMING_CACHE", &optarg) &&
+      options_mgr_->IsOn(optarg))
+  {
+    unsigned nfiles = kDefaultNfiles;
+    if (options_mgr_->GetValue("CVMFS_NFILES", &optarg))
+      nfiles = String2Uint64(optarg);
+    size_t buffer_size = StreamingCacheManager::kDefaultBufferSize;
+    if (options_mgr_->GetValue("CVMFS_STREAMING_CACHE_BUFFER_SIZE", &optarg))
+      buffer_size = String2Uint64(optarg);
+    cache_mgr_ = new StreamingCacheManager(nfiles, cache_mgr_, NULL, NULL,
+                                           buffer_size, statistics_);
+  }
+
+  return true;
 }
 
 
@@ -1126,6 +1250,26 @@ bool MountPoint::ReloadBlacklists() {
   return result;
 }
 
+/**
+ * Disables kernel caching of symlinks.
+ * Symlink caching requires fuse >= 3.10 (FUSE_CAP_CACHE_SYMLINKS) and
+ * linux kernel >= 4.2. Some OS might backport it.
+ *
+ * NOTE: This function should only be called before or within cvmfs_init().
+ *
+ */
+void MountPoint::DisableCacheSymlinks() {
+  cache_symlinks_ = false;
+}
+
+/**
+ * Instead of invalidate dentries, they should be expired.
+ * Fixes issues with mount-on-top mounts and symlink caching.
+ */
+void MountPoint::EnableFuseExpireEntry() {
+  fuse_expire_entry_ = true;
+}
+
 
 /**
  * The option_mgr parameter can be NULL, in which case the global option manager
@@ -1152,6 +1296,13 @@ MountPoint *MountPoint::Create(
     return mountpoint.Release();
   if (!mountpoint->CreateDownloadManagers())
     return mountpoint.Release();
+  if (file_system->cache_mgr()->id() == kStreamingCacheManager) {
+    StreamingCacheManager *streaming_cachemgr =
+      dynamic_cast<StreamingCacheManager *>(file_system->cache_mgr());
+    streaming_cachemgr->SetRegularDownloadManager(mountpoint->download_mgr());
+    streaming_cachemgr->SetExternalDownloadManager(
+      mountpoint->external_download_mgr());
+  }
   if (!mountpoint->CreateResolvConfWatcher()) {
     return mountpoint.Release();
   }
@@ -1163,7 +1314,8 @@ MountPoint *MountPoint::Create(
 
   mountpoint->ReEvaluateAuthz();
   mountpoint->CreateTables();
-  mountpoint->SetupBehavior();
+  if (!mountpoint->SetupBehavior())
+    return mountpoint.Release();
 
   mountpoint->boot_status_ = loader::kFailOk;
   return mountpoint.Release();
@@ -1255,16 +1407,42 @@ bool MountPoint::CreateCatalogManager() {
 
 bool MountPoint::CreateDownloadManagers() {
   string optarg;
-  download_mgr_ = new download::DownloadManager();
-  const bool use_system_proxy = false;
-  download_mgr_->Init(kDefaultNumConnections, use_system_proxy,
-                      perf::StatisticsTemplate("download", statistics_));
+  download_mgr_ = new download::DownloadManager(kDefaultNumConnections,
+                             perf::StatisticsTemplate("download", statistics_));
   download_mgr_->SetCredentialsAttachment(authz_attachment_);
 
-  if (options_mgr_->GetValue("CVMFS_SERVER_URL", &optarg)) {
+  // must be set before proxy and host chains are being initialized
+  // error output is handled in SetShardingPolicy
+  if (options_mgr_->GetValue("CVMFS_PROXY_SHARDING_POLICY", &optarg)) {
+    if (optarg.compare("EXTERNAL") == 0) {
+      download_mgr_->SetShardingPolicy(download::kShardingPolicyExternal);
+      download_mgr_->SetFqrn(fqrn());
+    }
+  }
+
+  if (options_mgr_->GetValue("CVMFS_FAILOVER_INDEFINITELY", &optarg) &&
+      options_mgr_->IsOn(optarg)) {
+      download_mgr_->SetFailoverIndefinitely();
+  }
+
+  if (options_mgr_->GetValue("CVMFS_METALINK_URL", &optarg)) {
+    download_mgr_->SetMetalinkChain(optarg);  
+    // host chain will be set later when the metalink server is contacted
+    download_mgr_->SetHostChain("");
+    // metalink requires redirects
+    download_mgr_->EnableRedirects();
+  } else if (options_mgr_->GetValue("CVMFS_SERVER_URL", &optarg)) {
     download_mgr_->SetHostChain(optarg);
   }
 
+  if (options_mgr_->GetValue("_CVMFS_DEVEL_IGNORE_SIGNATURE_FAILURES", &optarg)
+      && options_mgr_->IsOn(optarg)) {
+    download_mgr_->EnableIgnoreSignatureFailures();
+    LogCvmfs(kLogCvmfs, kLogDebug | kLogSyslogWarn,
+      "Development option: Activate ignore signature failures during download. "
+      "DO NOT USE IN PRODUCTION");
+  }
+
   SetupDnsTuning(download_mgr_);
   SetupHttpTuning();
 
@@ -1306,6 +1484,62 @@ bool MountPoint::CreateDownloadManagers() {
       download_mgr_->SetHostChain(host_chain);
     }
   }
+
+  if (options_mgr_->GetValue("CVMFS_USE_SSL_SYSTEM_CA", &optarg) &&
+      options_mgr_->IsOn(optarg)) {
+    download_mgr_->UseSystemCertificatePath();
+  }
+
+  if (options_mgr_->GetValue("CVMFS_PROXY_SHARD", &optarg) &&
+      options_mgr_->IsOn(optarg)) {
+    download_mgr_->ShardProxies();
+  }
+
+  // configure http tracing header
+  if (options_mgr_->GetValue("CVMFS_HTTP_TRACING", &optarg) &&
+      options_mgr_->IsOn(optarg)) {
+    download_mgr_->EnableHTTPTracing();
+    if (options_mgr_->GetValue("CVMFS_HTTP_TRACING_HEADERS", &optarg)) {
+      if (optarg.size() > 1000) {
+        LogCvmfs(kLogCvmfs, kLogSyslogErr | kLogDebug,
+           "CVMFS_HTTP_TRACING_HEADERS too large ( max 1000 chars, given %ld )",
+           optarg.size());
+      } else {
+        std::vector<std::string> tokens = SplitString(optarg, '|');
+        sanitizer::AlphaNumSanitizer sanitizer;
+
+        for (unsigned int i = 0; i < tokens.size(); i++) {
+          std::string token = Trim(tokens[i]);
+
+          std::vector<std::string> key_val = SplitString(token, ':');
+
+          if (key_val.size() != 2) {
+            LogCvmfs(kLogCvmfs, kLogSyslogErr | kLogDebug,
+              "Http tracing header: Skipping current token part of "
+              "CVMFS_HTTP_TRACING_HEADERS! Invalid "
+              "<key:value> pair. Token: %s", token.c_str());
+            continue;
+          }
+
+          std::string prefix = "X-CVMFS-";
+          std::string key = Trim(key_val[0]);
+
+          if (!sanitizer.IsValid(key)) {
+            LogCvmfs(kLogCvmfs, kLogSyslogErr | kLogDebug,
+            "Http tracing header: Skipping current token part of "
+            "CVMFS_HTTP_TRACING_HEADERS! Invalid key. Only alphanumeric keys "
+            "are allowed (a-z, A-Z, 0-9). Token: %s", token.c_str());
+            continue;
+          }
+
+          std::string final_token = prefix + key + ": " + Trim(key_val[1]);
+
+          download_mgr_->AddHTTPTracingHeader(final_token);
+        }
+      }
+    }
+  }
+
   return SetupExternalDownloadMgr(do_geosort);
 }
 
@@ -1318,7 +1552,7 @@ bool MountPoint::CreateResolvConfWatcher() {
              "DNS roaming is enabled for this repository.");
     // Create a file watcher to update the DNS settings of the download
     // managers when there are changes to /etc/resolv.conf
-    resolv_conf_watcher_ = platform_file_watcher();
+    resolv_conf_watcher_ = file_watcher::FileWatcher::Create();
 
     if (resolv_conf_watcher_) {
       ResolvConfEventHandler *handler =
@@ -1339,13 +1573,11 @@ void MountPoint::CreateFetchers() {
     backoff_throttle_,
     perf::StatisticsTemplate("fetch", statistics_));
 
-  const bool is_external_data = true;
   external_fetcher_ = new cvmfs::Fetcher(
     file_system_->cache_mgr(),
     external_download_mgr_,
     backoff_throttle_,
-    perf::StatisticsTemplate("fetch-external", statistics_),
-    is_external_data);
+    perf::StatisticsTemplate("fetch-external", statistics_));
 }
 
 
@@ -1370,15 +1602,12 @@ bool MountPoint::CreateSignatureManager() {
     boot_status_ = loader::kFailSignature;
     return false;
   }
-  LogCvmfs(kLogCvmfs, kLogDebug, "CernVM-FS: using public key(s) %s",
-           public_keys.c_str());
 
-  if (options_mgr_->GetValue("CVMFS_TRUSTED_CERTS", &optarg)) {
-    if (!signature_mgr_->LoadTrustedCaCrl(optarg)) {
-      boot_error_ = "failed to load trusted certificates";
-      boot_status_ = loader::kFailSignature;
-      return false;
-    }
+  if (public_keys.size() > 0) {
+    LogCvmfs(kLogCvmfs, kLogDebug, "CernVM-FS: using public key(s) %s",
+                                   public_keys.c_str());
+  } else {
+    LogCvmfs(kLogCvmfs, kLogDebug | kLogSyslogWarn, "no public key loaded");
   }
 
   return true;
@@ -1404,12 +1633,23 @@ void MountPoint::CreateStatistics() {
   statistics_->Register("inode_tracker.n_miss_path",
                         "overall number of unsuccessful path lookups");
 
-  statistics_->Register("nentry_tracker.n_insert",
+  statistics_->Register("dentry_tracker.n_insert",
                         "overall number of added negative cache entries");
-  statistics_->Register("nentry_tracker.n_remove",
+  statistics_->Register("dentry_tracker.n_remove",
                         "overall number of evicted negative cache entries");
-  statistics_->Register("nentry_tracker.n_prune",
+  statistics_->Register("dentry_tracker.n_prune",
                         "overall number of prune calls");
+
+  statistics_->Register("page_cache_tracker.n_insert",
+                        "overall number of added page cache entries");
+  statistics_->Register("page_cache_tracker.n_remove",
+                        "overall number of evicted page cache entries");
+  statistics_->Register("page_cache_tracker.n_open_direct",
+                        "overall number of direct I/O open calls");
+  statistics_->Register("page_cache_tracker.n_open_flush",
+    "overall number of open calls where the file's page cache gets flushed");
+  statistics_->Register("page_cache_tracker.n_open_cached",
+    "overall number of open calls where the file's page cache is reused");
 }
 
 
@@ -1441,7 +1681,10 @@ void MountPoint::CreateTables() {
                                          statistics_);
 
   inode_tracker_ = new glue::InodeTracker();
-  nentry_tracker_ = new glue::NentryTracker();
+  dentry_tracker_ = new glue::DentryTracker();
+  page_cache_tracker_ = new glue::PageCacheTracker();
+  if (file_system_->IsNfsSource())
+    page_cache_tracker_->Disable();
 }
 
 /**
@@ -1504,7 +1747,7 @@ bool MountPoint::DetermineRootHash(shash::Any *root_hash) {
   UnlinkGuard history_file(history_path);
   UniquePtr<history::History> tag_db(
     history::SqliteHistory::Open(history_path));
-  if (!tag_db) {
+  if (!tag_db.IsValid()) {
     LogCvmfs(kLogCvmfs, kLogDebug | kLogSyslog,
              "failed to open history database (%s)", history_path.c_str());
     boot_error_ = "failed to open history database";
@@ -1569,12 +1812,10 @@ bool MountPoint::FetchHistory(std::string *history_path) {
     return false;
   }
 
-  int fd = fetcher_->Fetch(
-    history_hash,
-    CacheManager::kSizeUnknown,
-    "tag database for " + fqrn_,
-    zlib::kZlibDefault,
-    CacheManager::kTypeRegular);
+  CacheManager::Label label;
+  label.flags = CacheManager::kLabelHistory;
+  label.path = fqrn_;
+  int fd = fetcher_->Fetch(CacheManager::LabeledObject(history_hash, label));
   if (fd < 0) {
     boot_error_ = "failed to download history: " + StringifyInt(-fd);
     boot_status_ = loader::kFailHistory;
@@ -1622,6 +1863,7 @@ MountPoint::MountPoint(
   , file_system_(file_system)
   , options_mgr_(options_mgr)
   , statistics_(NULL)
+  , telemetry_aggr_(NULL)
   , authz_fetcher_(NULL)
   , authz_session_mgr_(NULL)
   , authz_attachment_(NULL)
@@ -1640,14 +1882,20 @@ MountPoint::MountPoint(
   , md5path_cache_(NULL)
   , tracer_(NULL)
   , inode_tracker_(NULL)
-  , nentry_tracker_(NULL)
+  , dentry_tracker_(NULL)
+  , page_cache_tracker_(NULL)
+  , statfs_cache_(NULL)
   , resolv_conf_watcher_(NULL)
   , max_ttl_sec_(kDefaultMaxTtlSec)
   , kcache_timeout_sec_(static_cast<double>(kDefaultKCacheTtlSec))
   , fixed_catalog_(false)
-  , hide_magic_xattrs_(false)
   , enforce_acls_(false)
+  , cache_symlinks_(false)
+  , fuse_expire_entry_(false)
   , has_membership_req_(false)
+  , talk_socket_path_(std::string("./cvmfs_io.") + fqrn)
+  , talk_socket_uid_(0)
+  , talk_socket_gid_(0)
 {
   int retval = pthread_mutex_init(&lock_max_ttl_, NULL);
   assert(retval == 0);
@@ -1657,7 +1905,8 @@ MountPoint::MountPoint(
 MountPoint::~MountPoint() {
   pthread_mutex_destroy(&lock_max_ttl_);
 
-  delete nentry_tracker_;
+  delete page_cache_tracker_;
+  delete dentry_tracker_;
   delete inode_tracker_;
   delete tracer_;
   delete md5path_cache_;
@@ -1670,14 +1919,10 @@ MountPoint::~MountPoint() {
   delete inode_annotation_;
   delete external_fetcher_;
   delete fetcher_;
-  if (external_download_mgr_ != NULL) {
-    external_download_mgr_->Fini();
-    delete external_download_mgr_;
-  }
-  if (download_mgr_ != NULL) {
-    download_mgr_->Fini();
-    delete download_mgr_;
-  }
+
+  delete external_download_mgr_;
+  delete download_mgr_;
+
   if (signature_mgr_ != NULL) {
     signature_mgr_->Fini();
     delete signature_mgr_;
@@ -1692,8 +1937,11 @@ MountPoint::~MountPoint() {
   delete authz_attachment_;
   delete authz_session_mgr_;
   delete authz_fetcher_;
+  delete telemetry_aggr_;
   delete statistics_;
   delete uuid_;
+
+  delete statfs_cache_;
 }
 
 
@@ -1712,8 +1960,7 @@ void MountPoint::SetMaxTtlMn(unsigned value_minutes) {
   max_ttl_sec_ = value_minutes * 60;
 }
 
-
-void MountPoint::SetupBehavior() {
+bool MountPoint::SetupBehavior() {
   string optarg;
 
   if (options_mgr_->GetValue("CVMFS_MAX_TTL", &optarg))
@@ -1727,17 +1974,121 @@ void MountPoint::SetupBehavior() {
   LogCvmfs(kLogCvmfs, kLogDebug, "kernel caches expire after %d seconds",
            static_cast<int>(kcache_timeout_sec_));
 
-  if (options_mgr_->GetValue("CVMFS_HIDE_MAGIC_XATTRS", &optarg)
-      && options_mgr_->IsOn(optarg))
-  {
-    hide_magic_xattrs_ = true;
+  uint64_t statfs_time_cache_valid = 0;
+  if (options_mgr_->GetValue("CVMFS_STATFS_CACHE_TIMEOUT", &optarg)) {
+    statfs_time_cache_valid = static_cast<uint64_t>(String2Uint64(optarg));
   }
+  LogCvmfs(kLogCvmfs, kLogDebug, "statfs cache expires after %d seconds",
+           static_cast<int>(statfs_time_cache_valid));
+  statfs_cache_ = new StatfsCache(statfs_time_cache_valid);
+
+  MagicXattrManager::EVisibility xattr_visibility =
+    MagicXattrManager::kVisibilityRootOnly;
+  if (options_mgr_->GetValue("CVMFS_HIDE_MAGIC_XATTRS", &optarg)) {
+    if (options_mgr_->IsOn(optarg))
+      xattr_visibility = MagicXattrManager::kVisibilityNever;
+    else if (options_mgr_->IsOff(optarg))
+      xattr_visibility = MagicXattrManager::kVisibilityAlways;
+  }
+  if (options_mgr_->GetValue("CVMFS_MAGIC_XATTRS_VISIBILITY", &optarg)) {
+    if (ToUpper(optarg) == "ROOTONLY") {
+      xattr_visibility = MagicXattrManager::kVisibilityRootOnly;
+    } else if (ToUpper(optarg) == "NEVER") {
+      xattr_visibility = MagicXattrManager::kVisibilityNever;
+    } else if (ToUpper(optarg) == "ALWAYS") {
+      xattr_visibility = MagicXattrManager::kVisibilityAlways;
+    } else {
+      LogCvmfs(kLogCvmfs, kLogSyslogWarn | kLogDebug,
+               "unsupported setting: CVMFS_MAGIC_XATTRS_VISIBILITY=%s",
+               optarg.c_str());
+    }
+  }
+
+  std::set<gid_t> protected_xattr_gids;
+  if (options_mgr_->GetValue("CVMFS_XATTR_PRIVILEGED_GIDS", &optarg)) {
+    std::vector<string> tmp = SplitString(optarg, ',');
+
+    for (size_t i = 0; i < tmp.size(); i++) {
+      std::string trimmed = Trim(tmp[i]);
+      LogCvmfs(kLogCvmfs, kLogDebug,
+               "Privileged gid for xattr added: %s", trimmed.c_str());
+      protected_xattr_gids.insert(static_cast<gid_t>(String2Uint64(trimmed)));
+    }
+  }
+  std::set<std::string> protected_xattrs;
+  if (options_mgr_->GetValue("CVMFS_XATTR_PROTECTED_XATTRS", &optarg)) {
+    std::vector<string> tmp = SplitString(optarg, ',');
+
+    for (size_t i = 0; i < tmp.size(); i++) {
+      std::string trimmed = Trim(tmp[i]);
+      LogCvmfs(kLogCvmfs, kLogDebug,
+               "Protected xattr added: %s", trimmed.c_str());
+      protected_xattrs.insert(trimmed);
+    }
+
+    // root has always access to xattr
+    if (protected_xattr_gids.count(0) < 1) {
+      protected_xattr_gids.insert(0);
+      LogCvmfs(kLogCvmfs, kLogDebug,
+              "Automatically added root to have access to protected xattrs.");
+    }
+  }
+  magic_xattr_mgr_ = new MagicXattrManager(this, xattr_visibility,
+                                    protected_xattrs, protected_xattr_gids);
+
 
   if (options_mgr_->GetValue("CVMFS_ENFORCE_ACLS", &optarg)
       && options_mgr_->IsOn(optarg))
   {
     enforce_acls_ = true;
   }
+
+  if (options_mgr_->GetValue("CVMFS_CACHE_SYMLINKS", &optarg)
+      && options_mgr_->IsOn(optarg))
+  {
+    cache_symlinks_ = true;
+  }
+
+
+
+  if (options_mgr_->GetValue("CVMFS_TALK_SOCKET", &optarg)) {
+    talk_socket_path_ = optarg;
+  }
+  if (options_mgr_->GetValue("CVMFS_TALK_OWNER", &optarg)) {
+    bool retval = GetUidOf(optarg, &talk_socket_uid_, &talk_socket_gid_);
+    if (!retval) {
+      boot_error_ = "unknown owner of cvmfs_talk socket: " + optarg;
+      boot_status_ = loader::kFailOptions;
+      return false;
+    }
+  }
+
+  // this can be later be changed to switch through different
+  // telemetryAggregators
+  if (options_mgr_->GetValue("CVMFS_TELEMETRY_SEND", &optarg)
+      && options_mgr_->IsOn(optarg)) {
+    int telemetry_send_rate_sec = kDefaultTelemetrySendRateSec;
+    if (options_mgr_->GetValue("CVMFS_TELEMETRY_RATE", &optarg)) {
+      telemetry_send_rate_sec = static_cast<int>(String2Uint64(optarg));
+
+      // minimum send rate: 5sec
+      if (telemetry_send_rate_sec < kMinimumTelemetrySendRateSec) {
+        telemetry_send_rate_sec = kMinimumTelemetrySendRateSec;
+      }
+
+      telemetry_aggr_ = perf::TelemetryAggregator::Create(statistics_,
+                                                        telemetry_send_rate_sec,
+                                                        options_mgr_,
+                                                        this,
+                                                        fqrn_,
+                                                        perf::kTelemetryInflux);
+      LogCvmfs(kLogTelemetry, kLogSyslog | kLogDebug,
+               "Enable telemetry to report every %d seconds",
+               telemetry_send_rate_sec);
+    }
+  }
+
+  return true;
 }
 
 
@@ -1789,7 +2140,7 @@ bool MountPoint::SetupExternalDownloadMgr(bool dogeosort) {
   string optarg;
   external_download_mgr_ =
     download_mgr_->Clone(perf::StatisticsTemplate("download-external",
-      statistics_));
+      statistics_), "external");
 
   unsigned timeout;
   unsigned timeout_direct;
@@ -1802,7 +2153,13 @@ bool MountPoint::SetupExternalDownloadMgr(bool dogeosort) {
   }
   external_download_mgr_->SetTimeout(timeout, timeout_direct);
 
-  if (options_mgr_->GetValue("CVMFS_EXTERNAL_URL", &optarg)) {
+  if (options_mgr_->GetValue("CVMFS_EXTERNAL_METALINK", &optarg)) {
+    external_download_mgr_->SetMetalinkChain(optarg);  
+    // host chain will be set later when the metalink server is contacted
+    external_download_mgr_->SetHostChain("");
+    // metalink requires redirects
+    external_download_mgr_->EnableRedirects();
+  } else if (options_mgr_->GetValue("CVMFS_EXTERNAL_URL", &optarg)) {
     external_download_mgr_->SetHostChain(optarg);
     if (dogeosort) {
       std::vector<std::string> host_chain;
@@ -1872,8 +2229,13 @@ void MountPoint::SetupHttpTuning() {
 
   if (options_mgr_->GetValue("CVMFS_LOW_SPEED_LIMIT", &optarg))
     download_mgr_->SetLowSpeedLimit(String2Uint64(optarg));
-  if (options_mgr_->GetValue("CVMFS_PROXY_RESET_AFTER", &optarg))
+  if (options_mgr_->GetValue("CVMFS_PROXY_RESET_AFTER", &optarg)) {
     download_mgr_->SetProxyGroupResetDelay(String2Uint64(optarg));
+    // Use the proxy reset delay as the default for the metalink reset delay
+    download_mgr_->SetMetalinkResetDelay(String2Uint64(optarg));
+  }
+  if (options_mgr_->GetValue("CVMFS_METALINK_RESET_AFTER", &optarg))
+    download_mgr_->SetMetalinkResetDelay(String2Uint64(optarg));
   if (options_mgr_->GetValue("CVMFS_HOST_RESET_AFTER", &optarg))
     download_mgr_->SetHostResetDelay(String2Uint64(optarg));
 
@@ -1935,6 +2297,13 @@ bool MountPoint::SetupOwnerMaps() {
   {
     g_claim_ownership = true;
   }
+  if (options_mgr_->GetValue("CVMFS_WORLD_READABLE", &optarg) &&
+      options_mgr_->IsOn(optarg))
+  {
+    g_world_readable = true;
+  }
+
+
 
   return true;
 }
