@@ -14,7 +14,7 @@ namespace notify {
 /**
  * Implementation of Publisher based on HTTP
  *
- * Messsages are published to the notification system backend using cURL
+ * Messages are published to the notification system backend using cURL
  */
 class PublisherHTTP : public Publisher {
  public:
@@ -25,8 +25,6 @@ class PublisherHTTP : public Publisher {
 
  private:
   std::string server_url_;
-
-  bool initialized_;
 };
 
 }  // namespace notify
