@@ -7,7 +7,7 @@
 #
 
 
-# This file depends on fuctions implemented in the following files:
+# This file depends on functions implemented in the following files:
 # - cvmfs_server_util.sh
 # - cvmfs_server_json.sh
 
@@ -96,7 +96,7 @@ get_or_guess_multiple_repository_names() {
     else
       for name in $names; do
         if ! contains "$repo_names" $(basename $name); then
-          repo_names="$(basename $name) $repo_names"
+          repo_names="$repo_names $(basename $name)"
         fi
       done
     fi
@@ -165,6 +165,9 @@ get_repository_name() {
 # loads the configuration for a specific repository
 load_repo_config() {
   local name=$1
+  if [ ! -e /etc/cvmfs/repositories.d/${name}/server.conf ]; then
+    die "Error: The repository $name does not exist."
+  fi
   . /etc/cvmfs/repositories.d/${name}/server.conf
   if [ x"$CVMFS_REPOSITORY_TYPE" = x"stratum0" ]; then
     . /etc/cvmfs/repositories.d/${name}/client.conf
@@ -210,6 +213,7 @@ get_tag_hash() {
     -t ${CVMFS_SPOOL_DIR}/tmp                 \
     -p /etc/cvmfs/keys/${repository_name}.pub \
     -f $repository_name                       \
+    $(get_swissknife_proxy)                   \
     $(get_follow_http_redirects_flag) -x      \
     -n "$tag" 2>/dev/null | cut -d" " -f2
 }
@@ -230,8 +234,9 @@ get_tag_branch() {
     -t ${CVMFS_SPOOL_DIR}/tmp                 \
     -p /etc/cvmfs/keys/${repository_name}.pub \
     -f $repository_name                       \
+    $(get_swissknife_proxy)                   \
     $(get_follow_http_redirects_flag) -x      \
-    -n "$tag" 2>/dev/null | cut -d" " -f7)
+    -n "$tag" 2>/dev/null | cut -d" " -f6)
   if [ "x$branch" = "x(default)" ]; then
     branch=
   fi
@@ -268,8 +273,9 @@ get_head_of() {
     -t ${CVMFS_SPOOL_DIR}/tmp                 \
     -p /etc/cvmfs/keys/${repository_name}.pub \
     -f $repository_name                       \
+    $(get_swissknife_proxy)                   \
     $(get_follow_http_redirects_flag) -x      \
-    | cut -d" " -f1,2,7 | grep " $branch\$" | head -n 1
+    | cut -d" " -f1,2,6 | grep " $branch\$" | head -n 1
 }
 
 
@@ -332,7 +338,7 @@ set_ro_root_hash() {
 get_repo_info_from_url() {
   local url="$1"
   shift 1
-  __swissknife info $(get_follow_http_redirects_flag) -r "$url" $@ 2>/dev/null
+  __swissknife info $(get_follow_http_redirects_flag) $(get_swissknife_proxy) -r "$url" $@ 2>/dev/null
 }
 
 
@@ -350,9 +356,8 @@ get_repo_info() {
 is_in_transaction() {
   local name=$1
   load_repo_config $name
-  # ignore if the lock is "stale" because the process starting
-  #  transactions goes away
-  check_lock ${CVMFS_SPOOL_DIR}/in_transaction ignore_stale
+  local tx_flag="${CVMFS_SPOOL_DIR}/in_transaction"
+  check_flag "${tx_flag}"
 }
 
 
@@ -503,19 +508,15 @@ is_due_auto_garbage_collection() {
 # download a given file from the backend storage
 # @param name  the name of the repository to download from
 # @param url   the url to download from
-# @param noproxy  (optional)
 get_item() {
   local name="$1"
   local url="$2"
-  local noproxy="$3"
 
   load_repo_config $name
 
-  if [ x"$noproxy" != x"" ]; then
-    unset http_proxy && curl -f $(get_follow_http_redirects_flag) "$url" 2>/dev/null | tr -d '\0'
-  else
-    curl -f $(get_follow_http_redirects_flag) "$url" 2>/dev/null | tr -d '\0'
-  fi
+  curl -f -H "Cache-Control: max-age=0" $(get_curl_proxy) \
+       $(get_x509_cert_settings) $(get_follow_http_redirects_flag) \
+       "$url" 2>/dev/null | tr -d '\0'
 }
 
 # read an item from local or backend repository storage to stdout
@@ -530,11 +531,11 @@ read_repo_item() {
   load_repo_config $name
 
   if is_local_upstream $CVMFS_UPSTREAM_STORAGE; then
-    cat $(get_upstream_config $upstream)/"$item" 2>/dev/null
+    cat $(get_upstream_config $CVMFS_UPSTREAM_STORAGE)/"$item" 2>/dev/null
   elif is_stratum0 $name; then
-    get_item $name $CVMFS_STRATUM0/"$item" noproxy
+    get_item $name $CVMFS_STRATUM0/"$item"
   else
-    get_item $name $CVMFS_STRATUM1/"$item" noproxy
+    get_item $name $CVMFS_STRATUM1/"$item"
   fi
 }
 
@@ -547,6 +548,27 @@ get_follow_http_redirects_flag() {
 }
 
 
+# Parse special CA path settings for curl invocation
+get_x509_cert_settings() {
+  if [ x"$X509_CERT_BUNDLE" != "x" ]; then
+      echo "--cacert $X509_CERT_BUNDLE"
+  fi
+}
+
+# Parse proxy server for curl command
+get_curl_proxy() {
+  if [ x"$CVMFS_SERVER_PROXY" != x"" ]; then
+    echo "-x $CVMFS_SERVER_PROXY"
+  fi
+}
+
+# Parse proxy server for cvmfs_swissknife command
+get_swissknife_proxy() {
+  if [ x"$CVMFS_SERVER_PROXY" != x"" ]; then
+    echo "-@ $CVMFS_SERVER_PROXY"
+  fi
+}
+
 get_expiry_from_string() {
   local whitelist="$1"
 
@@ -573,11 +595,11 @@ get_expiry_from_string() {
 # figures out the time to expiry of the repository's whitelist
 #
 # @param stratum0  path/URL to stratum0 storage
-# @return          number of seconds until expiry (negativ if already expired)
+# @return          number of seconds until expiry (negative if already expired)
 get_expiry() {
   local name=$1
   local stratum0=$2
-  get_expiry_from_string "$(get_item $name $stratum0/.cvmfswhitelist 'noproxy')"
+  get_expiry_from_string "$(get_item $name $stratum0/.cvmfswhitelist)"
 }
 
 
@@ -647,6 +669,7 @@ sign_manifest() {
           -u $CVMFS_STRATUM0                   \
           -m $unsigned_manifest                \
           -t ${CVMFS_SPOOL_DIR}/tmp            \
+          $(get_swissknife_proxy)              \
           -r $CVMFS_UPSTREAM_STORAGE $return_early"
 
   if [ x"$metainfo_file" != x"" ]; then
@@ -669,12 +692,10 @@ sign_manifest() {
 open_transaction() {
   local name=$1
   load_repo_config $name
-  local tx_lock="${CVMFS_SPOOL_DIR}/in_transaction"
+  local tx_flag="${CVMFS_SPOOL_DIR}/in_transaction"
 
   is_stratum0 $name                    || die "Cannot open transaction on Stratum1"
-  # ignore if lock is "stale" because the process that created it
-  #  goes away
-  acquire_lock "$tx_lock" ignore_stale || die "Failed to create transaction lock"
+  set_flag "$tx_flag"                  || die "Failed to create transaction flag"
   run_suid_helper open $name           || die "Failed to make /cvmfs/$name writable"
 
   to_syslog_for_repo $name "opened transaction"
@@ -696,7 +717,7 @@ close_transaction() {
   is_in_transaction $name || return 0
 
   load_repo_config $name
-  local tx_lock="${CVMFS_SPOOL_DIR}/in_transaction"
+  local tx_flag="${CVMFS_SPOOL_DIR}/in_transaction"
   local tmp_dir="${CVMFS_SPOOL_DIR}/tmp"
   local current_scratch_dir="${CVMFS_SPOOL_DIR}/scratch/current"
   local wastebin_scratch_dir="${CVMFS_SPOOL_DIR}/scratch/wastebin"
@@ -745,7 +766,7 @@ close_transaction() {
   [ ! -z "$tmp_dir" ] && find "${tmp_dir}" -mindepth 1 -not -path '*/receiver*' | xargs rm -fR
   run_suid_helper rdonly_mount $name > /dev/null
   run_suid_helper rw_mount $name
-  release_lock "$tx_lock"
+  clear_flag "$tx_flag"
 
   # Remove session_token file, used for gateway transactions, if it exists
   rm -f ${CVMFS_SPOOL_DIR}/session_token
@@ -818,6 +839,55 @@ acquire_update_lock()
 }
 
 
+# Release a gc lock
+#
+# @param name             the repository to release
+
+release_gc_lock() {
+  local name=$1
+
+  load_repo_config $name
+  release_lock ${CVMFS_SPOOL_DIR}/is_collecting || echo "Warning: failed to release gc lock"
+}
+
+# Acquire a gc lock for a repository.  Always pair with a call to
+#   release_gc_lock if returns successful.
+#
+# @param name               the repository to lock
+# @param check_type         check type, either check or gc
+# @param abort_on_conflict  0 to wait for lock, 1 to abort if already acquired.
+#                           Default 0.
+# @return                   0 if lock successfully acquired
+
+acquire_gc_lock()
+{
+  local name=$1
+  local check_type=$2
+  local abort_on_conflict=${3:-0}
+
+  load_repo_config $name
+  local gc_lock=${CVMFS_SPOOL_DIR}/is_collecting
+
+  # look for another check or gc in progress
+  if ! acquire_lock $gc_lock; then
+    if [ $abort_on_conflict -eq 1 ]; then
+      echo "A check or other gc on $name is in progress... aborting"
+      to_syslog_for_repo $name "did not $check_type (check or gc in progress)"
+      return 1
+    fi
+
+    echo "Waiting for gc on $name to finish..."
+    if ! wait_and_acquire_lock $gc_lock; then
+      echo "failed to acquire gc lock"
+      to_syslog_for_repo $name "did not $check_type (locking issues)"
+      return 1
+    fi
+  fi
+
+  # The lock is now acquired
+}
+
+
 handle_read_only_file_descriptors_on_mount_point() {
   local name=$1
   local open_fd_dialog=${2:-1}
@@ -902,6 +972,7 @@ create_config_files_for_new_repository() {
   local external_data=${11}
   local voms_authz=${12}
   local auto_tag_timespan="${13}"
+  local proxy_url=${14}
 
   # other configurations
   local spool_dir="/var/spool/cvmfs/${name}"
@@ -939,6 +1010,9 @@ CVMFS_AUTO_REPAIR_MOUNTPOINT=true
 CVMFS_AUTOCATALOGS=false
 CVMFS_ASYNC_SCRATCH_CLEANUP=true
 CVMFS_PRINT_STATISTICS=false
+CVMFS_UPLOAD_STATS_DB=false
+CVMFS_UPLOAD_STATS_PLOTS=false
+CVMFS_IGNORE_XDIR_HARDLINKS=true
 EOF
 
   if [ x"$voms_authz" != x"" ]; then
@@ -953,6 +1027,10 @@ CVMFS_AUTO_GC=true
 EOF
   fi
 
+  if [ x"$proxy_url" != x"" ]; then
+    echo "CVMFS_SERVER_PROXY=$proxy_url" >> $server_conf
+  fi
+
   if [ $configure_apache -eq 1 ] && is_local_upstream $upstream; then
     local repository_dir=$(get_upstream_config $upstream)
     # make sure that the config file does not exist, yet
@@ -968,18 +1046,26 @@ CVMFS_RELOAD_SOCKETS=$cache_dir
 CVMFS_QUOTA_LIMIT=4000
 CVMFS_MOUNT_DIR=/cvmfs
 CVMFS_SERVER_URL=$stratum0
-CVMFS_HTTP_PROXY=DIRECT
+CVMFS_HTTP_PROXY=${proxy_url:-DIRECT}
 CVMFS_PUBLIC_KEY=/etc/cvmfs/keys/${name}.pub
-CVMFS_TRUSTED_CERTS=${repo_cfg_dir}/trusted_certs
 CVMFS_CHECK_PERMISSIONS=yes
 CVMFS_IGNORE_SIGNATURE=no
 CVMFS_AUTO_UPDATE=no
 CVMFS_NFS_SOURCE=no
-CVMFS_HIDE_MAGIC_XATTRS=yes
+CVMFS_MAGIC_XATTRS_VISIBILITY=never
 CVMFS_FOLLOW_REDIRECTS=yes
 CVMFS_SERVER_CACHE_MODE=yes
 CVMFS_NFILES=65536
+CVMFS_TALK_SOCKET=/var/spool/cvmfs/${name}/cvmfs_io
+CVMFS_TALK_OWNER=$cvmfs_user
+CVMFS_USE_SSL_SYSTEM_CA=true
 EOF
+
+  if [ "x$X509_CERT_BUNDLE" != "x" ]; then
+    cat >> $client_conf << EOF
+X509_CERT_BUNDLE=$X509_CERT_BUNDLE
+EOF
+  fi
 }
 
 
@@ -1108,7 +1194,7 @@ setup_and_mount_new_repository() {
   if [ x"$CVMFS_UNION_FS_TYPE" = x"overlayfs" ]; then
     echo -n "(overlayfs) "
     cat >> /etc/fstab << EOF
-cvmfs2#$name $rdonly_dir fuse allow_other,config=/etc/cvmfs/repositories.d/${name}/client.conf:${CVMFS_SPOOL_DIR}/client.local,cvmfs_suid,noauto 0 0 # added by CernVM-FS for $name
+cvmfs2#$name $rdonly_dir fuse allow_other,fsname=$name,config=/etc/cvmfs/repositories.d/${name}/client.conf:${CVMFS_SPOOL_DIR}/client.local,cvmfs_suid,noauto 0 0 # added by CernVM-FS for $name
 overlay_$name /cvmfs/$name overlay upperdir=${scratch_dir},lowerdir=${rdonly_dir},workdir=$ofs_workdir,noauto,nodev,ro 0 0 # added by CernVM-FS for $name
 EOF
   else
@@ -1117,16 +1203,18 @@ EOF
       selinux_context=",context=\"system_u:object_r:default_t:s0\""
     fi
     cat >> /etc/fstab << EOF
-cvmfs2#$name $rdonly_dir fuse allow_other,config=/etc/cvmfs/repositories.d/${name}/client.conf:${CVMFS_SPOOL_DIR}/client.local,cvmfs_suid,noauto 0 0 # added by CernVM-FS for $name
+cvmfs2#$name $rdonly_dir fuse allow_other,fsname=$name,config=/etc/cvmfs/repositories.d/${name}/client.conf:${CVMFS_SPOOL_DIR}/client.local,cvmfs_suid,noauto 0 0 # added by CernVM-FS for $name
 aufs_$name /cvmfs/$name aufs br=${scratch_dir}=rw:${rdonly_dir}=rr,udba=none,noauto,nodev,ro$selinux_context 0 0 # added by CernVM-FS for $name
 EOF
   fi
   local user_shell="$(get_user_shell $name)"
   $user_shell "touch ${CVMFS_SPOOL_DIR}/client.local"
 
-  # avoid racing against apache
+  # avoid racing against apache; we can safely ignore the certificate validation
+  # at this step, we only want to check that the endpoint is up.
+  # NB: Normally, we are anyway dealing with HTTP URLs at this point.
   local waiting=0
-  while ! curl -sIf ${CVMFS_STRATUM0}/.cvmfspublished > /dev/null && \
+  while ! curl $(get_curl_proxy) --insecure -sIf ${CVMFS_STRATUM0}/.cvmfspublished > /dev/null && \
         [ $http_timeout -gt 0 ]; do
     [ $waiting -eq 1 ] || echo -n "waiting for apache... "
     waiting=1
@@ -1158,7 +1246,17 @@ is_stratum0_garbage_collectable() {
 }
 
 
-# checks if a manifest ist present
+# checks if a manifest is present
+#
+# @param url  the url of the repository to be checked
+# @return      0 if it is empty
+is_empty_repository_from_url() {
+  local url=$1
+  [ x"$(get_repo_info_from_url "$url" -e)" = x"yes" ]
+}
+
+
+# checks if a manifest is present
 #
 # @param name  the name of the repository to be checked
 # @return      0 if it is empty
@@ -1304,5 +1402,3 @@ _run_catalog_migration() {
   sign_manifest $name $manifest      || die "Signing failed";
   set_ro_root_hash $name $trunk_hash || die "Root hash update failed";
 }
-
-
