@@ -117,16 +117,16 @@ check_overlayfs() {
 
 
 # check if at least one of the supported union file systems is available
-# currently AUFS get preference over OverlayFS if both are available
+# currently OverlayFS get preference over AUFS if both are available
 #
 # @return   0 if at least one was found (name through stdout); abort otherwise
 get_available_union_fs() {
-  if check_aufs; then
-    echo "aufs"
-  elif check_overlayfs; then
+  if check_overlayfs; then
     echo "overlayfs"
+  elif check_aufs; then
+    echo "aufs"
   else
-    die "neither AUFS nor OverlayFS detected on the system!"
+    die "neither OverlayFS nor AUFS detected on the system!"
   fi
 }
 
@@ -146,6 +146,21 @@ __swissknife() {
 }
 
 
+__publish_cmd() {
+  local might_be_debugging="$1"
+  if [ ! -z $might_be_debugging ]; then
+    echo "$CVMFS_SERVER_PUBLISH_DEBUG"
+  else
+    echo "$CVMFS_SERVER_PUBLISH"
+  fi
+}
+
+
+__publish() {
+  $(__publish_cmd) $@
+}
+
+
 # checks if a given list of strings contains a specific item
 #
 # @param haystack   the list to be searched
@@ -222,7 +237,7 @@ is_mounted() {
   # Use canonicalize-missing because the mount point can be broken and
   # inaccessible
   absolute_mnt="$(readlink --canonicalize-missing $mountpoint)" || return 2
-  local mnt_record="$(cat /proc/mounts 2>/dev/null | grep " $absolute_mnt ")"
+  local mnt_record="$(cat /proc/mounts 2>/dev/null | grep -F " $absolute_mnt ")"
   if [ x"$mnt_record" = x"" ]; then
     return 1
   fi
@@ -274,7 +289,7 @@ count_wr_fds() {
   local path=$1
   local cnt=0
   for line in $(get_fd_modes $path); do
-    if echo "$line" | grep -qe '^\a[wu]$'; then cnt=$(( $cnt + 1 )); fi
+    if echo "$line" | grep -qe '^rw'; then cnt=$(( $cnt + 1 )); fi
   done
   echo $cnt
 }
@@ -347,72 +362,57 @@ __hc_transition() {
 
 ### Locking functions
 
-# Helper functions for file locking including detection of stale locks
-# Note: The implementation idea was found here:
-#       http://rute.2038bug.com/node23.html.gz
-__is_valid_lock() {
+
+acquire_lock() {
   local path="$1"
-  local ignore_stale="$2"
-
   local lock_file="${path}.lock"
-  cvmfs_sys_file_is_regular $lock_file || return 1 # lock doesn't exist
-  [ -z "$ignore_stale" ] || return 0 # lock is there (skip the stale test)
-
-  local stale_pid=$(cat $lock_file 2>/dev/null)
-  [ -n "$stale_pid" ] && [ $stale_pid -gt 0 ]     && \
-  kill -0 $stale_pid 2>/dev/null
-}
-
-
-acquire_lock() { # hardlink creation is guaranteed to be atomic!
-  local path="$1"
-  local ignore_stale="$2"
-
-  local pid="$$"
-  local temp_file="${path}.${pid}"
-  local lock_file="${path}.lock"
-  echo $pid > $temp_file || return 1 # probably no access to $path
-
-  if ln $temp_file $lock_file 2>/dev/null; then
-    rm -f $temp_file 2>/dev/null
-    return 0 # lock acquired
-  fi
-
-  if __is_valid_lock "$path" "$ignore_stale"; then
-    rm -f $temp_file 2>/dev/null
-    return 1 # lock couldn't be acquired and appears valid
-  fi
-
-  rm -f $lock_file 2>/dev/null # lock was stale and can be removed
-  if ln $temp_file $lock_file; then
-    rm -f $temp_file 2>/dev/null
-    return 0 # lock acquired
-  fi
-
-  rm -f $temp_file 2>/dev/null
-  return 1 # lock couldn't be acquired after removing stale lock (lost the race)
+  exec 9<>${lock_file}
+  flock -n 9
 }
 
 
 wait_and_acquire_lock() {
   local path="$1"
-  while ! acquire_lock "$path"; do
-    sleep 5
-  done
+  local lock_file="${path}.lock"
+  exec 9<>${lock_file}
+  flock 9
 }
 
 
 release_lock() {
   local path="$1"
   local lock_file="${path}.lock"
-  rm -f $lock_file 2>/dev/null
+  rm -f $lock_file
+  exec 9<&-
 }
 
 
 check_lock() {
   local path="$1"
-  local ignore_stale="$2"
-  __is_valid_lock "${path}" "$ignore_stale"
+  if ! acquire_lock "${path}" ; then
+    return 0
+  fi
+  release_lock "${path}"
+  return 1
+}
+
+
+set_flag() {
+  local path="$1"
+  local flag_file="${path}.lock"
+  touch "${flag_file}"
+}
+
+clear_flag() {
+  local path="$1"
+  local flag_file="${path}.lock"
+  rm "${flag_file}"
+}
+
+check_flag() {
+  local path="$1"
+  local flag_file="${path}.lock"
+  [ -f "${flag_file}" ]
 }
 
 
@@ -445,7 +445,7 @@ cvmfs_version_string() {
 
 # Tracks changes to the organization of files and directories.
 # Stored in CVMFS_CREATOR_VERSION.  Started with 137.
-cvmfs_layout_revision() { echo "141"; }
+cvmfs_layout_revision() { echo "143"; }
 
 version_major() { echo $1 | cut --delimiter=. --fields=1 | grep -oe '^[0-9]\+'; }
 version_minor() { echo $1 | cut --delimiter=. --fields=2 | grep -oe '^[0-9]\+'; }
@@ -553,7 +553,7 @@ check_upstream_validity() {
 # @return  0 if overlayfs is installed and viable
 #          1 if it is not viable, and stdout contains a reason
 # This should probably now be called check_overlayfs_viability except
-#   that for backward compatiblity we need to keep the variable that
+#   that for backward compatibility we need to keep the variable that
 #   overrides it, CVMFS_DONT_CHECK_OVERLAYFS_VERSION, and changing the
 #   function name would make the variable name not make sense.
 check_overlayfs_version() {
@@ -634,13 +634,15 @@ _setcap_if_needed() {
 }
 
 
-# grants CAP_SYS_ADMIN to cvmfs_swissknife if it is necessary
+# grants CAP_SYS_ADMIN to cvmfs_swissknife and cvmfs_publish if it is necessary
 # Note: OverlayFS uses trusted extended attributes that are not readable by a
 #       normal unprivileged process
 ensure_swissknife_suid() {
   local unionfs="$1"
   local sk_bin="/usr/bin/$CVMFS_SERVER_SWISSKNIFE"
   local sk_dbg_bin="/usr/bin/${CVMFS_SERVER_SWISSKNIFE}_debug"
+  local pb_bin="/usr/bin/cvmfs_publish"
+  local pb_dbg_bin="/usr/bin/cvmfs_publish_debug"
   local cap_read="cap_dac_read_search"
   local cap_overlay="cap_sys_admin"
 
@@ -649,9 +651,13 @@ ensure_swissknife_suid() {
   if [ x"$unionfs" = x"overlayfs" ]; then
     _setcap_if_needed "$sk_bin"     "${cap_read},${cap_overlay}+p" || return 3
     _setcap_if_needed "$sk_dbg_bin" "${cap_read},${cap_overlay}+p" || return 4
+    _setcap_if_needed "$pb_bin"     "${cap_read},${cap_overlay}+p" || return 5
+    _setcap_if_needed "$pb_dbg_bin" "${cap_read},${cap_overlay}+p" || return 6
   else
     _setcap_if_needed "$sk_bin"     "${cap_read}+p" || return 1
     _setcap_if_needed "$sk_dbg_bin" "${cap_read}+p" || return 2
+    _setcap_if_needed "$pb_bin"     "${cap_read}+p" || return 7
+    _setcap_if_needed "$pb_dbg_bin" "${cap_read}+p" || return 8
   fi
 }
 
@@ -727,6 +733,54 @@ has_selinux() {
   $GETENFORCE_BIN | grep -qi "enforc" || return 1
 }
 
+# Functions to retrieve information about the operating system.
+# Used in cvmfs_server_json.sh to populate meta.json.
+# Fails gracefully if no os-release file is found (values become empty strings)
+_os_set_etc_release_filename() {
+  _CVMFS_OS_RELEASE_FILENAME=''
+  if test -e "/etc/os-release"; then
+    _CVMFS_OS_RELEASE_FILENAME="/etc/os-release";
+  elif test -e "/usr/lib/os-release"; then
+    _CVMFS_OS_RELEASE_FILENAME="/usr/lib/os-release";
+  fi
+}
+
+# Pick a specific field from the os-release file and return it.
+# 1. _os_set_etc_release_filename is called on every check, this is probably overkill,
+#    but the question would be where to call this once and for all. In theory one could
+#    call it right after the declaration above?
+# 2. If a fieldname isn't provided, die due to wrongful usage of the (internal) function.
+# 3. If we couldn't find an os-release file, return nothing.
+# 4. If we grep fails to find anything, return nothing. We may want to check against known
+#    keys from https://www.freedesktop.org/software/systemd/man/os-release.html?
+_os_etc_release_get_field() {
+  _os_set_etc_release_filename
+
+  local fieldname=$1
+  if [ "x${fieldname}" = "x" ]; then
+    die "Internal error: _os_etc_release_get_field expects a field name to search for!"
+  fi
+
+  if [ "x${_CVMFS_OS_RELEASE_FILENAME}" = "x" ]; then
+    # If we are unable to find a proper /etc/os-release file return nothing.
+    :
+  else
+    sed -n "s/\"//g;s/^${fieldname}=//p" ${_CVMFS_OS_RELEASE_FILENAME}
+  fi
+}
+
+_os_id() {
+  _os_etc_release_get_field 'ID'
+}
+
+_os_version_id() {
+  _os_etc_release_get_field 'VERSION_ID'
+}
+
+_os_pretty_name() {
+  _os_etc_release_get_field 'PRETTY_NAME'
+}
+
 
 set_selinux_httpd_context_if_needed() {
   local directory="$1"
@@ -785,7 +839,7 @@ If you go for production, backup your masterkey from /etc/cvmfs/keys/!"
 
 get_fd_modes() {
   local path=$1
-  $LSOF_BIN -Fan 2>/dev/null | grep -B1 -e "^n$path" | grep -e '^a.*'
+  cvmfs_publish lsof $path
 }
 
 # gets the number of open read-only file descriptors beneath a given path
@@ -796,7 +850,7 @@ count_rd_only_fds() {
   local path=$1
   local cnt=0
   for line in $(get_fd_modes $path); do
-    if echo "$line" | grep -qe '^\ar\?$';  then cnt=$(( $cnt + 1 )); fi
+    if echo "$line" | grep -qe '^ro';  then cnt=$(( $cnt + 1 )); fi
   done
   echo $cnt
 }
@@ -819,7 +873,7 @@ is_master_replica() {
   local name=$1
   local is_master_replica
 
-  if [ $(echo $name | cut --bytes=1-7) = "http://" ]; then
+  if [ $(echo $name | cut --bytes=1-7) = "http://" -o $(echo $name | cut --bytes=1-8) = "https://" ]; then
     is_master_replica=$(get_repo_info_from_url $name -m -L)
   else
     load_repo_config $name
@@ -859,8 +913,12 @@ minpidof() {
 }
 
 
+# Return true if systemd is to be used, that is, if neither the service
+# command nor the supervisorctl command has been located.  Note that
+# the service command is only looked for if systemctl is missing, but
+# the supervisorctl command will take precedence over systemctl.
 is_systemd() {
-  [ x"$SERVICE_BIN" = x"false" ]
+  [ x"$SERVICE_BIN" = x"false" ] && [ x"$SUPERVISOR_BIN" = x"false" ]
 }
 
 
@@ -903,12 +961,16 @@ _to_syslog_for_geoip() {
 
 _update_geodb_install() {
   local retcode=0
-  local datname="$2"
-  local dburl="${CVMFS_UPDATEGEO_URLBASE}?edition_id=${CVMFS_UPDATEGEO_DB%.*}&suffix=tar.gz&license_key=$CVMFS_GEO_LICENSE_KEY"
+  local dburl="${CVMFS_UPDATEGEO_URLBASE}?suffix=tar.gz"
   local dbfile="${CVMFS_UPDATEGEO_DIR}/${CVMFS_UPDATEGEO_DB}"
   local download_target=${dbfile}.tgz
   local untar_dir=${dbfile}.untar
 
+  if [ -z "$CVMFS_GEO_ACCOUNT_ID" ]; then
+      echo "CVMFS_GEO_ACCOUNT_ID not set" >&2
+      _to_syslog_for_geoip "CVMFS_GEO_ACCOUNT_ID not set"
+      return 1
+  fi
   if [ -z "$CVMFS_GEO_LICENSE_KEY" ]; then
       echo "CVMFS_GEO_LICENSE_KEY not set" >&2
       _to_syslog_for_geoip "CVMFS_GEO_LICENSE_KEY not set"
@@ -918,8 +980,9 @@ _update_geodb_install() {
   _to_syslog_for_geoip "started update from $dburl"
 
   # downloading the GeoIP database file
-  curl -sS  --connect-timeout 10 \
+  curl -L -sS  --connect-timeout 10 \
             --max-time 60        \
+            -u "${CVMFS_GEO_ACCOUNT_ID}:${CVMFS_GEO_LICENSE_KEY}" \
             "$dburl" > $download_target || true
   if ! tar tzf $download_target >/dev/null 2>&1; then
     local msg
@@ -955,6 +1018,11 @@ _update_geodb_install() {
     return 3
   fi
 
+  if [ -w "$(get_global_info_v1_path)" ]; then
+    # update repositories.json for the new geodb timestamp, if possible
+    update_global_repository_info || die "fail (update global repository info)"
+  fi
+
   # get rid of other files in the untar
   rm -rf $untar_dir
 
@@ -1050,11 +1118,11 @@ cvmfs_server_update_geodb() {
 # @return   0 if the command was recognized
 is_subcommand() {
   local subcommand="$1"
-  local supported_commands="mkfs add-replica import publish rollback rmfs alterfs    \
-    resign list info tag list-tags lstags check transaction abort snapshot           \
-    skeleton migrate list-catalogs diff checkout update-geodb gc catalog-chown \
-    eliminate-hardlinks update-info update-repoinfo mount fix-permissions \
-    masterkeycard ingest merge-stats print-stats"
+  local supported_commands="mkfs add-replica import publish rollback rmfs alterfs   \
+    resign list info tag list-tags lstags check transaction enter abort snapshot    \
+    skeleton migrate list-catalogs diff checkout update-geodb gc catalog-chown      \
+    eliminate-hardlinks eliminate-bulk-hashes fix-stats update-info update-repoinfo \
+    mount fix-permissions masterkeycard ingest merge-stats print-stats"
 
   for possible_command in $supported_commands; do
     if [ x"$possible_command" = x"$subcommand" ]; then
@@ -1104,12 +1172,13 @@ Supported Commands:
                   [-k path to existing keychain] [-p no apache config]
                   [-R require masterkeycard key ]
                   [-V VOMS authorization] [-X (external data)]
+                  [-x proxy url]
                   <fully qualified repository name>
                   Creates a new repository with a given name
   add-replica     [-u stratum1 upstream storage] [-o owner] [-w stratum1 url]
                   [-a silence apache warning] [-z enable garbage collection]
                   [-n alias name] [-s S3 config file] [-p no apache config]
-                  [-g snapshot group]
+                  [-g snapshot group] [-P pass-through repository]
                   <stratum 0 url> <public key | keys directory>
                   Creates a Stratum 1 replica of a Stratum 0 repository
   import          [-w stratum0 url] [-o owner] [-u upstream storage]
@@ -1118,10 +1187,11 @@ Supported Commands:
                   [-k path to keys] [-g chown backend] [-r recreate whitelist]
                   [-p no apache config] [-t recreate repo key and certificate]
                   [ -R recreate whitelist and require masterkeycard ]
+                  [-x proxy url]
                   <fully qualified repository name>
                   Imports an old CernVM-FS repository into a fresh repo
   publish         [-p pause for tweaks] [-n manual revision number] [-v verbose]
-                  [-a tag name] [-c tag channel] [-m tag description]
+                  [-a tag name] [-m tag description]
                   [-X (force external data) | -N (force native data)]
                   [-Z compression algorithm] [-F authz info file]
                   [-f use force remount if necessary]
@@ -1173,7 +1243,7 @@ Supported Commands:
   info            <fully qualified name>
                   Print summary about the repository
   tag             Create and manage named snapshots
-                  [-a create tag <name>] [-c channel] [-m message] [-h hash]
+                  [-a create tag <name>] [-m message] [-h hash]
                   [-r remove tag <name>] [-f don't ask again]
                   [-i inspect tag <name>] [-x machine readable]
                   [-b list branch hierarchy] [-x machine readable]
@@ -1187,12 +1257,12 @@ Supported Commands:
                   [-t tag (check given tag instead of trunk)]
                   [-s path to nested catalog subtree to check]
                   [-r repair reflog problems]
+                  [-a check all active local repos, log to checks.log |
+                    <fully qualified name> ]
                   <fully qualified name>
                   Checks if the repository is sane
-  transaction     [-r (retry if unable to acquire lease]
-                  [-i INT (initial retry delay seconds)]
-                  [-m INT (max retry delay seconds)]
-                  [-n INT (max number of retries)]
+  transaction     [-t (timeout in seconds for waiting if the repository is busy, 0=infinite)]
+                  [-T /template-from=/template-to]
                   <fully qualified name>
                   Start to edit a repository
   snapshot        [-t fail if other snapshot is in progress]
