@@ -6,10 +6,11 @@
 #define CVMFS_SIGNING_TOOL_H_
 
 #include <string>
+#include <vector>
 
-#include "hash.h"
+#include "crypto/hash.h"
 #include "server_tool.h"
-#include "util_concurrency.h"
+#include "util/future.h"
 
 class ServerTool;
 
@@ -37,9 +38,12 @@ class SigningTool {
              const std::string &repo_name = "", const std::string &pwd = "",
              const std::string &meta_info = "",
              const std::string &reflog_chksum_path = "",
+             const std::string &proxy = "",
              const bool garbage_collectable = false,
              const bool bootstrap_shortcuts = false,
-             const bool return_early = false);
+             const bool return_early = false,
+             const std::vector<shash::Any> reflog_catalogs =
+              std::vector<shash::Any>());
 
  protected:
   void CertificateUploadCallback(const upload::SpoolerResult &result);
