@@ -6,7 +6,7 @@
 
 #define __STDC_FORMAT_MACROS
 
-#include "cvmfs_config.h"
+
 #include "swissknife_check.h"
 
 #include <inttypes.h>
@@ -20,22 +20,40 @@
 #include <vector>
 
 #include "catalog_sql.h"
-#include "compression.h"
-#include "download.h"
+#include "compression/compression.h"
 #include "file_chunk.h"
 #include "history_sqlite.h"
-#include "logging.h"
 #include "manifest.h"
+#include "network/download.h"
 #include "reflog.h"
 #include "sanitizer.h"
 #include "shortstring.h"
+#include "util/exception.h"
+#include "util/logging.h"
 #include "util/pointer.h"
 #include "util/posix.h"
 
 using namespace std;  // NOLINT
 
+// for map of duplicate entries; as in kvstore.cc
+static inline uint32_t hasher_any(const shash::Any &key) {
+  // We'll just do the same thing as hasher_md5, since every hash is at
+  // least as large.
+  return *const_cast<uint32_t *>(
+             reinterpret_cast<const uint32_t *>(key.digest) + 1);
+}
+
+
 namespace swissknife {
 
+CommandCheck::CommandCheck()
+    : check_chunks_(false)
+    , no_duplicates_map_(false)
+    , is_remote_(false) {
+    const shash::Any hash_null;
+    duplicates_map_.Init(16, hash_null, hasher_any);
+}
+
 bool CommandCheck::CompareEntries(const catalog::DirectoryEntry &a,
                                   const catalog::DirectoryEntry &b,
                                   const bool compare_names,
@@ -64,12 +82,12 @@ bool CommandCheck::CompareEntries(const catalog::DirectoryEntry &a,
     }
   }
   if (diffs & Difference::kLinkcount) {
-    LogCvmfs(kLogCvmfs, kLogStderr, "linkcounts differ: %lu / %lu",
+    LogCvmfs(kLogCvmfs, kLogStderr, "linkcounts differ: %u / %u",
              a.linkcount(), b.linkcount());
     retval = false;
   }
   if (diffs & Difference::kHardlinkGroup) {
-    LogCvmfs(kLogCvmfs, kLogStderr, "hardlink groups differ: %lu / %lu",
+    LogCvmfs(kLogCvmfs, kLogStderr, "hardlink groups differ: %u / %u",
              a.hardlink_group(), b.hardlink_group());
     retval = false;
   }
@@ -79,7 +97,7 @@ bool CommandCheck::CompareEntries(const catalog::DirectoryEntry &a,
     retval = false;
   }
   if (diffs & Difference::kMode) {
-    LogCvmfs(kLogCvmfs, kLogStderr, "modes differ: %lu / %lu",
+    LogCvmfs(kLogCvmfs, kLogStderr, "modes differ: %u / %u",
              a.mode(), b.mode());
     retval = false;
   }
@@ -110,6 +128,18 @@ bool CommandCheck::CompareEntries(const catalog::DirectoryEntry &a,
              a.name().c_str(), b.name().c_str());
     retval = false;
   }
+  if (!is_transition_point) {
+      if (diffs & Difference::kUid) {
+        LogCvmfs(kLogCvmfs, kLogStderr, "uids differ: %d / %d (%s / %s)",
+                 a.uid(), b.uid(), a.name().c_str(), b.name().c_str());
+        retval = false;
+      }
+      if (diffs & Difference::kGid) {
+        LogCvmfs(kLogCvmfs, kLogStderr, "gids differ: %d / %d (%s / %s)",
+                 a.gid(), b.gid(), a.name().c_str(), b.name().c_str());
+        retval = false;
+      }
+    }
 
   return retval;
 }
@@ -142,7 +172,7 @@ bool CommandCheck::CompareCounters(const catalog::Counters &a,
 
 
 /**
- * Checks for existance of a file either locally or via HTTP head
+ * Checks for existence of a file either locally or via HTTP head
  */
 bool CommandCheck::Exists(const string &file)
 {
@@ -150,6 +180,7 @@ bool CommandCheck::Exists(const string &file)
     return FileExists(file) || SymlinkExists(file);
   } else {
     const string url = repo_base_path_ + "/" + file;
+    LogCvmfs(kLogCvmfs, kLogVerboseMsg, "[Exists::url] %s", url.c_str());
     download::JobInfo head(&url, false);
     return download_manager()->Fetch(&head) == download::kFailOk;
   }
@@ -167,17 +198,16 @@ string CommandCheck::FetchPath(const string &path) {
 
   const string url = repo_base_path_ + "/" + path;
   if (is_remote_) {
-    download::JobInfo download_job(&url, false, false, f, NULL);
+    cvmfs::FileSink filesink(f);
+    download::JobInfo download_job(&url, false, false, NULL, &filesink);
     download::Failures retval = download_manager()->Fetch(&download_job);
     if (retval != download::kFailOk) {
-      LogCvmfs(kLogCvmfs, kLogStderr, "failed to read %s", url.c_str());
-      abort();
+      PANIC(kLogStderr, "failed to read %s", url.c_str());
     }
   } else {
     bool retval = CopyPath2File(url, f);
     if (!retval) {
-      LogCvmfs(kLogCvmfs, kLogStderr, "failed to read %s", url.c_str());
-      abort();
+      PANIC(kLogStderr, "failed to read %s", url.c_str());
     }
   }
 
@@ -267,7 +297,7 @@ bool CommandCheck::InspectHistory(history::History *history) {
 
   bool result = true;
 
-  map<string, unsigned> initial_revisions;
+  map<string, uint64_t> initial_revisions;
   sanitizer::BranchSanitizer sanitizer;
   for (unsigned i = 0; i < branches.size(); ++i) {
     if (!sanitizer.IsValid(branches[i].branch)) {
@@ -282,7 +312,7 @@ bool CommandCheck::InspectHistory(history::History *history) {
   // TODO(jblomer): same root hash implies same size and revision
   for (unsigned i = 0; i < tags.size(); ++i) {
     used_branches.insert(tags[i].branch);
-    map<string, unsigned>::const_iterator iter =
+    const map<string, uint64_t>::const_iterator iter =
       initial_revisions.find(tags[i].branch);
     if (iter == initial_revisions.end()) {
       LogCvmfs(kLogCvmfs, kLogStderr, "invalid branch %s in tag %s",
@@ -290,8 +320,8 @@ bool CommandCheck::InspectHistory(history::History *history) {
       result = false;
     } else {
       if (tags[i].revision < iter->second) {
-        LogCvmfs(kLogCvmfs, kLogStderr, "invalid revision %u of tag %s",
-               tags[i].revision, tags[i].name.c_str());
+        LogCvmfs(kLogCvmfs, kLogStderr, "invalid revision %" PRIu64
+                 " of tag %s", tags[i].revision, tags[i].name.c_str());
         result = false;
       }
     }
@@ -308,6 +338,10 @@ bool CommandCheck::InspectHistory(history::History *history) {
 
 /**
  * Recursive catalog walk-through
+ *
+ * TODO(vavolkl): This method is large and does a lot of checks
+ * that could be split into smaller ones.
+ *
  */
 bool CommandCheck::Find(const catalog::Catalog *catalog,
                         const PathString &path,
@@ -335,12 +369,23 @@ bool CommandCheck::Find(const catalog::Catalog *catalog,
   bool found_nested_marker = false;
 
   for (unsigned i = 0; i < entries.size(); ++i) {
+    // for performance reasons, keep track of files already checked
+    // and only run requests once per hash
+    const bool entry_needs_check =
+          !entries[i].checksum().IsNull() && !entries[i].IsExternalFile() &&
+         // fallback cli option can force the entry to be checked
+          (no_duplicates_map_ ||
+             !duplicates_map_.Contains(entries[i].checksum()));
+    if (entry_needs_check && !no_duplicates_map_)
+        duplicates_map_.Insert(entries[i].checksum(), 1);
+
     PathString full_path(path);
     full_path.Append("/", 1);
     full_path.Append(entries[i].name().GetChars(),
                      entries[i].name().GetLength());
-    LogCvmfs(kLogCvmfs, kLogVerboseMsg, "[path] %s",
-             full_path.c_str());
+    LogCvmfs(kLogCvmfs, kLogVerboseMsg, "[path] %s [needs check] %i",
+             full_path.c_str(), entry_needs_check);
+
 
     // Name must not be empty
     if (entries[i].name().IsEmpty()) {
@@ -370,8 +415,7 @@ bool CommandCheck::Find(const catalog::Catalog *catalog,
     }
 
     // Check if the chunk is there
-    if (check_chunks_ &&
-        !entries[i].checksum().IsNull() && !entries[i].IsExternalFile())
+    if (check_chunks_ && entry_needs_check)
     {
       string chunk_path = "data/" + entries[i].checksum().MakePath();
       if (entries[i].IsDirectory())
@@ -414,6 +458,13 @@ bool CommandCheck::Find(const catalog::Catalog *catalog,
     }
 
     // Checks depending of entry type
+    if (!entries[i].IsRegular()) {
+      if (entries[i].IsDirectIo()) {
+        LogCvmfs(kLogCvmfs, kLogStderr, "invalid direct i/o flag found: %s",
+                 full_path.c_str());
+        retval = false;
+      }
+    }
     if (entries[i].IsDirectory()) {
       computed_counters->self.directories++;
       num_subdirs++;
@@ -478,8 +529,8 @@ bool CommandCheck::Find(const catalog::Catalog *catalog,
       }
       // Right size of symbolic link?
       if (entries[i].size() != entries[i].symlink().GetLength()) {
-        LogCvmfs(kLogCvmfs, kLogStderr, "wrong symbolic link size for %s; ",
-                 "expected %s, got %s", full_path.c_str(),
+        LogCvmfs(kLogCvmfs, kLogStderr, "wrong symbolic link size for %s; "
+                 "expected %u, got %lu", full_path.c_str(),
                  entries[i].symlink().GetLength(), entries[i].size());
         retval = false;
       }
@@ -558,17 +609,27 @@ bool CommandCheck::Find(const catalog::Catalog *catalog,
         aggregated_file_size += this_chunk.size();
 
         // are all data chunks in the data store?
-        if (check_chunks_) {
+        if (check_chunks_ && !entries[i].IsExternalFile()) {
           const shash::Any &chunk_hash = this_chunk.content_hash();
-          const string chunk_path = "data/" + chunk_hash.MakePath();
-          if (!Exists(chunk_path)) {
-            LogCvmfs(kLogCvmfs, kLogStderr, "partial data chunk %s (%s -> "
-                                            "offset: %d | size: %d) missing",
-                     this_chunk.content_hash().ToStringWithSuffix().c_str(),
-                     full_path.c_str(),
-                     this_chunk.offset(),
-                     this_chunk.size());
-            retval = false;
+          // for performance reasons, only perform the check once
+          // and skip if the hash has been checked before
+          bool chunk_needs_check = true;
+          if (!no_duplicates_map_ && !duplicates_map_.Contains(chunk_hash)) {
+            duplicates_map_.Insert(chunk_hash, 1);
+          } else if (!no_duplicates_map_) {
+            chunk_needs_check = false;
+          }
+          if (chunk_needs_check) {
+           const string chunk_path = "data/" + chunk_hash.MakePath();
+           if (!Exists(chunk_path)) {
+              LogCvmfs(kLogCvmfs, kLogStderr, "partial data chunk %s (%s -> "
+                       "offset: %ld | size: %lu) missing",
+                       this_chunk.content_hash().ToStringWithSuffix().c_str(),
+                       full_path.c_str(),
+                       this_chunk.offset(),
+                       this_chunk.size());
+              retval = false;
+            }
           }
         }
       }
@@ -576,7 +637,7 @@ bool CommandCheck::Find(const catalog::Catalog *catalog,
       // is the aggregated chunk size equal to the actual file size?
       if (aggregated_file_size != entries[i].size()) {
         LogCvmfs(kLogCvmfs, kLogStderr, "chunks of file %s produce a size "
-                                        "mismatch. Calculated %d bytes | %d "
+                                        "mismatch. Calculated %zu bytes | %lu "
                                         "bytes expected",
                  full_path.c_str(),
                  aggregated_file_size,
@@ -598,7 +659,7 @@ bool CommandCheck::Find(const catalog::Catalog *catalog,
   // Check directory linkcount
   if (this_directory.linkcount() != num_subdirs + 2) {
     LogCvmfs(kLogCvmfs, kLogStderr, "wrong linkcount for %s; "
-             "expected %lu, got %lu",
+             "expected %u, got %u",
              path.c_str(), num_subdirs + 2, this_directory.linkcount());
     retval = false;
   }
@@ -609,7 +670,7 @@ bool CommandCheck::Find(const catalog::Catalog *catalog,
   {
     if (i->second[0].linkcount() != i->second.size()) {
       LogCvmfs(kLogCvmfs, kLogStderr, "hardlink linkcount wrong for %s, "
-               "expected %lu, got %lu",
+               "expected %lu, got %u",
                (path.ToString() + "/" + i->second[0].name().ToString()).c_str(),
                i->second.size(), i->second[0].linkcount());
       retval = false;
@@ -624,10 +685,13 @@ string CommandCheck::DownloadPiece(const shash::Any catalog_hash) {
   string source = "data/" + catalog_hash.MakePath();
   const string dest = temp_directory_ + "/" + catalog_hash.ToString();
   const string url = repo_base_path_ + "/" + source;
-  download::JobInfo download_catalog(&url, true, false, &dest, &catalog_hash);
+
+  cvmfs::PathSink pathsink(dest);
+  download::JobInfo download_catalog(&url, true, false, &catalog_hash,
+                                     &pathsink);
   download::Failures retval = download_manager()->Fetch(&download_catalog);
   if (retval != download::kFailOk) {
-    LogCvmfs(kLogCvmfs, kLogStderr, "failed to download catalog %s (%d)",
+    LogCvmfs(kLogCvmfs, kLogStderr, "failed to download object %s (%d)",
              catalog_hash.ToString().c_str(), retval);
     return "";
   }
@@ -664,7 +728,12 @@ catalog::Catalog* CommandCheck::FetchCatalog(const string      &path,
   catalog::Catalog *catalog =
                    catalog::Catalog::AttachFreely(path, tmp_file, catalog_hash);
   int64_t catalog_file_size = GetFileSize(tmp_file);
-  assert(catalog_file_size > 0);
+  if (catalog_file_size <= 0) {
+
+    LogCvmfs(kLogCvmfs, kLogStderr, "Error downloading catalog %s at %s %s",
+             catalog_hash.ToString().c_str(), path.c_str(), tmp_file.c_str() );
+    assert(catalog_file_size > 0);
+  }
   unlink(tmp_file.c_str());
 
   if ((catalog_size > 0) && (uint64_t(catalog_file_size) != catalog_size)) {
@@ -690,8 +759,7 @@ bool CommandCheck::FindSubtreeRootCatalog(const string &subtree_path,
   typedef vector<string> Tokens;
   const Tokens path_tokens = SplitString(subtree_path, '/');
 
-  string      current_path = "";
-  bool        found        = false;
+  string current_path = "";
 
   Tokens::const_iterator i    = path_tokens.begin();
   Tokens::const_iterator iend = path_tokens.end();
@@ -712,12 +780,11 @@ bool CommandCheck::FindSubtreeRootCatalog(const string &subtree_path,
           break;
         }
       } else {
-        found = true;
+        return true;
       }
     }
   }
-
-  return found;
+  return false;
 }
 
 
@@ -731,7 +798,7 @@ bool CommandCheck::InspectTree(const string                  &path,
                                const catalog::DirectoryEntry *transition_point,
                                catalog::DeltaCounters        *computed_counters)
 {
-  LogCvmfs(kLogCvmfs, kLogStdout, "[inspecting catalog] %s at %s",
+  LogCvmfs(kLogCvmfs, kLogStdout | kLogInform, "[inspecting catalog] %s at %s",
            catalog_hash.ToString().c_str(), path == "" ? "/" : path.c_str());
 
   const catalog::Catalog *catalog = FetchCatalog(path,
@@ -884,7 +951,6 @@ int CommandCheck::Main(const swissknife::ArgumentList &args) {
   string tag_name;
   string subtree_path = "";
   string pubkey_path = "";
-  string trusted_certs = "";
   string repo_name = "";
   string reflog_chksum_path = "";
 
@@ -894,9 +960,11 @@ int CommandCheck::Main(const swissknife::ArgumentList &args) {
     tag_name = *args.find('n')->second;
   if (args.find('c') != args.end())
     check_chunks_ = true;
+  if (args.find('d') != args.end())
+    no_duplicates_map_ = true;
   if (args.find('l') != args.end()) {
     unsigned log_level =
-      1 << (kLogLevel0 + String2Uint64(*args.find('l')->second));
+      kLogLevel0 << String2Uint64(*args.find('l')->second);
     if (log_level > kLogNone) {
       LogCvmfs(kLogCvmfs, kLogStderr, "invalid log level");
       return 1;
@@ -907,8 +975,6 @@ int CommandCheck::Main(const swissknife::ArgumentList &args) {
     pubkey_path = *args.find('k')->second;
   if (DirectoryExists(pubkey_path))
     pubkey_path = JoinStrings(FindFilesBySuffix(pubkey_path, ".pub"), ":");
-  if (args.find('z') != args.end())
-    trusted_certs = *args.find('z')->second;
   if (args.find('N') != args.end())
     repo_name = *args.find('N')->second;
 
@@ -919,12 +985,13 @@ int CommandCheck::Main(const swissknife::ArgumentList &args) {
     reflog_chksum_path = *args.find('R')->second;
 
   // Repository can be HTTP address or on local file system
-  is_remote_ = (repo_base_path_.substr(0, 7) == "http://");
+  is_remote_ = IsHttpUrl(repo_base_path_);
 
   // initialize the (swissknife global) download and signature managers
   if (is_remote_) {
     const bool follow_redirects = (args.count('L') > 0);
-    if (!this->InitDownloadManager(follow_redirects)) {
+    const string proxy = (args.count('@') > 0) ? *args.find('@')->second : "";
+    if (!this->InitDownloadManager(follow_redirects, proxy)) {
       return 1;
     }
 
@@ -934,7 +1001,7 @@ int CommandCheck::Main(const swissknife::ArgumentList &args) {
       return 1;
     }
 
-    if (!this->InitVerifyingSignatureManager(pubkey_path, trusted_certs)) {
+    if (!this->InitSignatureManager(pubkey_path)) {
       return 1;
     }
   }
@@ -959,6 +1026,21 @@ int CommandCheck::Main(const swissknife::ArgumentList &args) {
     return 1;
   }
 
+  // Check meta-info object
+  if (!manifest->meta_info().IsNull()) {
+    string tmp_file;
+    if (!is_remote_)
+      tmp_file = DecompressPiece(manifest->meta_info());
+    else
+      tmp_file = DownloadPiece(manifest->meta_info());
+    if (tmp_file == "") {
+      LogCvmfs(kLogCvmfs, kLogStderr, "failed to load repository metainfo %s",
+               manifest->meta_info().ToString().c_str());
+      return 1;
+    }
+    unlink(tmp_file.c_str());
+  }
+
   shash::Any reflog_hash;
   if (!reflog_chksum_path.empty()) {
     if (!manifest::Reflog::ReadChecksum(reflog_chksum_path, &reflog_hash)) {
@@ -976,7 +1058,7 @@ int CommandCheck::Main(const swissknife::ArgumentList &args) {
                ".cvmfsreflog present but no checksum provided, aborting");
       return 1;
     }
-    bool retval = InspectReflog(reflog_hash, manifest);
+    bool retval = InspectReflog(reflog_hash, manifest.weak_ref());
     if (!retval) {
       LogCvmfs(kLogCvmfs, kLogStderr, "failed to verify reflog");
       return 1;
@@ -1014,7 +1096,7 @@ int CommandCheck::Main(const swissknife::ArgumentList &args) {
       return 1;
     }
     tag_db->TakeDatabaseFileOwnership();
-    successful = InspectHistory(tag_db) && successful;
+    successful = InspectHistory(tag_db.weak_ref()) && successful;
   }
 
   if (manifest->has_alt_catalog_path()) {
@@ -1060,6 +1142,7 @@ int CommandCheck::Main(const swissknife::ArgumentList &args) {
     return 1;
   }
 
+
   catalog::DeltaCounters computed_counters;
   successful = InspectTree(subtree_path,
                            root_hash,
