@@ -6,7 +6,7 @@
  * We take all three volumes (namely union, overlay and repository) into
  * account to sync the changes back into the repository.
  *
- * On the repository side we have a catalogs directory that mimicks the
+ * On the repository side we have a catalogs directory that mimics the
  * shadow directory structure and stores compressed and uncompressed
  * versions of all catalogs.  The raw data are stored in the data
  * subdirectory in zlib-compressed form.  They are named with their SHA-1
@@ -16,12 +16,13 @@
  * to allow Apache to follow the symlinks.
  */
 
+// NOLINTNEXTLINE
 #define _FILE_OFFSET_BITS 64
-
+// NOLINTNEXTLINE
 #define __STDC_FORMAT_MACROS
 
 #include "swissknife_sync.h"
-#include "cvmfs_config.h"
+
 
 #include <errno.h>
 #include <fcntl.h>
@@ -38,19 +39,21 @@
 #include "catalog_mgr_ro.h"
 #include "catalog_mgr_rw.h"
 #include "catalog_virtual.h"
-#include "download.h"
-#include "logging.h"
 #include "manifest.h"
+#include "monitor.h"
+#include "network/download.h"
 #include "path_filters/dirtab.h"
-#include "platform.h"
 #include "reflog.h"
 #include "sanitizer.h"
 #include "statistics.h"
+#include "statistics_database.h"
 #include "swissknife_capabilities.h"
 #include "sync_mediator.h"
 #include "sync_union.h"
 #include "sync_union_aufs.h"
 #include "sync_union_overlayfs.h"
+#include "util/logging.h"
+#include "util/platform.h"
 #include "util/string.h"
 
 using namespace std;  // NOLINT
@@ -108,7 +111,7 @@ int swissknife::CommandCreate::Main(const swissknife::ArgumentList &args) {
   const string reflog_chksum_path = *args.find('R')->second;
   if (args.find('l') != args.end()) {
     unsigned log_level =
-        1 << (kLogLevel0 + String2Uint64(*args.find('l')->second));
+      kLogLevel0 << String2Uint64(*args.find('l')->second);
     if (log_level > kLogNone) {
       LogCvmfs(kLogCvmfs, kLogStderr, "invalid log level");
       return 1;
@@ -138,15 +141,15 @@ int swissknife::CommandCreate::Main(const swissknife::ArgumentList &args) {
 
   UniquePtr<manifest::Manifest> manifest(
       catalog::WritableCatalogManager::CreateRepository(
-          dir_temp, volatile_content, voms_authz, spooler));
+          dir_temp, volatile_content, voms_authz, spooler.weak_ref()));
   if (!manifest.IsValid()) {
-    PrintError("Failed to create new repository");
+    PrintError("Swissknife Sync: Failed to create new repository");
     return 1;
   }
 
   UniquePtr<manifest::Reflog> reflog(CreateEmptyReflog(dir_temp, repo_name));
   if (!reflog.IsValid()) {
-    PrintError("Failed to create fresh Reflog");
+    PrintError("Swissknife Sync: Failed to create fresh Reflog");
     return 1;
   }
 
@@ -159,7 +162,7 @@ int swissknife::CommandCreate::Main(const swissknife::ArgumentList &args) {
   spooler->WaitForUpload();
   unlink(reflog_path.c_str());
   if (spooler->GetNumberOfErrors()) {
-    LogCvmfs(kLogCvmfs, kLogStderr, "Failed to upload reflog");
+    LogCvmfs(kLogCvmfs, kLogStderr, "Swissknife Sync: Failed to upload reflog");
     return 4;
   }
   assert(!reflog_chksum_path.empty());
@@ -171,7 +174,7 @@ int swissknife::CommandCreate::Main(const swissknife::ArgumentList &args) {
   manifest->set_has_alt_catalog_path(needs_bootstrap_shortcuts);
 
   if (!manifest->Export(manifest_path)) {
-    PrintError("Failed to create new repository");
+    PrintError("Swissknife Sync: Failed to create new repository");
     return 5;
   }
 
@@ -186,7 +189,7 @@ int swissknife::CommandUpload::Main(const swissknife::ArgumentList &args) {
   if (args.find('a') != args.end()) {
     hash_algorithm = shash::ParseHashAlgorithm(*args.find('a')->second);
     if (hash_algorithm == shash::kAny) {
-      PrintError("unknown hash algorithm");
+      PrintError("Swissknife Sync: Unknown hash algorithm");
       return 1;
     }
   }
@@ -198,7 +201,8 @@ int swissknife::CommandUpload::Main(const swissknife::ArgumentList &args) {
   spooler->WaitForUpload();
 
   if (spooler->GetNumberOfErrors() > 0) {
-    LogCvmfs(kLogCatalog, kLogStderr, "failed to upload %s", source.c_str());
+    LogCvmfs(kLogCatalog, kLogStderr, "Swissknife Sync: failed to upload %s",
+                                      source.c_str());
     return 1;
   }
 
@@ -218,15 +222,17 @@ int swissknife::CommandPeek::Main(const swissknife::ArgumentList &args) {
   const bool success = spooler->Peek(file_to_peek);
 
   if (spooler->GetNumberOfErrors() > 0) {
-    LogCvmfs(kLogCatalog, kLogStderr, "failed to peek for %s",
+    LogCvmfs(kLogCatalog, kLogStderr, "Swissknife Sync: failed to peek for %s",
              file_to_peek.c_str());
     return 2;
   }
   if (!success) {
-    LogCvmfs(kLogCatalog, kLogStdout, "%s not found", file_to_peek.c_str());
+    LogCvmfs(kLogCatalog, kLogStdout, "Swissknife Sync: %s not found",
+                                      file_to_peek.c_str());
     return 1;
   }
-  LogCvmfs(kLogCatalog, kLogStdout, "%s available", file_to_peek.c_str());
+  LogCvmfs(kLogCatalog, kLogStdout, "Swissknife Sync: %s available",
+                                    file_to_peek.c_str());
 
   delete spooler;
 
@@ -245,7 +251,7 @@ int swissknife::CommandRemove::Main(const ArgumentList &args) {
   spooler->WaitForUpload();
 
   if (spooler->GetNumberOfErrors() > 0) {
-    LogCvmfs(kLogCatalog, kLogStderr, "failed to delete %s",
+    LogCvmfs(kLogCatalog, kLogStderr, "Swissknife Sync: failed to delete %s",
              file_to_delete.c_str());
     return 1;
   }
@@ -268,24 +274,28 @@ int swissknife::CommandApplyDirtab::Main(const ArgumentList &args) {
   // check if there is a dirtab file
   if (!FileExists(dirtab_file)) {
     LogCvmfs(kLogCatalog, kLogVerboseMsg,
-             "Didn't find a dirtab at '%s'. Skipping...", dirtab_file.c_str());
+                   "Swissknife Sync: Didn't find a dirtab at '%s'. Skipping...",
+                   dirtab_file.c_str());
     return 0;
   }
 
   // parse dirtab file
   catalog::Dirtab *dirtab = catalog::Dirtab::Create(dirtab_file);
   if (!dirtab->IsValid()) {
-    LogCvmfs(kLogCatalog, kLogStderr, "Invalid or not readable dirtab '%s'",
-             dirtab_file.c_str());
+    LogCvmfs(kLogCatalog, kLogStderr,
+                         "Swissknife Sync: Invalid or not readable dirtab '%s'",
+                         dirtab_file.c_str());
     return 1;
   }
-  LogCvmfs(kLogCatalog, kLogVerboseMsg, "Found %d rules in dirtab '%s'",
-           dirtab->RuleCount(), dirtab_file.c_str());
+  LogCvmfs(kLogCatalog, kLogVerboseMsg,
+                              "Swissknife Sync: Found %lu rules in dirtab '%s'",
+                              dirtab->RuleCount(), dirtab_file.c_str());
 
   // initialize catalog infrastructure
   const bool auto_manage_catalog_files = true;
   const bool follow_redirects = (args.count('L') > 0);
-  if (!InitDownloadManager(follow_redirects)) {
+  const string proxy = (args.count('@') > 0) ? *args.find('@')->second : "";
+  if (!InitDownloadManager(follow_redirects, proxy)) {
     return 1;
   }
   catalog::SimpleCatalogManager catalog_manager(
@@ -381,17 +391,20 @@ void swissknife::CommandApplyDirtab::DetermineNestedCatalogCandidates(
 
     if (glob_retval == 0) {
       // found some candidates... filtering by cvmfs catalog structure
-      LogCvmfs(kLogCatalog, kLogDebug, "Found %d entries for pathspec (%s)",
-               glob_res.gl_pathc, glob_string.c_str());
+      LogCvmfs(kLogCatalog, kLogDebug,
+                         "Swissknife Sync: Found %lu entries for pathspec (%s)",
+                         glob_res.gl_pathc, glob_string.c_str());
       FilterCandidatesFromGlobResult(dirtab, glob_res.gl_pathv,
                                      glob_res.gl_pathc, catalog_manager,
                                      nested_catalog_candidates);
     } else if (glob_retval == GLOB_NOMATCH) {
-      LogCvmfs(kLogCvmfs, kLogStderr, "WARNING: cannot apply pathspec %s",
-               glob_string.c_str());
+      LogCvmfs(kLogCvmfs, kLogStderr,
+                           "Swissknife Sync: WARNING: cannot apply pathspec %s",
+                           glob_string.c_str());
     } else {
-      LogCvmfs(kLogCvmfs, kLogStderr, "Failed to run glob matching (%s)",
-               glob_string.c_str());
+      LogCvmfs(kLogCvmfs, kLogStderr,
+                            "Swissknife Sync: Failed to run glob matching (%s)",
+                            glob_string.c_str());
     }
 
     globfree(&glob_res);
@@ -411,11 +424,18 @@ void swissknife::CommandApplyDirtab::FilterCandidatesFromGlobResult(
     // check if path points to a directory
     platform_stat64 candidate_info;
     const int lstat_retval = platform_lstat(candidate.c_str(), &candidate_info);
+    if (lstat_retval != 0) {
+      LogCvmfs(kLogCatalog, kLogDebug | kLogStderr | kLogSyslogErr,
+               "Swissknife Sync: "
+               "Error in processing .cvmfsdirtab: cannot access %s (%d)",
+               candidate.c_str(), errno);
+      abort();
+    }
     assert(lstat_retval == 0);
     if (!S_ISDIR(candidate_info.st_mode)) {
       // The GLOB_ONLYDIR flag is only a hint, non-directories can still be
       // returned
-      LogCvmfs(kLogCatalog, kLogDebug,
+      LogCvmfs(kLogCatalog, kLogDebug, "Swissknife Sync: "
                "The '%s' dirtab entry does not point to a directory "
                "but to a file or a symbolic link",
                candidate_rel.c_str());
@@ -431,8 +451,9 @@ void swissknife::CommandApplyDirtab::FilterCandidatesFromGlobResult(
 
     // check that the path isn't excluded in the dirtab
     if (dirtab.IsOpposing(candidate_rel)) {
-      LogCvmfs(kLogCatalog, kLogDebug, "Candidate '%s' is excluded by dirtab",
-               candidate_rel.c_str());
+      LogCvmfs(kLogCatalog, kLogDebug,
+                        "Swissknife Sync: Candidate '%s' is excluded by dirtab",
+                        candidate_rel.c_str());
       continue;
     }
 
@@ -441,17 +462,17 @@ void swissknife::CommandApplyDirtab::FilterCandidatesFromGlobResult(
     // a new directory and thus not in any catalog yet.
     catalog::DirectoryEntry dirent;
     const bool lookup_success = catalog_manager->LookupPath(
-        candidate_rel, catalog::kLookupSole, &dirent);
+        candidate_rel, catalog::kLookupDefault, &dirent);
     if (!lookup_success) {
       LogCvmfs(kLogCatalog, kLogDebug,
-               "Didn't find '%s' in catalogs, could "
+               "Swissknife Sync: Didn't find '%s' in catalogs, could "
                "be a new directory and nested catalog.",
                candidate_rel.c_str());
       nested_catalog_candidates->push_back(candidate);
     } else if (!dirent.IsNestedCatalogMountpoint() &&
                !dirent.IsNestedCatalogRoot()) {
       LogCvmfs(kLogCatalog, kLogDebug,
-               "Found '%s' in catalogs but is not a "
+               "Swissknife Sync: Found '%s' in catalogs but is not a "
                "nested catalog yet.",
                candidate_rel.c_str());
       nested_catalog_candidates->push_back(candidate);
@@ -462,17 +483,17 @@ void swissknife::CommandApplyDirtab::FilterCandidatesFromGlobResult(
       //       space to verify that it was touched (copy-on-write)
       //       Otherwise we would force the cvmfs client behind the union
       //       file-
-      //       system to (potentially) unncessarily fetch catalogs
+      //       system to (potentially) unnecessarily fetch catalogs
       if (DirectoryExists(scratch_dir_ + candidate_rel) &&
           !FileExists(union_dir_ + candidate_rel + "/.cvmfscatalog")) {
         LogCvmfs(kLogCatalog, kLogStdout,
-                 "WARNING: '%s' should be a nested "
+                 "Swissknife Sync: WARNING: '%s' should be a nested "
                  "catalog according to the dirtab. "
                  "Recreating...",
                  candidate_rel.c_str());
         nested_catalog_candidates->push_back(candidate);
       } else {
-        LogCvmfs(kLogCatalog, kLogDebug,
+        LogCvmfs(kLogCatalog, kLogDebug, "Swissknife Sync: "
                  "Found '%s' in catalogs and it already is a nested catalog.",
                  candidate_rel.c_str());
       }
@@ -502,7 +523,7 @@ bool swissknife::CommandApplyDirtab::CreateCatalogMarkers(
     const int fd = open(marker_path.c_str(), O_CREAT, mode);
     if (fd < 0) {
       LogCvmfs(kLogCvmfs, kLogStderr,
-               "Failed to create nested catalog marker "
+               "Swissknife Sync: Failed to create nested catalog marker "
                "at '%s' (errno: %d)",
                marker_path.c_str(), errno);
       success = false;
@@ -512,8 +533,9 @@ bool swissknife::CommandApplyDirtab::CreateCatalogMarkers(
 
     // inform the user if requested
     if (verbose_) {
-      LogCvmfs(kLogCvmfs, kLogStdout, "Auto-creating nested catalog in %s",
-               k->c_str());
+      LogCvmfs(kLogCvmfs, kLogStdout,
+                          "Swissknife Sync: Auto-creating nested catalog in %s",
+                          k->c_str());
     }
   }
 
@@ -544,10 +566,11 @@ bool swissknife::CommandSync::ReadFileChunkingArgs(
 
     if (arg != args.end()) {
       size_t arg_value = static_cast<size_t>(String2Uint64(*arg->second));
-      if (arg_value > 0)
+      if (arg_value > 0) {
         *i->save_to = arg_value;
-      else
+      } else {
         return false;
+      }
     }
   }
 
@@ -556,6 +579,20 @@ bool swissknife::CommandSync::ReadFileChunkingArgs(
 }
 
 int swissknife::CommandSync::Main(const swissknife::ArgumentList &args) {
+  string start_time = GetGMTimestamp();
+
+  // Spawn monitoring process (watchdog)
+  std::string watchdog_dir = "/tmp";
+  char watchdog_path[PATH_MAX];
+  std::string timestamp = GetGMTimestamp("%Y.%m.%d-%H.%M.%S");
+  int path_size = snprintf(watchdog_path, sizeof(watchdog_path),
+                           "%s/cvmfs-swissknife-sync-stacktrace.%s.%d",
+                           watchdog_dir.c_str(), timestamp.c_str(), getpid());
+  assert(path_size > 0);
+  assert(path_size < PATH_MAX);
+  UniquePtr<Watchdog> watchdog(Watchdog::Create(NULL));
+  watchdog->Spawn(std::string(watchdog_path));
+
   SyncParameters params;
 
   // Initialization
@@ -585,13 +622,16 @@ int swissknife::CommandSync::Main(const swissknife::ArgumentList &args) {
   if (args.find('V') != args.end()) params.voms_authz = true;
   if (args.find('F') != args.end()) params.authz_file = *args.find('F')->second;
   if (args.find('k') != args.end()) params.include_xattrs = true;
+  if (args.find('j') != args.end()) params.enable_mtime_ns = true;
   if (args.find('Y') != args.end()) params.external_data = true;
+  if (args.find('W') != args.end()) params.direct_io = true;
   if (args.find('S') != args.end()) {
     bool retval = catalog::VirtualCatalog::ParseActions(
         *args.find('S')->second, &params.virtual_dir_actions);
     if (!retval) {
-      LogCvmfs(kLogCvmfs, kLogStderr, "invalid virtual catalog options: %s",
-               args.find('S')->second->c_str());
+      LogCvmfs(kLogCvmfs, kLogStderr,
+                         "Swissknife Sync: Invalid virtual catalog options: %s",
+                          args.find('S')->second->c_str());
       return 1;
     }
   }
@@ -599,7 +639,7 @@ int swissknife::CommandSync::Main(const swissknife::ArgumentList &args) {
     unsigned log_level =
         1 << (kLogLevel0 + String2Uint64(*args.find('z')->second));
     if (log_level > kLogNone) {
-      LogCvmfs(kLogCvmfs, kLogStderr, "invalid log level");
+      LogCvmfs(kLogCvmfs, kLogStderr, "Swissknife Sync: invalid log level");
       return 1;
     }
     SetLogVerbosity(static_cast<LogLevels>(log_level));
@@ -613,7 +653,7 @@ int swissknife::CommandSync::Main(const swissknife::ArgumentList &args) {
   if (args.find('p') != args.end()) {
     params.use_file_chunking = true;
     if (!ReadFileChunkingArgs(args, &params)) {
-      PrintError("Failed to read file chunk size values");
+      PrintError("Swissknife Sync: Failed to read file chunk size values");
       return 2;
     }
   }
@@ -624,7 +664,7 @@ int swissknife::CommandSync::Main(const swissknife::ArgumentList &args) {
   if (args.find('e') != args.end()) {
     hash_algorithm = shash::ParseHashAlgorithm(*args.find('e')->second);
     if (hash_algorithm == shash::kAny) {
-      PrintError("unknown hash algorithm");
+      PrintError("Swissknife Sync: Unknown hash algorithm");
       return 1;
     }
   }
@@ -633,10 +673,6 @@ int swissknife::CommandSync::Main(const swissknife::ArgumentList &args) {
         zlib::ParseCompressionAlgorithm(*args.find('Z')->second);
   }
 
-  if (args.find('C') != args.end()) {
-    params.trusted_certs = *args.find('C')->second;
-  }
-
   if (args.find('E') != args.end()) params.enforce_limits = true;
   if (args.find('Q') != args.end()) {
     params.nested_kcatalog_limit = String2Uint64(*args.find('Q')->second);
@@ -657,7 +693,7 @@ int swissknife::CommandSync::Main(const swissknife::ArgumentList &args) {
   if (args.find('v') != args.end()) {
     sanitizer::IntegerSanitizer sanitizer;
     if (!sanitizer.IsValid(*args.find('v')->second)) {
-      PrintError("invalid revision number");
+      PrintError("Swissknife Sync: Invalid revision number");
       return 1;
     }
     params.manual_revision = String2Uint64(*args.find('v')->second);
@@ -690,22 +726,24 @@ int swissknife::CommandSync::Main(const swissknife::ArgumentList &args) {
   }
 
   if (args.find('D') != args.end()) {
-    params.repo_tag.name_ = *args.find('D')->second;
-  }
-
-  if (args.find('G') != args.end()) {
-    params.repo_tag.channel_ = *args.find('G')->second;
+    params.repo_tag.SetName(*args.find('D')->second);
   }
 
   if (args.find('J') != args.end()) {
-    params.repo_tag.description_ = *args.find('J')->second;
+    params.repo_tag.SetDescription(*args.find('J')->second);
   }
 
+  if (args.find('G') != args.end()) {
+    params.cache_dir = "/var/spool/cvmfs/" + params.repo_name + "/cache.server";
+  }
+
+  const bool upload_statsdb = (args.count('I') > 0);
+
   if (!CheckParams(params)) return 2;
   // This may fail, in which case a warning is printed and the process continues
   ObtainDacReadSearchCapability();
 
-  perf::StatisticsTemplate publish_statistics("Publish", this->statistics());
+  perf::StatisticsTemplate publish_statistics("publish", this->statistics());
 
   // Start spooler
   upload::SpoolerDefinition spooler_definition(
@@ -731,25 +769,20 @@ int swissknife::CommandSync::Main(const swissknife::ArgumentList &args) {
   if (!spooler_catalogs.IsValid()) return 3;
 
   const bool follow_redirects = (args.count('L') > 0);
-  if (!InitDownloadManager(follow_redirects)) {
+  const string proxy = (args.count('@') > 0) ? *args.find('@')->second : "";
+  if (!InitDownloadManager(follow_redirects, proxy)) {
     return 3;
   }
 
-  if (!InitVerifyingSignatureManager(params.public_keys,
-                                     params.trusted_certs)) {
+  if (!InitSignatureManager(params.public_keys)) {
     return 3;
   }
 
-  bool with_gateway =
-      spooler_definition.driver_type == upload::SpoolerDefinition::Gateway;
   /*
    * Note: If the upstream is of type gateway, due to the possibility of
-   * concurrent
-   *       release managers, it's possible to have a different local and remote
-   * root
-   *       hashes. We proceed by loading the remote manifest but we give an
-   * empty base
-   *       hash.
+   * concurrent release managers, it's possible to have a different local and
+   * remote root hashes. We proceed by loading the remote manifest but we give
+   * an empty base hash.
    */
   UniquePtr<manifest::Manifest> manifest;
   if (params.branched_catalog) {
@@ -760,36 +793,39 @@ int swissknife::CommandSync::Main(const swissknife::ArgumentList &args) {
     manifest = this->OpenLocalManifest(params.manifest_path);
     params.base_hash = manifest->catalog_hash();
   } else {
-    if (with_gateway) {
-      manifest =
-          FetchRemoteManifest(params.stratum0, params.repo_name, shash::Any());
-    } else {
-      manifest = FetchRemoteManifest(params.stratum0, params.repo_name,
-                                     params.base_hash);
-    }
+    // TODO(jblomer): revert to params.base_hash if spooler driver type is not
+    // upload::SpoolerDefinition::Gateway
+    manifest =
+      FetchRemoteManifest(params.stratum0, params.repo_name, shash::Any());
   }
-  if (!manifest) {
+  if (!manifest.IsValid()) {
     return 3;
   }
 
+  StatisticsDatabase *stats_db =
+    StatisticsDatabase::OpenStandardDB(params.repo_name);
+
   const std::string old_root_hash = manifest->catalog_hash().ToString(true);
 
   catalog::WritableCatalogManager catalog_manager(
-      params.base_hash, params.stratum0, params.dir_temp, spooler_catalogs,
-      download_manager(), params.enforce_limits, params.nested_kcatalog_limit,
-      params.root_kcatalog_limit, params.file_mbyte_limit, statistics(),
-      params.is_balanced, params.max_weight, params.min_weight);
+         params.base_hash, params.stratum0, params.dir_temp,
+         spooler_catalogs.weak_ref(), download_manager(), params.enforce_limits,
+         params.nested_kcatalog_limit, params.root_kcatalog_limit,
+         params.file_mbyte_limit, statistics(), params.is_balanced,
+         params.max_weight, params.min_weight, params.cache_dir);
   catalog_manager.Init();
 
   publish::SyncMediator mediator(&catalog_manager, &params, publish_statistics);
+  LogCvmfs(kLogPublish, kLogStdout, "Swissknife Sync: Processing changes...");
 
-  // Should be before the syncronization starts to avoid race of GetTTL with
+  // Should be before the synchronization starts to avoid race of GetTTL with
   // other sqlite operations
   if ((params.ttl_seconds > 0) &&
       ((params.ttl_seconds != catalog_manager.GetTTL()) ||
        !catalog_manager.HasExplicitTTL())) {
-    LogCvmfs(kLogCvmfs, kLogStdout, "Setting repository TTL to %" PRIu64 "s",
-             params.ttl_seconds);
+    LogCvmfs(kLogCvmfs, kLogStdout,
+                      "Swissknife Sync: Setting repository TTL to %" PRIu64 "s",
+                      params.ttl_seconds);
     catalog_manager.SetTTL(params.ttl_seconds);
   }
 
@@ -803,14 +839,15 @@ int swissknife::CommandSync::Main(const swissknife::ArgumentList &args) {
       sync = new publish::SyncUnionAufs(&mediator, params.dir_rdonly,
                                         params.dir_union, params.dir_scratch);
     } else {
-      LogCvmfs(kLogCvmfs, kLogStderr, "unknown union file system: %s",
-               params.union_fs_type.c_str());
+      LogCvmfs(kLogCvmfs, kLogStderr,
+                               "Swissknife Sync: unknown union file system: %s",
+                               params.union_fs_type.c_str());
       return 3;
     }
 
     if (!sync->Initialize()) {
       LogCvmfs(kLogCvmfs, kLogStderr,
-               "Initialization of the synchronisation "
+               "Swissknife Sync: Initialization of the synchronisation "
                "engine failed");
       return 4;
     }
@@ -825,13 +862,13 @@ int swissknife::CommandSync::Main(const swissknife::ArgumentList &args) {
 
   if (!params.authz_file.empty()) {
     LogCvmfs(kLogCvmfs, kLogDebug,
-             "Adding contents of authz file %s to"
+             "Swissknife Sync: Adding contents of authz file %s to"
              " root catalog.",
              params.authz_file.c_str());
     int fd = open(params.authz_file.c_str(), O_RDONLY);
     if (fd == -1) {
       LogCvmfs(kLogCvmfs, kLogStderr,
-               "Unable to open authz file (%s)"
+               "Swissknife Sync: Unable to open authz file (%s)"
                "from the publication process: %s",
                params.authz_file.c_str(), strerror(errno));
       return 7;
@@ -842,8 +879,9 @@ int swissknife::CommandSync::Main(const swissknife::ArgumentList &args) {
     close(fd);
 
     if (!read_successful) {
-      LogCvmfs(kLogCvmfs, kLogStderr, "Failed to read authz file (%s): %s",
-               params.authz_file.c_str(), strerror(errno));
+      LogCvmfs(kLogCvmfs, kLogStderr,
+                          "Swissknife Sync: Failed to read authz file (%s): %s",
+                          params.authz_file.c_str(), strerror(errno));
       return 8;
     }
 
@@ -851,21 +889,30 @@ int swissknife::CommandSync::Main(const swissknife::ArgumentList &args) {
   }
 
   if (!mediator.Commit(manifest.weak_ref())) {
-    PrintError("something went wrong during sync");
+    PrintError("Swissknife Sync: Something went wrong during sync");
+    if (!params.dry_run) {
+      stats_db->StorePublishStatistics(this->statistics(), start_time, false);
+      if (upload_statsdb) {
+        stats_db->UploadStatistics(params.spooler);
+      }
+    }
     return 5;
   }
 
-  perf::Counter *revision_counter = statistics()->Register("Publish.revision",
+  perf::Counter *revision_counter = statistics()->Register("publish.revision",
                                                   "Published revision number");
-  revision_counter->Set(catalog_manager.GetRootCatalog()->revision());
+  revision_counter->Set(static_cast<int64_t>(
+    catalog_manager.GetRootCatalog()->revision()));
 
   // finalize the spooler
-  LogCvmfs(kLogCvmfs, kLogStdout, "Wait for all uploads to finish");
+  LogCvmfs(kLogCvmfs, kLogStdout,
+                             "Swissknife Sync: Wait for all uploads to finish");
   params.spooler->WaitForUpload();
   spooler_catalogs->WaitForUpload();
   params.spooler->FinalizeSession(false);
 
-  LogCvmfs(kLogCvmfs, kLogStdout, "Exporting repository manifest");
+  LogCvmfs(kLogCvmfs, kLogStdout,
+                              "Swissknife Sync: Exporting repository manifest");
 
   // We call FinalizeSession(true) this time, to also trigger the commit
   // operation on the gateway machine (if the upstream is of type "gw").
@@ -875,13 +922,27 @@ int swissknife::CommandSync::Main(const swissknife::ArgumentList &args) {
 
   if (!spooler_catalogs->FinalizeSession(true, old_root_hash, new_root_hash,
                                          params.repo_tag)) {
-    PrintError("Failed to commit transaction.");
+    PrintError("Swissknife Sync: Failed to commit transaction.");
+    if (!params.dry_run) {
+      stats_db->StorePublishStatistics(this->statistics(), start_time, false);
+      if (upload_statsdb) {
+        stats_db->UploadStatistics(params.spooler);
+      }
+    }
     return 9;
-                                         }
+  }
+
+  if (!params.dry_run) {
+    stats_db->StorePublishStatistics(this->statistics(), start_time, true);
+    if (upload_statsdb) {
+      stats_db->UploadStatistics(params.spooler);
+    }
+  }
+
   delete params.spooler;
 
   if (!manifest->Export(params.manifest_path)) {
-    PrintError("Failed to create new repository");
+    PrintError("Swissknife Sync: Failed to create new repository");
     return 6;
   }
 
