@@ -12,11 +12,17 @@
  * The talk module runs in a separate thread.
  */
 
-#include "cvmfs_config.h"
+#ifndef __STDC_FORMAT_MACROS
+#define __STDC_FORMAT_MACROS
+#endif
+
+
 #include "talk.h"
 
 #include <errno.h>
+#include <inttypes.h>
 #include <pthread.h>
+#include <stdint.h>
 #include <sys/socket.h>
 #include <sys/stat.h>
 #include <sys/types.h>
@@ -34,21 +40,22 @@
 #include "cache_posix.h"
 #include "catalog_mgr_client.h"
 #include "cvmfs.h"
-#include "download.h"
 #include "duplex_sqlite3.h"
 #include "fuse_remount.h"
 #include "glue_buffer.h"
 #include "loader.h"
-#include "logging.h"
+#include "lru_md.h"
 #include "monitor.h"
 #include "mountpoint.h"
+#include "network/download.h"
 #include "nfs_maps.h"
 #include "options.h"
-#include "platform.h"
 #include "quota.h"
 #include "shortstring.h"
 #include "statistics.h"
 #include "tracer.h"
+#include "util/logging.h"
+#include "util/platform.h"
 #include "util/pointer.h"
 #include "wpad.h"
 
@@ -90,12 +97,33 @@ TalkManager *TalkManager::Create(
 }
 
 
+string TalkManager::FormatMetalinkInfo(download::DownloadManager *download_mgr)
+{
+  vector<string> metalink_chain;
+  unsigned active_metalink;
+
+  download_mgr->GetMetalinkInfo(&metalink_chain, &active_metalink);
+  if (metalink_chain.size() == 0)
+    return "No metalinks defined\n";
+
+  string metalink_str;
+  for (unsigned i = 0; i < metalink_chain.size(); ++i) {
+    metalink_str += "  [" + StringifyInt(i) + "] " + metalink_chain[i] + "\n";
+  }
+  metalink_str += "Active metalink " + StringifyInt(active_metalink) + ": " +
+              metalink_chain[active_metalink] + "\n";
+  return metalink_str;
+}
+
 string TalkManager::FormatHostInfo(download::DownloadManager *download_mgr) {
   vector<string> host_chain;
   vector<int> rtt;
   unsigned active_host;
 
   download_mgr->GetHostInfo(&host_chain, &rtt, &active_host);
+  if (host_chain.size() == 0)
+    return "No hosts defined\n";
+
   string host_str;
   for (unsigned i = 0; i < host_chain.size(); ++i) {
     host_str += "  [" + StringifyInt(i) + "] " + host_chain[i] + " (";
@@ -181,7 +209,7 @@ void *TalkManager::MainResponder(void *data) {
     if (buf[bytes_read-1] == '\0')
       bytes_read--;
     const string line = string(buf, bytes_read);
-    LogCvmfs(kLogTalk, kLogDebug, "received %s (length %u)",
+    LogCvmfs(kLogTalk, kLogDebug, "received %s (length %lu)",
              line.c_str(), line.length());
 
     if (line == "tracebuffer flush") {
@@ -240,6 +268,27 @@ void *TalkManager::MainResponder(void *data) {
           talk_mgr->Answer(con_fd, StringifyInt(rate) + "\n");
         }
       }
+    } else if (line.substr(0, 15) == "cache limit set") {
+      if (line.length() < 16) {
+        talk_mgr->Answer(con_fd, "Usage: cache limit set <MB>\n");
+      } else {
+        QuotaManager *quota_mgr = file_system->cache_mgr()->quota_mgr();
+        const uint64_t size = String2Uint64(line.substr(16));
+        if (size < 1000) {
+            talk_mgr->Answer(con_fd, "New limit too low (minimum 1000)\n");
+        } else {
+          if(quota_mgr->SetLimit(size * 1024*1024)) {
+              file_system->options_mgr()->SetValueFromTalk("CVMFS_QUOTA_LIMIT", StringifyUint(size));
+              talk_mgr->Answer(con_fd, "OK\n");
+          } else {
+              talk_mgr->Answer(con_fd, "Limit not reset\n");
+          }
+        }
+      }
+    } else if (line == "cache limit get") {
+      std::string limit_from_options;
+      file_system->options_mgr()->GetValue("CVMFS_QUOTA_LIMIT", &limit_from_options);
+            talk_mgr->Answer(con_fd, limit_from_options + "\n");
     } else if (line.substr(0, 7) == "cleanup") {
       QuotaManager *quota_mgr = file_system->cache_mgr()->quota_mgr();
       if (!quota_mgr->HasCapability(QuotaManager::kCapShrink)) {
@@ -283,6 +332,23 @@ void *TalkManager::MainResponder(void *data) {
       }
     } else if (line == "mountpoint") {
       talk_mgr->Answer(con_fd, cvmfs::loader_exports_->mount_point + "\n");
+    } else if (line == "device id") {
+      if (cvmfs::loader_exports_->version >= 5)
+        talk_mgr->Answer(con_fd, cvmfs::loader_exports_->device_id + "\n");
+      else
+        talk_mgr->Answer(con_fd, "0:0\n");
+    } else if (line.substr(0, 13) == "send mount fd") {
+      // Hidden command intended to be used only by the cvmfs mount helper
+      if (line.length() < 15) {
+        talk_mgr->Answer(con_fd, "EINVAL\n");
+      } else {
+        std::string socket_path = line.substr(14);
+        bool retval = cvmfs::SendFuseFd(socket_path);
+        talk_mgr->Answer(con_fd, retval ? "OK\n" : "Failed\n");
+        LogCvmfs(kLogCvmfs, kLogDebug | kLogSyslog,
+                 "Attempt to send fuse connection info to new mount (via %s)%s",
+                 socket_path.c_str(), retval ? "" : " -- failed!");
+      }
     } else if (line.substr(0, 7) == "remount") {
       FuseRemounter::Status status;
       if (line == "remount sync")
@@ -308,6 +374,22 @@ void *TalkManager::MainResponder(void *data) {
         default:
           talk_mgr->Answer(con_fd, "internal error\n");
       }
+    } else if (line.substr(0, 6) == "chroot") {
+      if (line.length() < 8) {
+        talk_mgr->Answer(con_fd, "Usage: chroot <hash>\n");
+      } else {
+        std::string root_hash = Trim(line.substr(7), true /* trim_newline */);
+        FuseRemounter::Status status = remounter->ChangeRoot(
+          MkFromHexPtr(shash::HexPtr(root_hash), shash::kSuffixCatalog));
+        switch (status) {
+          case FuseRemounter::kStatusUp2Date:
+            talk_mgr->Answer(con_fd, "OK\n");
+            break;
+          default:
+            talk_mgr->Answer(con_fd, "Failed\n");
+            break;
+        }
+      }
     } else if (line == "detach nested catalogs") {
       mount_point->catalog_mgr()->DetachNested();
       talk_mgr->Answer(con_fd, "OK\n");
@@ -344,12 +426,21 @@ void *TalkManager::MainResponder(void *data) {
         mount_point->download_mgr()->SetDnsServer(host);
         talk_mgr->Answer(con_fd, "OK\n");
       }
+    } else if (line == "external metalink info") {
+      const string external_metalink_info =
+        talk_mgr->FormatMetalinkInfo(mount_point->external_download_mgr());
+      talk_mgr->Answer(con_fd, external_metalink_info);
+    } else if (line == "metalink info") {
+      const string metalink_info =
+        talk_mgr->FormatMetalinkInfo(mount_point->download_mgr());
+      talk_mgr->Answer(con_fd, metalink_info);
     } else if (line == "external host info") {
-      string external_host_info =
+      const string external_host_info =
         talk_mgr->FormatHostInfo(mount_point->external_download_mgr());
       talk_mgr->Answer(con_fd, external_host_info);
     } else if (line == "host info") {
-      string host_info = talk_mgr->FormatHostInfo(mount_point->download_mgr());
+      const string host_info =
+        talk_mgr->FormatHostInfo(mount_point->download_mgr());
       talk_mgr->Answer(con_fd, host_info);
     } else if (line == "host probe") {
       mount_point->download_mgr()->ProbeHosts();
@@ -360,12 +451,34 @@ void *TalkManager::MainResponder(void *data) {
         talk_mgr->Answer(con_fd, "OK\n");
       else
         talk_mgr->Answer(con_fd, "Failed\n");
+    } else if (line == "external metalink switch") {
+      mount_point->external_download_mgr()->SwitchMetalink();
+      talk_mgr->Answer(con_fd, "OK\n");
+    } else if (line == "metalink switch") {
+      mount_point->download_mgr()->SwitchMetalink();
+      talk_mgr->Answer(con_fd, "OK\n");
     } else if (line == "external host switch") {
       mount_point->external_download_mgr()->SwitchHost();
       talk_mgr->Answer(con_fd, "OK\n");
     } else if (line == "host switch") {
       mount_point->download_mgr()->SwitchHost();
       talk_mgr->Answer(con_fd, "OK\n");
+    } else if (line.substr(0, 21) == "external metalink set") {
+      if (line.length() < 23) {
+        talk_mgr->Answer(con_fd, "Usage: external metalink set <URL>\n");
+      } else {
+        const std::string host = line.substr(22);
+        mount_point->external_download_mgr()->SetMetalinkChain(host);
+        talk_mgr->Answer(con_fd, "OK\n");
+      }
+    } else if (line.substr(0, 12) == "metalink set") {
+      if (line.length() < 14) {
+        talk_mgr->Answer(con_fd, "Usage: metalink set <URL>\n");
+      } else {
+        const std::string host = line.substr(13);
+        mount_point->download_mgr()->SetMetalinkChain(host);
+        talk_mgr->Answer(con_fd, "OK\n");
+      }
     } else if (line.substr(0, 17) == "external host set") {
       if (line.length() < 19) {
         talk_mgr->Answer(con_fd, "Usage: external host set <URL>\n");
@@ -457,6 +570,18 @@ void *TalkManager::MainResponder(void *data) {
       }
     } else if (line == "open catalogs") {
       talk_mgr->Answer(con_fd, mount_point->catalog_mgr()->PrintHierarchy());
+    } else if (line == "drop metadata caches") {
+      // For testing
+      mount_point->inode_cache()->Pause();
+      mount_point->path_cache()->Pause();
+      mount_point->md5path_cache()->Pause();
+      mount_point->inode_cache()->Drop();
+      mount_point->path_cache()->Drop();
+      mount_point->md5path_cache()->Drop();
+      mount_point->inode_cache()->Resume();
+      mount_point->path_cache()->Resume();
+      mount_point->md5path_cache()->Resume();
+      talk_mgr->Answer(con_fd, "OK\n");
     } else if (line == "internal affairs") {
       int current;
       int highwater;
@@ -481,8 +606,10 @@ void *TalkManager::MainResponder(void *data) {
       // Manually setting the inode tracker numbers
       glue::InodeTracker::Statistics inode_stats =
         mount_point->inode_tracker()->GetStatistics();
-      glue::NentryTracker::Statistics nentry_stats =
-        mount_point->nentry_tracker()->GetStatistics();
+      glue::DentryTracker::Statistics dentry_stats =
+        mount_point->dentry_tracker()->GetStatistics();
+      glue::PageCacheTracker::Statistics page_cache_stats =
+        mount_point->page_cache_tracker()->GetStatistics();
       mount_point->statistics()->Lookup("inode_tracker.n_insert")->Set(
         atomic_read64(&inode_stats.num_inserts));
       mount_point->statistics()->Lookup("inode_tracker.n_remove")->Set(
@@ -495,12 +622,22 @@ void *TalkManager::MainResponder(void *data) {
         atomic_read64(&inode_stats.num_hits_path));
       mount_point->statistics()->Lookup("inode_tracker.n_miss_path")->Set(
         atomic_read64(&inode_stats.num_misses_path));
-      mount_point->statistics()->Lookup("nentry_tracker.n_insert")->Set(
-        nentry_stats.num_insert);
-      mount_point->statistics()->Lookup("nentry_tracker.n_remove")->Set(
-        nentry_stats.num_remove);
-      mount_point->statistics()->Lookup("nentry_tracker.n_prune")->Set(
-        nentry_stats.num_prune);
+      mount_point->statistics()->Lookup("dentry_tracker.n_insert")->Set(
+        dentry_stats.num_insert);
+      mount_point->statistics()->Lookup("dentry_tracker.n_remove")->Set(
+        dentry_stats.num_remove);
+      mount_point->statistics()->Lookup("dentry_tracker.n_prune")->Set(
+        dentry_stats.num_prune);
+      mount_point->statistics()->Lookup("page_cache_tracker.n_insert")->Set(
+        page_cache_stats.n_insert);
+      mount_point->statistics()->Lookup("page_cache_tracker.n_remove")->Set(
+        page_cache_stats.n_remove);
+      mount_point->statistics()->Lookup("page_cache_tracker.n_open_direct")->
+        Set(page_cache_stats.n_open_direct);
+      mount_point->statistics()->Lookup("page_cache_tracker.n_open_flush")->
+        Set(page_cache_stats.n_open_flush);
+      mount_point->statistics()->Lookup("page_cache_tracker.n_open_cached")->
+        Set(page_cache_stats.n_open_cached);
 
       if (file_system->cache_mgr()->id() == kPosixCacheManager) {
         PosixCacheManager *cache_mgr =
@@ -572,6 +709,8 @@ void *TalkManager::MainResponder(void *data) {
 
       result += "Lookup\n" + file_system->hist_fs_lookup()->ToString();
       result += "Forget\n" + file_system->hist_fs_forget()->ToString();
+      result += "Multi-Forget\n"
+                + file_system->hist_fs_forget_multi()->ToString();
       result += "Getattr\n" + file_system->hist_fs_getattr()->ToString();
       result += "Readlink\n" + file_system->hist_fs_readlink()->ToString();
       result += "Opendir\n" + file_system->hist_fs_opendir()->ToString();
@@ -639,7 +778,7 @@ void *TalkManager::MainResponder(void *data) {
       mount_point->inode_tracker()->EndEnumerate(&cursor);
       talk_mgr->Answer(con_fd, result);
     } else if (line == "version") {
-      string version_str = string(VERSION) + " (CernVM-FS Fuse Module)\n" +
+      const string version_str = string(CVMFS_VERSION) + " (CernVM-FS Fuse Module)\n" +
         cvmfs::loader_exports_->loader_version + " (Loader)\n";
       talk_mgr->Answer(con_fd, version_str);
     } else if (line == "version patchlevel") {
@@ -662,7 +801,7 @@ void *TalkManager::MainResponder(void *data) {
   }
 
   return NULL;
-}
+}  // NOLINT(readability/fn_size)
 
 string TalkManager::FormatLatencies(const MountPoint &mount_point,
                                     FileSystem *file_system) {
@@ -709,6 +848,8 @@ string TalkManager::FormatLatencies(const MountPoint &mount_point,
   names.push_back("lookup");
   hist.push_back(file_system->hist_fs_forget());
   names.push_back("forget");
+  hist.push_back(file_system->hist_fs_forget_multi());
+  names.push_back("forget_multi");
   hist.push_back(file_system->hist_fs_getattr());
   names.push_back("getattr");
   hist.push_back(file_system->hist_fs_readlink());
@@ -729,11 +870,11 @@ string TalkManager::FormatLatencies(const MountPoint &mount_point,
   for (unsigned int j = 0; j < hist.size(); j++) {
     Log2Histogram *h = hist[j];
     unsigned int format_index =
-        snprintf(buffer, bufSize, "\"%s\",\"%s\",%ld,\"%s\"", repo.c_str(),
-                 names[j].c_str(), h->N(), "nanoseconds");
+      snprintf(buffer, bufSize, "\"%s\",\"%s\",%" PRIu64 ",\"nanoseconds\"",
+               repo.c_str(), names[j].c_str(), h->N());
     for (unsigned int i = 0; i < qs.size(); i++) {
       format_index += snprintf(buffer + format_index, bufSize - format_index,
-                               ",%d", h->GetQuantile(qs[i]));
+                               ",%u", h->GetQuantile(qs[i]));
     }
     format_index +=
         snprintf(buffer + format_index, bufSize - format_index, "\n");
@@ -763,7 +904,7 @@ TalkManager::TalkManager(
 TalkManager::~TalkManager() {
   if (!socket_path_.empty()) {
     int retval = unlink(socket_path_.c_str());
-    if (retval != 0) {
+    if ((retval != 0) && (errno != ENOENT)) {
       LogCvmfs(kLogTalk, kLogSyslogWarn,
                "Could not remove cvmfs_io socket from cache directory (%d)",
                errno);
