@@ -5,11 +5,12 @@
  */
 
 #ifndef __STDC_FORMAT_MACROS
+// NOLINTNEXTLINE
 #define __STDC_FORMAT_MACROS
 #endif
 
 #include "string.h"
-#include "cvmfs_config.h"
+
 
 #include <errno.h>
 #include <fcntl.h>
@@ -39,7 +40,7 @@ const char b64_table[] = {'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K',
 /**
  * Decode Base64 and Base64Url
  */
-const signed char db64_table[] = {
+const int8_t db64_table[] = {
     -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
     -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
     -1, -1, -1, -1, -1, 62, -1, 62, -1, 63, 52, 53, 54, 55, 56, 57, 58, 59, 60,
@@ -119,6 +120,24 @@ string StringifyTime(const time_t seconds, const bool utc) {
   return string(buffer);
 }
 
+/**
+ * Converts seconds since UTC 0 into something like 12 Sep 14:59:37 CDT
+ */
+string StringifyLocalTime(const time_t seconds) {
+  struct tm timestamp;
+  localtime_r(&seconds, &timestamp);
+
+  const char *months[] = {"Jan", "Feb", "Mar", "Apr", "May", "Jun",
+                          "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"};
+  char buffer[26];
+  (void)/* cast to void ignores return and placates clang-tidy */
+   snprintf(buffer, sizeof(buffer), "%d %s %d %02d:%02d:%02d %s", timestamp.tm_mday,
+           months[timestamp.tm_mon], timestamp.tm_year + 1900,
+           timestamp.tm_hour, timestamp.tm_min, timestamp.tm_sec, timestamp.tm_zone);
+
+  return string(buffer);
+}
+
 
 /**
  * Current time in format Wed, 01 Mar 2006 12:00:00 GMT
@@ -191,7 +210,7 @@ string StringifyTimeval(const timeval value) {
 }
 
 /**
- * Parses a timstamp of the form YYYY-MM-DDTHH:MM:SSZ
+ * Parses a timestamp of the form YYYY-MM-DDTHH:MM:SSZ
  * Return 0 on error
  */
 time_t IsoTimestamp2UtcTime(const std::string &iso8601) {
@@ -206,12 +225,12 @@ time_t IsoTimestamp2UtcTime(const std::string &iso8601) {
 
   struct tm tm_wl;
   memset(&tm_wl, 0, sizeof(struct tm));
-  tm_wl.tm_year = String2Int64(iso8601.substr(0, 4)) - 1900;
-  tm_wl.tm_mon = String2Int64(iso8601.substr(5, 2)) - 1;
-  tm_wl.tm_mday = String2Int64(iso8601.substr(8, 2));
-  tm_wl.tm_hour = String2Int64(iso8601.substr(11, 2));
-  tm_wl.tm_min = String2Int64(iso8601.substr(14, 2));
-  tm_wl.tm_sec = String2Int64(iso8601.substr(17, 2));
+  tm_wl.tm_year = static_cast<int>(String2Int64(iso8601.substr(0, 4))) - 1900;
+  tm_wl.tm_mon = static_cast<int>(String2Int64(iso8601.substr(5, 2))) - 1;
+  tm_wl.tm_mday = static_cast<int>(String2Int64(iso8601.substr(8, 2)));
+  tm_wl.tm_hour = static_cast<int>(String2Int64(iso8601.substr(11, 2)));
+  tm_wl.tm_min = static_cast<int>(String2Int64(iso8601.substr(14, 2)));
+  tm_wl.tm_sec = static_cast<int>(String2Int64(iso8601.substr(17, 2)));
   utc_time = timegm(&tm_wl);
   if (utc_time < 0) return 0;
 
@@ -226,8 +245,10 @@ int64_t String2Int64(const string &value) {
 
 uint64_t String2Uint64(const string &value) {
   uint64_t result;
-  sscanf(value.c_str(), "%" PRIu64, &result);
-  return result;
+  if (sscanf(value.c_str(), "%" PRIu64, &result) == 1) {
+    return result;
+  }
+  return 0;
 }
 
 /**
@@ -284,8 +305,13 @@ bool HasSuffix(const std::string &str, const std::string &suffix,
              : std::equal(suffix.rbegin(), suffix.rend(), str.rbegin());
 }
 
-vector<string> SplitString(const string &str, const char delim,
-                           const unsigned max_chunks) {
+vector<string> SplitString(const string &str, char delim) {
+  return SplitStringBounded(0, str, delim);
+}
+
+vector<string> SplitStringBounded(
+  unsigned max_chunks, const string &str, char delim)
+{
   vector<string> result;
 
   // edge case... one chunk is always the whole string
@@ -360,7 +386,7 @@ bool ParseKeyvalPath(const string &filename, map<char, string> *content) {
   if (fd < 0) return false;
 
   unsigned char buffer[4096];
-  int num_bytes = read(fd, buffer, sizeof(buffer));
+  ssize_t num_bytes = read(fd, buffer, sizeof(buffer));
   close(fd);
 
   if ((num_bytes <= 0) || (unsigned(num_bytes) >= sizeof(buffer))) return false;
@@ -386,7 +412,7 @@ bool GetLineFile(FILE *f, std::string *line) {
     } else if (retval == EOF) {
       break;
     }
-    char c = retval;
+    char c = static_cast<char>(retval);
     if (c == '\n') break;
     line->push_back(c);
   }
@@ -394,7 +420,7 @@ bool GetLineFile(FILE *f, std::string *line) {
 }
 
 bool GetLineFd(const int fd, std::string *line) {
-  int retval;
+  ssize_t retval;
   char c;
   line->clear();
   while (true) {
@@ -439,13 +465,36 @@ string Trim(const string &raw, bool trim_newline) {
   return raw.substr(start_pos, end_pos - start_pos + 1);
 }
 
+std::string TrimString(
+  const std::string& path,
+  const std::string& toTrim,
+  const int trimMode)
+{
+  std::string trimmed = path;
+  if (trimmed != toTrim) {
+    while ((trimMode & kTrimLeading) &&
+           HasPrefix(trimmed, toTrim, true) &&
+           (trimmed.size() > toTrim.size()))
+    {
+      trimmed = trimmed.substr(toTrim.size());
+    }
+    while ((trimMode & kTrimTrailing) &&
+           HasSuffix(trimmed, toTrim, true) &&
+           (trimmed.size() > toTrim.size()))
+    {
+      trimmed = trimmed.substr(0, trimmed.size() - toTrim.size());
+    }
+  }
+  return trimmed;
+}
+
 /**
  * Converts all characters to upper case
  */
 string ToUpper(const string &mixed_case) {
   string result(mixed_case);
   for (unsigned i = 0, l = result.length(); i < l; ++i) {
-    result[i] = toupper(result[i]);
+    result[i] = static_cast<char>(toupper(result[i]));
   }
   return result;
 }
@@ -504,16 +553,17 @@ string Base64(const string &data) {
 string Base64Url(const string &data) {
   string base64 = Base64(data);
   for (unsigned i = 0, l = base64.length(); i < l; ++i) {
-    if (base64[i] == '+')
+    if (base64[i] == '+') {
       base64[i] = '-';
-    else if (base64[i] == '/')
+    } else if (base64[i] == '/') {
       base64[i] = '_';
+    }
   }
   return base64;
 }
 
-static bool Debase64Block(const unsigned char input[4],
-                          const signed char *d_table, unsigned char output[3]) {
+static bool Debase64Block(const unsigned char input[4], unsigned char output[3])
+{
   int32_t dec[4];
   for (int i = 0; i < 4; ++i) {
     dec[i] = db64_table[input[i]];
@@ -541,7 +591,7 @@ bool Debase64(const string &data, string *decoded) {
 
   while (pos < length) {
     unsigned char decoded_block[3];
-    bool retval = Debase64Block(data_ptr + pos, db64_table, decoded_block);
+    bool retval = Debase64Block(data_ptr + pos, decoded_block);
     if (!retval) return false;
     decoded->append(reinterpret_cast<char *>(decoded_block), 3);
     pos += 4;
@@ -560,7 +610,7 @@ bool Debase64(const string &data, string *decoded) {
 string Tail(const string &source, unsigned num_lines) {
   if (source.empty() || (num_lines == 0)) return "";
 
-  unsigned l = source.length();
+  int l = static_cast<int>(source.length());
   int i = l - 1;
   for (; i >= 0; --i) {
     char c = source.data()[i];
@@ -580,7 +630,7 @@ string Tail(const string &source, unsigned num_lines) {
   * @param format format if timestamp (YYYY-MM-DD HH:MM:SS by default)
   * @return a timestamp string on success, empty string on failure
   */
-std::string GetGMTimestamp(std::string format) {
+std::string GetGMTimestamp(const std::string &format) {
   struct tm time_ptr;
   char date_and_time[100];
   time_t t = time(NULL);
