@@ -1,15 +1,21 @@
 #!/bin/sh
 
+
+
 # source the common platform independent functionality and option parsing
 script_location=$(dirname $(readlink --canonicalize $0))
 . ${script_location}/common_setup.sh
 
+# use archive, see https://linux.web.cern.ch/els7/
+sudo sed -i 's#http://linuxsoft.cern.ch/epel/7/#http://linuxsoft.cern.ch/internal/archive/epel/7/#g' /etc/yum.repos.d/epel.repo || die "fail (patching epel mirror)"
+
 echo "enabling epel yum repository..."
 install_from_repo epel-release
 
 # install CernVM-FS RPM packages
 echo "installing RPM packages... "
 install_rpm "$CONFIG_PACKAGES"
+install_rpm $LIBS_PACKAGE
 install_rpm $CLIENT_PACKAGE
 install_rpm $UNITTEST_PACKAGE
 install_rpm $FUSE3_PACKAGE
@@ -29,10 +35,15 @@ echo "installing additional RPM packages..."
 install_from_repo gcc
 install_from_repo gcc-c++
 install_from_repo wget
+install_from_repo git
+install_from_repo perl
 
 # traffic shaping
 install_from_repo trickle
 
+# For 006 kernel compile
+install_from_repo openssl-devel
+
 # Migration test needs lsb_release
 echo "install lsb_release..."
 install_from_repo redhat-lsb-core
