@@ -10,9 +10,9 @@
 #include <sstream>
 
 #include "catalog_rw.h"
-#include "compression.h"
+#include "compression/compression.h"
+#include "crypto/hash.h"
 #include "directory_entry.h"
-#include "hash.h"
 #include "options.h"
 #include "testutil.h"
 #include "util/posix.h"
@@ -66,9 +66,7 @@ DirSpec::DirSpec() : items_(), dirs_() {
 bool DirSpec::AddFile(const std::string& name, const std::string& parent,
                       const std::string& digest, const size_t size,
                       const XattrList& xattrs, shash::Suffix suffix) {
-  shash::Any hash = shash::Any(
-      shash::kSha1, reinterpret_cast<const unsigned char*>(digest.c_str()),
-      suffix);
+  shash::Any hash = shash::MkFromHexPtr(shash::HexPtr(digest), suffix);
   if (!HasDir(parent)) {
     return false;
   }
@@ -109,7 +107,9 @@ bool DirSpec::AddDirectory(const std::string& name, const std::string& parent,
 
 bool DirSpec::AddNestedCatalog(const std::string& name) {
   bool ret = AddNC(name);
+  if (!ret) return ret;
   nested_catalogs_.push_back(name);
+  AddFile(".cvmfscatalog", name, "0000000000000000000000000000000000000001", 0);
   return ret;
 }
 
@@ -135,17 +135,22 @@ void DirSpec::ToString(std::string* out) {
   for (DirSpec::ItemList::const_iterator it = items_.begin();
        it != items_.end(); ++it) {
     const DirSpecItem& item = it->second;
-    char item_type = ' ';
     if (item.entry_base().IsRegular()) {
-      item_type = 'F';
+      ostr << "F ";
     } else if (item.entry_base().IsDirectory()) {
-      item_type = 'D';
+      ostr << "D ";
+    } else if (item.entry_base().IsLink()) {
+      ostr << "S ";
     }
     std::string parent = item.parent();
     AddLeadingSlash(&parent);
 
-    ostr << item_type << " " << item.entry_base().GetFullPath(parent).c_str()
-         << std::endl;
+    ostr << item.entry_base().GetFullPath(parent).c_str();
+    if (item.entry_base().IsLink()) {
+      ostr << " -> " << item.entry_base().symlink().c_str();
+    }
+
+    ostr << std::endl;
   }
   *out = ostr.str();
 }
@@ -155,6 +160,11 @@ const DirSpecItem* DirSpec::Item(const std::string& full_path) const {
   if (it != items_.end()) {
     return &it->second;
   }
+  std::string no_slash(full_path);
+  RemoveLeadingSlash(&no_slash);
+  if (no_slash != full_path) {
+    return Item(full_path);
+  }
   return NULL;
 }
 
@@ -167,6 +177,7 @@ static void RemoveItemHelper(
   if (it != spec.items().end()) {
     const DirSpecItem item = it->second;
     acc->push_back(full_path);
+
     if (item.entry_base().IsDirectory()) {
       std::string rel_full_path(full_path);
       RemoveLeadingSlash(&rel_full_path);
@@ -183,8 +194,10 @@ static void RemoveItemHelper(
 }
 
 void DirSpec::RemoveItemRec(const std::string& full_path) {
+  std::string path(full_path);
+  RemoveLeadingSlash(&path);
   std::vector<std::string> acc(0);
-  RemoveItemHelper(*this, full_path, &acc);
+  RemoveItemHelper(*this, path, &acc);
 
   for (size_t i = 0u; i < acc.size(); ++i) {
     const DirSpecItem* item = Item(acc[i]);
@@ -192,6 +205,15 @@ void DirSpec::RemoveItemRec(const std::string& full_path) {
       RmDir(std::string(item->entry_base().name().c_str()), item->parent());
     }
     items_.erase(acc[i]);
+
+    DirSpec::NestedCatalogList::iterator n;
+    for (n = nested_catalogs_.begin(); n != nested_catalogs_.end();) {
+      if (*n == acc[i]) {
+        n = nested_catalogs_.erase(n);
+      } else {
+        ++n;
+      }
+    }
   }
 }
 
@@ -240,7 +262,7 @@ CatalogTestTool::CatalogTestTool(const std::string& name)
     : name_(name), manifest_(), spooler_(), history_() {}
 
 bool CatalogTestTool::Init() {
-  if (!InitDownloadManager(true)) {
+  if (!InitDownloadManager(true, "")) {
     return false;
   }
 
@@ -260,21 +282,20 @@ bool CatalogTestTool::Init() {
     return false;
   }
 
-  perf::Statistics stats;
-  manifest_ = CreateRepository(temp_dir_, spooler_);
+  manifest_ = CreateRepository(temp_dir_, spooler_.weak_ref());
 
   if (!manifest_.IsValid()) {
     return false;
   }
 
   shash::Any history_hash(shash::kSha1);
-  CreateHistory(stratum0_, manifest_, &history_hash);
+  CreateHistory(stratum0_, manifest_.weak_ref(), &history_hash);
 
   manifest_->set_certificate(hash_cert);
   manifest_->set_history(history_hash);
   manifest_->set_repository_name("keys.cern.ch");
   manifest_->set_publish_timestamp(time(NULL));
-  CreateManifest(stratum0_, manifest_);
+  CreateManifest(stratum0_, manifest_.weak_ref());
 
 
   history_.clear();
@@ -283,16 +304,19 @@ bool CatalogTestTool::Init() {
   return true;
 }
 
-// Note: we always apply the dir spec to the revision corresponding to the
-// original,
-//       empty repository.
-bool CatalogTestTool::Apply(const std::string& id, const DirSpec& spec) {
-  perf::Statistics stats;
-  UniquePtr<catalog::WritableCatalogManager> catalog_mgr(
-      CreateCatalogMgr(history_.front().second, "file://" + stratum0_,
-                       temp_dir_, spooler_, download_manager(), &stats));
+void CatalogTestTool::UpdateManifest() {
+  CreateManifest(stratum0_, manifest_.weak_ref());
+}
 
-  if (!catalog_mgr.IsValid()) {
+// Note: we always apply the dir spec to the revision corresponding to the
+// original, empty repository.
+bool CatalogTestTool::Apply(const std::string& id, const DirSpec& spec) {
+  statistics_ = new perf::Statistics();
+  catalog_mgr_ =
+    CreateCatalogMgr(history_.front().second, "file://" + stratum0_,
+                     temp_dir_, spooler_.weak_ref(), download_manager(),
+                     statistics_.weak_ref());
+  if (!catalog_mgr_.IsValid()) {
     return false;
   }
 
@@ -300,14 +324,21 @@ bool CatalogTestTool::Apply(const std::string& id, const DirSpec& spec) {
        it != spec.items().end(); ++it) {
     const DirSpecItem& item = it->second;
     if (item.entry_.IsRegular() || item.entry_.IsLink()) {
-      catalog_mgr->AddFile(item.entry_base(), item.xattrs(), item.parent());
+      catalog_mgr_->AddFile(item.entry_base(), item.xattrs(), item.parent());
     } else if (item.entry_.IsDirectory()) {
-      catalog_mgr->AddDirectory(
+      catalog_mgr_->AddDirectory(
         item.entry_base(), item.xattrs(), item.parent());
     }
   }
 
-  if (!catalog_mgr->Commit(false, 0, manifest_)) {
+  DirSpec::NestedCatalogList::const_iterator it;
+  for (it = spec.nested_catalogs().begin();
+       it != spec.nested_catalogs().end(); ++it) {
+    catalog_mgr_->CreateNestedCatalog(*it);
+  }
+
+
+  if (!catalog_mgr_->Commit(false, 0, manifest_.weak_ref())) {
     return false;
   }
 
@@ -320,12 +351,12 @@ bool CatalogTestTool::ApplyAtRootHash(
   const shash::Any& root_hash,
   const DirSpec& spec
 ) {
-  perf::Statistics stats;
-  UniquePtr<catalog::WritableCatalogManager> catalog_mgr(
-      CreateCatalogMgr(root_hash, "file://" + stratum0_, temp_dir_, spooler_,
-                       download_manager(), &stats));
-
-  if (!catalog_mgr.IsValid()) {
+  statistics_ = new perf::Statistics();
+  catalog_mgr_ =
+    CreateCatalogMgr(root_hash, "file://" + stratum0_, temp_dir_,
+                     spooler_.weak_ref(),
+                     download_manager(), statistics_.weak_ref());
+  if (!catalog_mgr_.IsValid()) {
     return false;
   }
 
@@ -333,9 +364,9 @@ bool CatalogTestTool::ApplyAtRootHash(
        it != spec.items().end(); ++it) {
     const DirSpecItem& item = it->second;
     if (item.entry_.IsRegular() || item.entry_.IsLink()) {
-      catalog_mgr->AddFile(item.entry_base(), item.xattrs(), item.parent());
+      catalog_mgr_->AddFile(item.entry_base(), item.xattrs(), item.parent());
     } else if (item.entry_.IsDirectory()) {
-      catalog_mgr->AddDirectory(
+      catalog_mgr_->AddDirectory(
         item.entry_base(), item.xattrs(), item.parent());
     }
   }
@@ -343,10 +374,10 @@ bool CatalogTestTool::ApplyAtRootHash(
   DirSpec::NestedCatalogList::const_iterator it;
   for (it = spec.nested_catalogs().begin();
        it != spec.nested_catalogs().end(); ++it) {
-    catalog_mgr->CreateNestedCatalog(*it);
+    catalog_mgr_->CreateNestedCatalog(*it);
   }
 
-  if (!catalog_mgr->Commit(false, 0, manifest_)) {
+  if (!catalog_mgr_->Commit(false, 0, manifest_.weak_ref())) {
     return false;
   }
 
@@ -360,8 +391,8 @@ bool CatalogTestTool::LookupNestedCatalogHash(
 ) {
   perf::Statistics stats;
   UniquePtr<catalog::WritableCatalogManager> catalog_mgr(
-      CreateCatalogMgr(root_hash, "file://" + stratum0_, temp_dir_, spooler_,
-                       download_manager(), &stats));
+      CreateCatalogMgr(root_hash, "file://" + stratum0_, temp_dir_,
+                       spooler_.weak_ref(), download_manager(), &stats));
   if (!catalog_mgr.IsValid()) {
     return false;
   }
@@ -372,7 +403,7 @@ bool CatalogTestTool::LookupNestedCatalogHash(
 
   catalog::DirectoryEntry entry;
   // This lookup is used to ensure the needed catalogs are mounted
-  catalog_mgr->LookupPath(path, catalog::kLookupSole, &entry);
+  catalog_mgr->LookupPath(path, catalog::kLookupDefault, &entry);
 
   p.Assign(&path[0], path.length());
   shash::Any hash = catalog_mgr->GetNestedCatalogHash(p);
@@ -398,13 +429,13 @@ bool CatalogTestTool::FindEntry(
 ) {
   perf::Statistics stats;
   UniquePtr<catalog::WritableCatalogManager> catalog_mgr(
-      CreateCatalogMgr(root_hash, "file://" + stratum0_, temp_dir_, spooler_,
-                       download_manager(), &stats));
+      CreateCatalogMgr(root_hash, "file://" + stratum0_, temp_dir_,
+                       spooler_.weak_ref(), download_manager(), &stats));
   if (!catalog_mgr.IsValid()) {
     return false;
   }
 
-  if (!catalog_mgr->LookupPath(path, catalog::kLookupSole, entry)) {
+  if (!catalog_mgr->LookupPath(path, catalog::kLookupDefault, entry)) {
     LogCvmfs(kLogCatalog, kLogStderr,
              "catalog for directory '%s' cannot be found",
              path.c_str());
@@ -418,8 +449,8 @@ bool CatalogTestTool::DirSpecAtRootHash(const shash::Any& root_hash,
                                         DirSpec* spec) {
   perf::Statistics stats;
   UniquePtr<catalog::WritableCatalogManager> catalog_mgr(
-      CreateCatalogMgr(root_hash, "file://" + stratum0_, temp_dir_, spooler_,
-                       download_manager(), &stats));
+      CreateCatalogMgr(root_hash, "file://" + stratum0_, temp_dir_,
+                       spooler_.weak_ref(), download_manager(), &stats));
 
   if (!catalog_mgr.IsValid()) {
     return false;
@@ -456,7 +487,7 @@ catalog::WritableCatalogManager* CatalogTestTool::CreateCatalogMgr(
   catalog::WritableCatalogManager* catalog_mgr =
       new catalog::WritableCatalogManager(root_hash, stratum0, temp_dir,
                                           spooler, dl_mgr, false, 0, 0, 0,
-                                          stats, false, 0, 0);
+                                          stats, false, 0, 0, "");
   catalog_mgr->Init();
 
   return catalog_mgr;
@@ -467,9 +498,10 @@ void CatalogTestTool::CreateHistory(
   manifest::Manifest *manifest,
   shash::Any *history_hash
 ) {
+  const string history_path = CreateTempPath(repo_path_ + "/history", 0600);
   {
     UniquePtr<history::SqliteHistory> history(
-      history::SqliteHistory::Create(repo_path_ + "/history",
+      history::SqliteHistory::Create(history_path,
                                      "keys.cern.ch"));
     ASSERT_TRUE(history.IsValid());
     history::History::Tag tag;
@@ -479,10 +511,10 @@ void CatalogTestTool::CreateHistory(
     ASSERT_TRUE(history->Insert(tag));
   }
   history_hash->suffix = shash::kSuffixHistory;
-  ASSERT_TRUE(zlib::CompressPath2Null(repo_path_ + "/history", history_hash));
+  ASSERT_TRUE(zlib::CompressPath2Null(history_path, history_hash));
   ASSERT_TRUE(
     zlib::CompressPath2Path(
-      repo_path_ + "/history",
+      history_path,
       repo_path_ + "/data/" + history_hash->MakePath()));
 }
 
