@@ -13,9 +13,9 @@
 #include <string>
 #include <vector>
 
-#include "dns.h"
-#include "logging.h"
-#include "platform.h"
+#include "network/dns.h"
+#include "util/logging.h"
+#include "util/platform.h"
 #include "util/pointer.h"
 #include "util/posix.h"
 #include "util/string.h"
@@ -235,6 +235,11 @@ TEST_F(T_Dns, ExtractHost) {
   EXPECT_EQ(ExtractHost("http://:"), "");
   EXPECT_EQ(ExtractHost("http://["), "");
   EXPECT_EQ(ExtractHost("http://[]"), "[]");
+  EXPECT_EQ(ExtractHost("http://user:pw@localhost:3128"), "localhost");
+  EXPECT_EQ(ExtractHost("http://user:pw@[::1]:3128"), "[::1]");
+  EXPECT_EQ(ExtractHost("http://user:pw@localhost/path"), "localhost");
+  EXPECT_EQ(ExtractHost("http://user:pw@localhost/p@th"), "localhost");
+  EXPECT_EQ(ExtractHost("http://user:pw@"), "");
 }
 
 
@@ -639,14 +644,14 @@ TEST_F(T_Dns, CaresResolverMany) {
   names.push_back("l.root-servers.net");
   names.push_back("m.root-servers.net");
   names.push_back("127.0.0.1");
-  names.push_back("nemo.root-servers.net");
+  names.push_back("nemo.example.com");
   vector<Host> hosts;
   default_resolver->ResolveMany(names, &hosts);
   ASSERT_EQ(hosts.size(), names.size());
   ExpectResolvedName(hosts[0], "a.root-servers.net",
                      "198.41.0.4", "[2001:503:ba3e::2:30]");
   ExpectResolvedName(hosts[1], "b.root-servers.net",
-                     "199.9.14.201", "[2001:500:200::b]");
+                     "170.247.170.2", "[2801:1b8:10::b]");
   ExpectResolvedName(hosts[2], "c.root-servers.net",
                      "192.33.4.12", "[2001:500:2::c]");
   ExpectResolvedName(hosts[3], "d.root-servers.net",
@@ -697,22 +702,6 @@ TEST_F(T_Dns, CaresResolverFinalDot) {
 }
 
 
-// TODO(jblomer): figure out why this fails on Travis
-TEST_F(T_Dns, CaresResolverLocalhost) {
-  Host host = default_resolver->Resolve("localhost");
-  // Not using ExpectResolvedName because the canonical name for localhost
-  // differs from system to system.
-  ASSERT_EQ(1U, host.ipv4_addresses().size());
-  EXPECT_EQ("127.0.0.1", *host.ipv4_addresses().begin());
-  if (host.HasIpv6()) {
-    ASSERT_EQ(1U, host.ipv6_addresses().size());
-    EXPECT_EQ("[::1]", *host.ipv6_addresses().begin());
-  } else {
-    EXPECT_EQ(0U, host.ipv6_addresses().size());
-  }
-}
-
-
 TEST_F(T_Dns, CaresResolverSearchDomainSlow) {
   Host host = default_resolver->Resolve("a");
   EXPECT_TRUE((host.status() == kFailUnknownHost) ||
@@ -720,7 +709,7 @@ TEST_F(T_Dns, CaresResolverSearchDomainSlow) {
               (host.status() == kFailInvalidResolvers));
 
   vector<string> new_domains;
-  new_domains.push_back("no.such.domain");
+  new_domains.push_back("example.com");
   new_domains.push_back("root-servers.net");
   bool retval = default_resolver->SetSearchDomains(new_domains);
   EXPECT_EQ(retval, true);
@@ -763,12 +752,22 @@ TEST_F(T_Dns, CaresResolverReadConfig) {
 
   vector<string> system_resolvers = default_resolver->resolvers();
   vector<string> system_domains = default_resolver->domains();
-  sort(system_resolvers.begin(), system_resolvers.end());
   sort(system_domains.begin(), system_domains.end());
-  sort(nameservers.begin(), nameservers.end());
   sort(domains.begin(), domains.end());
-  EXPECT_EQ(nameservers, system_resolvers);
-  EXPECT_EQ(domains, system_domains);
+
+  // On macOS, libresolv might filter out name servers listed in
+  // /etc/resolv.conf
+  EXPECT_FALSE(nameservers.empty());
+  bool found = false;
+  for (unsigned i = 0; i < nameservers.size(); ++i) {
+    if (std::find(system_resolvers.begin(), system_resolvers.end(),
+                  nameservers[i]) != system_resolvers.end())
+    {
+      found = true;
+      break;
+    }
+  }
+  EXPECT_TRUE(found);
 }
 
 
@@ -980,16 +979,12 @@ TEST_F(T_Dns, HostfileResolverMultipleAddresses) {
 }
 
 TEST_F(T_Dns, HostfileResolverBlankLines) {
-  g_log_messages.clear();
-  SetAltLogFunc(TDnsAltLogFunc);
   CreateHostfile("   \n  #comment\n\n\n127.0.0.1 localhost\n\n");
   Host host = hostfile_resolver->Resolve("localhost");
   EXPECT_EQ(kFailOk, host.status());
   set<string> expected_ipv4;
   expected_ipv4.insert("127.0.0.1");
   EXPECT_EQ(expected_ipv4, host.ipv4_addresses());
-  EXPECT_EQ(0U, g_log_messages.size());
-  SetAltLogFunc(NULL);
 }
 
 TEST_F(T_Dns, HostfileResolverTooLong) {
@@ -1110,7 +1105,7 @@ TEST_F(T_Dns, NormalResolverCombinedSlow) {
   ExpectResolvedName(hosts[0], "a.root-servers.net",
                      "198.41.0.4", "[2001:503:ba3e::2:30]");
   ExpectResolvedName(hosts[1], "b.root-servers.net",
-                     "199.9.14.201", "[2001:500:200::b]");
+                     "170.247.170.2", "[2801:1b8:10::b]");
   EXPECT_EQ(kFailOk, hosts[2].status());
   ExpectResolvedName(hosts[3], "127.0.0.1", "127.0.0.1", "");
   ExpectResolvedName(hosts[4], "[::1]", "", "[::1]");
