@@ -11,16 +11,16 @@
 #include <utility>
 #include <vector>
 
-#include "atomic.h"
 #include "c_file_sandbox.h"
 #include "c_mock_uploader.h"
+#include "crypto/hash.h"
 #include "file_chunk.h"
-#include "hash.h"
 #include "ingestion/item_mem.h"
 #include "ingestion/pipeline.h"
-#include "prng.h"
-#include "smalloc.h"
 #include "testutil.h"
+#include "util/atomic.h"
+#include "util/prng.h"
+#include "util/smalloc.h"
 #include "util/string.h"
 
 
