@@ -20,8 +20,9 @@
 using namespace std;  // NOLINT
 
 static void cvmfs_log_ignore(const char *msg) {
-  // Remove comment to debug test failures
-  // fprintf(stderr, "%s\n", msg);
+#ifdef DEBUGMSG
+  printf("(t_libcvmfs) %s\n", msg);
+#endif
 }
 
 class T_Libcvmfs : public ::testing::Test {
@@ -97,13 +98,19 @@ TEST_F(T_Libcvmfs, InitFailures) {
   stderr = fdevnull_;
   retval = cvmfs_init("bad option");
   stderr = save_stderr;
-  ASSERT_EQ(LIBCVMFS_FAIL_BADOPT, retval);
+  EXPECT_EQ(LIBCVMFS_FAIL_BADOPT, retval);
+  if (retval == LIBCVMFS_FAIL_OK)
+    cvmfs_fini();
 
-  retval = cvmfs_init((opt_cache_ + ",max_open_files=100000000").c_str());
-  ASSERT_EQ(LIBCVMFS_FAIL_NOFILES, retval);
+  retval = cvmfs_init((opt_cache_ + ",max_open_files=2000000000").c_str());
+  EXPECT_EQ(LIBCVMFS_FAIL_NOFILES, retval);
+  if (retval == LIBCVMFS_FAIL_OK)
+    cvmfs_fini();
 
   retval = cvmfs_init("");
-  ASSERT_EQ(LIBCVMFS_FAIL_MKCACHE, retval);
+  EXPECT_EQ(LIBCVMFS_FAIL_MKCACHE, retval);
+  if (retval == LIBCVMFS_FAIL_OK)
+    cvmfs_fini();
 
   sqlite3_shutdown();
 }
@@ -175,37 +182,51 @@ TEST_F(T_Libcvmfs, OptionAliases) {
   FILE *save_stderr = stderr;
 
   retval = cvmfs_init((opt_cache_ +
-    ",nofiles=100000000,max_open_files=100000000").c_str());
-  ASSERT_EQ(LIBCVMFS_FAIL_NOFILES, retval);
+    ",nofiles=2000000000,max_open_files=2000000000").c_str());
+  EXPECT_EQ(LIBCVMFS_FAIL_NOFILES, retval);
+  if (retval == LIBCVMFS_FAIL_OK)
+    cvmfs_fini();
 
-  retval = cvmfs_init((opt_cache_ + ",nofiles=100000000").c_str());
-  ASSERT_EQ(LIBCVMFS_FAIL_NOFILES, retval);
+  retval = cvmfs_init((opt_cache_ + ",nofiles=2000000000").c_str());
+  EXPECT_EQ(LIBCVMFS_FAIL_NOFILES, retval);
+  if (retval == LIBCVMFS_FAIL_OK)
+    cvmfs_fini();
 
-  retval = cvmfs_init((opt_cache_ + ",max_open_files=100000000").c_str());
-  ASSERT_EQ(LIBCVMFS_FAIL_NOFILES, retval);
+  retval = cvmfs_init((opt_cache_ + ",max_open_files=2000000000").c_str());
+  EXPECT_EQ(LIBCVMFS_FAIL_NOFILES, retval);
+  if (retval == LIBCVMFS_FAIL_OK)
+    cvmfs_fini();
 
   stderr = fdevnull_;
   retval = cvmfs_init((opt_cache_ + ",nofiles=999,max_open_files=888").c_str());
   stderr = save_stderr;
-  ASSERT_EQ(LIBCVMFS_FAIL_BADOPT, retval);
+  EXPECT_EQ(LIBCVMFS_FAIL_BADOPT, retval);
+  if (retval == LIBCVMFS_FAIL_OK)
+    cvmfs_fini();
 
   stderr = fdevnull_;
   retval =
     cvmfs_init((opt_cache_ + ",syslog_level=0,log_syslog_level=1").c_str());
   stderr = save_stderr;
-  ASSERT_EQ(LIBCVMFS_FAIL_BADOPT, retval);
+  EXPECT_EQ(LIBCVMFS_FAIL_BADOPT, retval);
+  if (retval == LIBCVMFS_FAIL_OK)
+    cvmfs_fini();
 
   stderr = fdevnull_;
   retval =
     cvmfs_init((opt_cache_ + ",log_file=abc,logfile=efg").c_str());
   stderr = save_stderr;
-  ASSERT_EQ(LIBCVMFS_FAIL_BADOPT, retval);
+  EXPECT_EQ(LIBCVMFS_FAIL_BADOPT, retval);
+  if (retval == LIBCVMFS_FAIL_OK)
+    cvmfs_fini();
 
   stderr = fdevnull_;
   retval =
     cvmfs_init((opt_cache_ + ",cachedir=/abc").c_str());
   stderr = save_stderr;
-  ASSERT_EQ(LIBCVMFS_FAIL_BADOPT, retval);
+  EXPECT_EQ(LIBCVMFS_FAIL_BADOPT, retval);
+  if (retval == LIBCVMFS_FAIL_OK)
+    cvmfs_fini();
 
   retval =
     cvmfs_init((opt_cache_ + ",cachedir=" + tmp_path_).c_str());
@@ -222,8 +243,11 @@ TEST_F(T_Libcvmfs, Initv2) {
   EXPECT_EQ(LIBCVMFS_ERR_OK, cvmfs_init_v2(opts));
   cvmfs_fini();
 
-  cvmfs_options_set(opts, "CVMFS_NFILES", "100000000");
-  EXPECT_EQ(LIBCVMFS_ERR_PERMISSION, cvmfs_init_v2(opts));
+  cvmfs_options_set(opts, "CVMFS_NFILES", "2000000000");
+  int retval = cvmfs_init_v2(opts);
+  EXPECT_EQ(LIBCVMFS_ERR_PERMISSION, retval);
+  if (retval == LIBCVMFS_FAIL_OK)
+    cvmfs_fini();
 
   cvmfs_options_fini(opts);
 }
@@ -278,12 +302,12 @@ TEST_F(T_Libcvmfs, TemplatingSlow) {
 
 namespace {
 // Create some default hashes for DirSpec
-const char* g_hashes[] = {"b026324c6904b2a9cb4b88d6d61c81d1000000",
-                         "26ab0db90d72e28ad0ba1e22ee510510000000",
-                         "6d7fce9fee471194aa8b5b6e47267f03000000",
-                         "48a24b70a0b376535542b996af517398000000",
-                         "1dcca23355272056f04fe8bf20edfce0000000",
-                         "11111111111111111111111111111111111111"};
+const char* g_hashes[] = {"b026324c6904b2a9cb4b88d6d61c81d100000000",
+                          "26ab0db90d72e28ad0ba1e22ee51051000000000",
+                          "6d7fce9fee471194aa8b5b6e47267f0300000000",
+                          "48a24b70a0b376535542b996af51739800000000",
+                          "1dcca23355272056f04fe8bf20edfce000000000",
+                          "1111111111111111111111111111111111111111"};
 
 const size_t g_file_size = 4096;
 }  // anonymous namespace
@@ -336,6 +360,7 @@ TEST_F(T_Libcvmfs, Stat) {
   // Create file structure
   DirSpec spec1 = MakeBaseSpec();
   EXPECT_TRUE(tester.ApplyAtRootHash(tester.manifest()->catalog_hash(), spec1));
+  tester.DestroyCatalogManager();
 
   // Find directory entry for use later
   catalog::DirectoryEntry entry;
@@ -395,6 +420,7 @@ TEST_F(T_Libcvmfs, Attr) {
   DirSpec spec = MakeBaseSpec();
   EXPECT_TRUE(spec.LinkFile("link", "dir", "file1", g_file_size));
   EXPECT_TRUE(tester.ApplyAtRootHash(tester.manifest()->catalog_hash(), spec));
+  tester.DestroyCatalogManager();
 
   // Find directory entry for use later
   catalog::DirectoryEntry entry;
@@ -485,6 +511,7 @@ TEST_F(T_Libcvmfs, Listdir) {
 
   // Apply the created DirSpec
   EXPECT_TRUE(tester.ApplyAtRootHash(tester.manifest()->catalog_hash(), spec1));
+  tester.DestroyCatalogManager();
 
   // Set CVMFS options to reflect created repository
   cvmfs_options_set(opts,
@@ -544,6 +571,7 @@ TEST_F(T_Libcvmfs, StatNestedCatalog) {
 
   // Apply the created DirSpec
   EXPECT_TRUE(tester.ApplyAtRootHash(tester.manifest()->catalog_hash(), spec));
+  tester.DestroyCatalogManager();
   char *d0_hash = strdup(tester.manifest()->catalog_hash().ToString().c_str());
 
   // Find the hash of the newly added Nested Catalog for comparison
@@ -629,6 +657,7 @@ TEST_F(T_Libcvmfs, ListNestedCatalogs) {
   // Apply created DirSpec
   EXPECT_TRUE(
     tester.ApplyAtRootHash(tester.manifest()->catalog_hash(), spec));
+  tester.DestroyCatalogManager();
 
   // Set CVMFS options to reflect created repository
   cvmfs_options_set(opts,
@@ -679,3 +708,153 @@ TEST_F(T_Libcvmfs, ListNestedCatalogs) {
   cvmfs_fini();
   cvmfs_options_fini(opts);
 }
+
+TEST_F(T_Libcvmfs, ListStat) {
+  // Initialize options
+  cvmfs_option_map *opts = cvmfs_options_init();
+
+  // Create and initialize repository named "stat"
+  CatalogTestTool tester("list_stat");
+  EXPECT_TRUE(tester.Init());
+
+  // Create file structure
+  DirSpec spec1 = MakeBaseSpec();
+  EXPECT_TRUE(tester.ApplyAtRootHash(tester.manifest()->catalog_hash(), spec1));
+  tester.DestroyCatalogManager();
+
+  // Set CVMFS options to reflect created repository
+  cvmfs_options_set(opts, "CVMFS_ROOT_HASH",
+                        tester.manifest()->catalog_hash().ToString().c_str());
+  cvmfs_options_set(opts, "CVMFS_SERVER_URL",
+                        ("file://" + tester.repo_name()).c_str());
+  cvmfs_options_set(opts, "CVMFS_HTTP_PROXY", "DIRECT");
+  cvmfs_options_set(opts, "CVMFS_PUBLIC_KEY",
+                        tester.public_key().c_str());
+  cvmfs_options_set(opts, "CVMFS_CACHE_DIR",
+                        (tester.repo_name()+"/data/txn").c_str());
+  cvmfs_options_set(opts, "CVMFS_MOUNT_DIR",
+                        ("/cvmfs" + tester.repo_name()).c_str());
+
+  // Initialize client repo based on options
+  ASSERT_EQ(LIBCVMFS_ERR_OK, cvmfs_init_v2(opts));
+
+  // Attach to client repo
+  cvmfs_context *ctx;
+  EXPECT_EQ(LIBCVMFS_ERR_OK,
+    cvmfs_attach_repo_v2((tester.repo_name().c_str()), opts, &ctx));
+
+  struct cvmfs_stat_t *buf = 0;
+  size_t buflen = 0;
+  size_t listlen = 0;
+
+  int retval = cvmfs_listdir_stat(ctx, "dir", &buf, &listlen, &buflen);
+  EXPECT_EQ(0, retval);
+
+  EXPECT_EQ(4U, listlen);
+
+  for (size_t i = 0; i < listlen; ++i) {
+    if (!strcmp(buf[i].name, "dir")) {
+      EXPECT_TRUE(buf[i].info.st_mode & S_IFDIR);
+    } else if (!strcmp(buf[i].name, "dir2")) {
+      EXPECT_TRUE(buf[i].info.st_mode & S_IFDIR);
+
+    } else if (!strcmp(buf[i].name, "dir3")) {
+      EXPECT_TRUE(buf[i].info.st_mode & S_IFDIR);
+
+    } else if (!strcmp(buf[i].name, "file1")) {
+      EXPECT_TRUE(buf[i].info.st_mode & S_IFREG);
+      EXPECT_EQ(static_cast<int>(g_file_size), buf[i].info.st_size);
+
+    } else {
+      FAIL() << "unexpected object in list: " << buf[i].name;
+    }
+  }
+
+  // Finalize and close repo and options
+  cvmfs_detach_repo(ctx);
+  cvmfs_fini();
+  cvmfs_options_fini(opts);
+}
+
+
+TEST_F(T_Libcvmfs, Remount) {
+  // Initialize options
+  cvmfs_option_map *opts = cvmfs_options_init();
+
+  // Create and initialize repository
+  CatalogTestTool tester("remount");
+  EXPECT_TRUE(tester.Init());
+
+  // Set CVMFS options to reflect created repository
+  cvmfs_options_set(opts, "CVMFS_SERVER_URL",
+                        ("file://" + tester.repo_name()).c_str());
+  cvmfs_options_set(opts, "CVMFS_HTTP_PROXY", "DIRECT");
+  cvmfs_options_set(opts, "CVMFS_PUBLIC_KEY",
+                        tester.public_key().c_str());
+  cvmfs_options_set(opts, "CVMFS_CACHE_DIR",
+                        (tester.repo_name()+"/data/txn").c_str());
+  cvmfs_options_set(opts, "CVMFS_MOUNT_DIR",
+                        ("/cvmfs" + tester.repo_name()).c_str());
+
+  // libcvmfs changes the sqlite global state, so we have to call it in another
+  // process in order to change the catalog along the way.
+  int pipe_send[2];
+  int pipe_recv[2];
+  MakePipe(pipe_send);
+  MakePipe(pipe_recv);
+  pid_t pid;
+  if ((pid = fork()) == 0) {
+    // Initialize client repo based on options
+    int retval = cvmfs_init_v2(opts);
+    EXPECT_EQ(LIBCVMFS_ERR_OK, retval);
+    if (retval != LIBCVMFS_ERR_OK) {
+      char c = '!';
+      WritePipe(pipe_send[1], &c, 1);
+      exit(1);
+    }
+
+    // Attach to client repo
+    cvmfs_context *ctx;
+    retval = cvmfs_attach_repo_v2("keys.cern.ch", opts, &ctx);
+    EXPECT_EQ(LIBCVMFS_ERR_OK, retval);
+    if (retval != LIBCVMFS_ERR_OK) {
+      char c = '!';
+      WritePipe(pipe_send[1], &c, 1);
+      exit(1);
+    }
+
+    EXPECT_EQ(0u, cvmfs_get_revision(ctx));
+    EXPECT_EQ(0, cvmfs_remount(ctx));
+    EXPECT_EQ(0u, cvmfs_get_revision(ctx));
+
+    char c = '1';
+    WritePipe(pipe_send[1], &c, 1);
+    ReadPipe(pipe_recv[0], &c, 1);
+
+    EXPECT_EQ(0, cvmfs_remount(ctx));
+    EXPECT_EQ(1u, cvmfs_get_revision(ctx));
+
+    // Finalize and close repo and options
+    cvmfs_detach_repo(ctx);
+    cvmfs_fini();
+    exit(HasFailure() ? 1 : 0);
+  }
+
+  char c;
+  ReadPipe(pipe_send[0], &c, 1);
+  if (c != '!') {
+    DirSpec spec1 = MakeBaseSpec();
+    EXPECT_TRUE(
+      tester.ApplyAtRootHash(tester.manifest()->catalog_hash(), spec1));
+    tester.UpdateManifest();
+    tester.DestroyCatalogManager();
+    WritePipe(pipe_recv[1], &c, 1);
+  }
+
+  int retcode = WaitForChild(pid);
+  EXPECT_EQ(0, retcode);
+
+  ClosePipe(pipe_send);
+  ClosePipe(pipe_recv);
+  cvmfs_options_fini(opts);
+}
