@@ -13,10 +13,10 @@
 #include <cstring>
 #include <string>
 
-#include "duplex_ssl.h"
-#include "hash.h"
-#include "prng.h"
-#include "smalloc.h"
+#include "crypto/hash.h"
+#include "crypto/openssl_version.h"
+#include "util/prng.h"
+#include "util/smalloc.h"
 #include "util/string.h"
 
 using namespace std;  // NOLINT
