@@ -10,6 +10,7 @@
 #include <pthread.h>
 #include <sys/resource.h>
 #include <sys/socket.h>
+#include <sys/syscall.h>
 #include <sys/un.h>
 #include <unistd.h>
 
@@ -18,14 +19,15 @@
 #include <limits>
 #include <vector>
 
-#include "atomic.h"
 #include "shortstring.h"
-#include "smalloc.h"
 #include "testutil.h"
 #include "util/algorithm.h"
+#include "util/atomic.h"
 #include "util/file_guard.h"
 #include "util/mmap_file.h"
+#include "util/pipe.h"
 #include "util/posix.h"
+#include "util/smalloc.h"
 #include "util/string.h"
 
 using namespace std;  // NOLINT
@@ -143,6 +145,13 @@ TEST_F(T_Util, GetShell) {
 }
 
 
+TEST_F(T_Util, GetUserNameOf) {
+  std::string name;
+  EXPECT_TRUE(GetUserNameOf(0, &name));
+  EXPECT_EQ("root", name);
+}
+
+
 TEST_F(T_Util, GetUidOf) {
   uid_t uid;
   gid_t gid;
@@ -159,6 +168,24 @@ TEST_F(T_Util, GetHomeDirectory) {
 }
 
 
+TEST_F(T_Util, GetArch) {
+  int fd_stdin;
+  int fd_stdout;
+  int fd_stderr;
+  ASSERT_TRUE(Shell(&fd_stdin, &fd_stdout, &fd_stderr));
+  ASSERT_TRUE(SafeWrite(fd_stdin, "uname -m\n", 18));
+  close(fd_stdin);
+  char arch[257];
+  ssize_t nbytes = SafeRead(fd_stdout, arch, 256);
+  close(fd_stdout);
+  close(fd_stderr);
+  ASSERT_GT(nbytes, 0);
+  ASSERT_LT(nbytes, 257);
+  arch[nbytes] = '\0';
+  EXPECT_EQ(Trim(arch, true /* trim_newline */), GetArch());
+}
+
+
 TEST_F(T_Util, GetGidOf) {
 #ifdef __APPLE__
   const std::string group_name = "wheel";
@@ -316,15 +343,23 @@ TEST_F(T_Util, String2Uint64) {
 
 TEST_F(T_Util, IsHttpUrl) {
   EXPECT_TRUE(IsHttpUrl("http://cvmfs-stratum-one.cern.ch/cvmfs/cms.cern.ch"));
+  EXPECT_TRUE(IsHttpUrl("https://cvmfs-stratum-one.cern.ch/cvmfs/cms.cern.ch"));
   EXPECT_TRUE(IsHttpUrl("http://"));
+  EXPECT_TRUE(IsHttpUrl("https://"));
   EXPECT_TRUE(IsHttpUrl("http://foobar"));
+  EXPECT_TRUE(IsHttpUrl("https://foobar"));
   EXPECT_TRUE(IsHttpUrl("HTTP://www.google.com"));
+  EXPECT_TRUE(IsHttpUrl("HTTPS://www.google.com"));
   EXPECT_TRUE(IsHttpUrl("HTtP://cvmfs-stratum-zero.cern.ch/ot/atlas"));
+  EXPECT_TRUE(IsHttpUrl("HTtPs://cvmfs-stratum-zero.cern.ch/ot/atlas"));
   EXPECT_FALSE(IsHttpUrl("http:/foobar"));
+  EXPECT_FALSE(IsHttpUrl("https:/foobar"));
   EXPECT_FALSE(IsHttpUrl("http"));
+  EXPECT_FALSE(IsHttpUrl("https"));
   EXPECT_FALSE(IsHttpUrl("/srv/cvmfs/cms.cern.ch"));
   EXPECT_FALSE(IsHttpUrl("srv/cvmfs/cms.cern.ch"));
   EXPECT_FALSE(IsHttpUrl("http//foobar"));
+  EXPECT_FALSE(IsHttpUrl("https//foobar"));
 }
 
 TEST_F(T_Util, MakeCannonicalPath) {
@@ -361,6 +396,49 @@ TEST_F(T_Util, GetFileName) {
 }
 
 
+TEST_F(T_Util, GetFileSystemInfo) {
+  if (!DirectoryExists("/proc")) {
+    printf("Skipping\n");
+    return;
+  }
+
+  FileSystemInfo fs_info;
+  fs_info = GetFileSystemInfo("/proc");
+  EXPECT_EQ(kFsTypeProc, fs_info.type);
+  fs_info = GetFileSystemInfo("/");
+  EXPECT_EQ(kFsTypeUnknown, fs_info.type);
+}
+
+
+TEST_F(T_Util, ReadSymlink) {
+  EXPECT_TRUE(ReadSymlink(".").empty());
+  EXPECT_EQ(0, symlink(".", "cvmfs_read_symlink_test"));
+  EXPECT_EQ(".", ReadSymlink("cvmfs_read_symlink_test"));
+}
+
+
+TEST_F(T_Util, ResolvePath) {
+  EXPECT_EQ("/", ResolvePath("/"));
+  EXPECT_EQ("/", ResolvePath(""));
+  EXPECT_EQ("/no/such/path", ResolvePath("/no/such/path"));
+  EXPECT_EQ("/", ResolvePath("/.//././."));
+  EXPECT_EQ("/", ResolvePath("/usr/.."));
+
+  EXPECT_EQ(0, symlink(".", "cvmfs_test_link"));
+  EXPECT_EQ(GetCurrentWorkingDirectory(), ResolvePath("cvmfs_test_link"));
+
+  EXPECT_EQ(0, symlink("/no/such/path", "cvmfs_test_link_dangling"));
+  EXPECT_EQ("/no/such/path", ResolvePath("cvmfs_test_link_dangling"));
+}
+
+
+TEST_F(T_Util, IsMountPoint) {
+  EXPECT_TRUE(IsMountPoint(""));
+  EXPECT_TRUE(IsMountPoint("/"));
+  EXPECT_FALSE(IsMountPoint("/no/such/file"));
+}
+
+
 TEST_F(T_Util, SplitPath) {
   string dirname;
   string filename;
@@ -384,7 +462,8 @@ TEST_F(T_Util, SplitPath) {
 
 
 TEST_F(T_Util, CreateFile) {
-  ASSERT_DEATH(CreateFile("myfakepath/otherfakepath.txt", 0777), ".*");
+  EXPECT_THROW(CreateFile("myfakepath/otherfakepath.txt", 0777),
+               std::runtime_error);
   string filename = sandbox + "/createfile.txt";
   CreateFile(filename, 0600);
   FILE* myfile = fopen(filename.c_str(), "w");
@@ -490,13 +569,60 @@ TEST_F(T_Util, ReadHalfPipe) {
   ClosePipe(fd);
 }
 
+namespace {
+
+struct ReadHalfPipeInfo {
+  int fd;
+  int length;
+  void *dst;
+};
+
+static void *MainReadHalfPipe(void *data) {
+  ReadHalfPipeInfo *info = reinterpret_cast<ReadHalfPipeInfo *>(data);
+  bool retval;
+  do {
+    retval = ReadHalfPipe(info->fd, info->dst, info->length, 10);
+  } while (retval == false);
+  return NULL;
+}
+
+}  // anonymous namespace
+
+TEST_F(T_Util, ReadHalfPipeTimeout) {
+  int fd[2];
+  void *buffer_output = scalloc(20, sizeof(char));
+  MakePipe(fd);
+
+  const int size = static_cast<int>(to_write.length());
+
+  ReadHalfPipeInfo info;
+  info.fd = fd[0];
+  info.length = size;
+  info.dst = buffer_output;
+  pthread_t thread_read;
+  const int retval =
+    pthread_create(&thread_read, NULL, MainReadHalfPipe, &info);
+  assert(retval == 0);
+
+  SafeSleepMs(250);
+  EXPECT_EQ(size, write(fd[1], to_write.data(), size));
+
+  pthread_join(thread_read, NULL);
+
+  EXPECT_EQ(0,
+    memcmp(const_cast<char *>(to_write.data()), buffer_output, size));
+  free(buffer_output);
+  ClosePipe(fd);
+}
+
 TEST_F(T_Util, ClosePipe) {
   int fd[2];
   UniquePtr<void> buffer_output(scalloc(20, sizeof(char)));
   MakePipe(fd);
   ClosePipe(fd);
   ASSERT_DEATH(WritePipe(fd[1], to_write.c_str(), to_write.length()), ".*");
-  ASSERT_DEATH(ReadPipe(fd[0], buffer_output, to_write.length()), ".*");
+  ASSERT_DEATH(ReadPipe(fd[0], buffer_output.weak_ref(), to_write.length()),
+               ".*");
 }
 
 
@@ -699,6 +825,51 @@ TEST_F(T_Util, SendMes2Socket) {
 }
 
 
+TEST_F(T_Util, SendRecvFd) {
+  int fd_server = MakeSocket(socket_address, 0700);
+  ASSERT_LE(0, fd_server);
+  FdGuard fd_guard_server(fd_server);
+  listen(fd_server, 1);
+
+  pid_t pid = fork();
+  switch (pid) {
+    case -1:
+      ASSERT_TRUE(false);
+    case 0:
+      struct sockaddr_un client_addr;
+      unsigned int client_length = sizeof(client_addr);
+      int fd_conn = accept(fd_server, (struct sockaddr *) &client_addr,
+                           &client_length);
+      if (fd_conn < 0)
+        _exit(1);
+      int fd_recv = RecvFdFromSocket(fd_conn);
+      if (fd_recv < 0)
+        _exit(10 - fd_recv);
+      char zero = 42;
+      int retval = read(fd_recv, &zero, 1);
+      if (retval != 1)
+        _exit(3);
+      if (zero != 0)
+        _exit(4);
+      _exit(0);
+  }
+
+  int fd_client = ConnectSocket(socket_address);
+  ASSERT_LE(0, fd_client);
+  FdGuard fd_guard_client(fd_client);
+
+  int fd_pass = open("/dev/zero", O_RDONLY);
+  ASSERT_LE(0, fd_pass);
+  EXPECT_TRUE(SendFd2Socket(fd_client, fd_pass));
+  EXPECT_EQ(0, WaitForChild(pid));
+}
+
+
+TEST_F(T_Util, GetHostname) {
+  EXPECT_FALSE(GetHostname().empty());
+}
+
+
 TEST_F(T_Util, TcpEndpoints) {
   EXPECT_EQ(-1, MakeTcpEndpoint("foobar", 0));
   int fd_server = MakeTcpEndpoint("", 12345);
@@ -1020,6 +1191,48 @@ TEST_F(T_Util, FindDirectories) {
 }
 
 
+TEST_F(T_Util, ListDirectory) {
+  std::vector<std::string> names;
+  std::vector<mode_t> modes;
+
+  EXPECT_FALSE(ListDirectory("/no/such/dir", &names, &modes));
+  string dir = sandbox + "/test-listdir";
+  ASSERT_TRUE(MkdirDeep(dir, 0700));
+
+  EXPECT_TRUE(ListDirectory(dir, &names, &modes));
+  EXPECT_TRUE(names.empty());
+  EXPECT_TRUE(modes.empty());
+
+  ASSERT_TRUE(MkdirDeep(dir + "/dir1/sub", 0700));
+  ASSERT_TRUE(MkdirDeep(dir + "/dir2", 0700));
+  EXPECT_TRUE(SymlinkForced(dir + "/dir1", dir + "/dirX"));
+  string temp_file = CreateTempPath(dir + "/tempfile", 0600);
+  ASSERT_FALSE(temp_file.empty());
+
+  EXPECT_TRUE(ListDirectory(dir, &names, &modes));
+  ASSERT_EQ(4U, names.size());
+  EXPECT_EQ("dir1", names[0]);
+  EXPECT_EQ("dir2", names[1]);
+  EXPECT_EQ("dirX", names[2]);
+  EXPECT_EQ(GetFileName(temp_file), names[3]);
+  EXPECT_TRUE(S_ISDIR(modes[0]));
+  EXPECT_TRUE(S_ISDIR(modes[1]));
+  EXPECT_TRUE(S_ISLNK(modes[2]));
+  EXPECT_TRUE(S_ISREG(modes[3]));
+}
+
+
+TEST_F(T_Util, FindExecutable) {
+  std::string ls = FindExecutable("ls");
+  ASSERT_FALSE(ls.empty());
+  EXPECT_EQ('/', ls[0]);
+  std::string ls_abs = FindExecutable(ls);
+  EXPECT_EQ(ls, ls_abs);
+  std::string fail = FindExecutable("no-such-exe");
+  EXPECT_TRUE(fail.empty());
+}
+
+
 TEST_F(T_Util, GetUmask) {
   unsigned test_umask = 0755;
   mode_t original_mask = umask(test_umask);
@@ -1059,8 +1272,8 @@ TEST_F(T_Util, StringifyDouble) {
 }
 
 TEST_F(T_Util, StringifyTime) {
-  time_t now = time(NULL);
-  time_t other = 1263843;
+  const time_t now = time(NULL);
+  const time_t other = 1263843;
 
   EXPECT_EQ(GetTimeString(now, true), StringifyTime(now, true));
   EXPECT_EQ(GetTimeString(now, false), StringifyTime(now, false));
@@ -1068,6 +1281,20 @@ TEST_F(T_Util, StringifyTime) {
   EXPECT_EQ(GetTimeString(other, false), StringifyTime(other, false));
 }
 
+TEST_F(T_Util, StringifyLocalTime) {
+  struct stat tmpbuf;
+  if(stat("/usr/share/zoneinfo", &tmpbuf)) {
+    // TODO(vvolkl): use GTEST_SKIP once externals are updated
+    printf("Skipping test, no tzdata available.\n");
+  } else {
+    const time_t other = 1263843;
+    setenv("TZ", "America/Los_Angeles", true);
+    tzset();
+    EXPECT_EQ(StringifyLocalTime(other), "15 Jan 1970 07:04:03 PST");
+    unsetenv("TZ");
+  }
+}
+
 TEST_F(T_Util, RfcTimestamp) {
   char *curr_locale = setlocale(LC_TIME, NULL);
   const char *format = "%a, %e %h %Y %H:%M:%S %Z";
@@ -1188,23 +1415,24 @@ TEST_F(T_Util, SplitString) {
   string str2 = "my::string:by:colons";
   vector<string> result;
 
-  result = SplitString(str1, ' ', 1u);
+  result = SplitStringBounded(1, str1, ' ');
   EXPECT_EQ(1u, result.size());
   EXPECT_EQ(str1, result[0]);
 
-  result = SplitString(str1, ' ', 2u);
+  result = SplitStringBounded(2, str1, ' ');
   EXPECT_EQ(2u, result.size());
   EXPECT_EQ("the", result[0]);
   EXPECT_EQ("string that will be cut in peaces", result[1]);
 
-  result = SplitString(str1, ';', 200u);
+  result = SplitStringBounded(200, str1, ';');
   EXPECT_EQ(1u, result.size());
   EXPECT_EQ(str1, result[0]);
 
-  result = SplitString(str2, ':', 200u);
+  result = SplitStringBounded(200, str2, ':');
   EXPECT_EQ(5u, result.size());
   EXPECT_EQ("", result[1]);
-  EXPECT_EQ(SplitString(str2, ':', 5u), SplitString(str2, ':', 5000u));
+  EXPECT_EQ(SplitStringBounded(5, str2, ':'),
+            SplitStringBounded(5000, str2, ':'));
 }
 
 TEST_F(T_Util, JoinStrings) {
@@ -1397,6 +1625,13 @@ TEST_F(T_Util, ReplaceAll) {
           "REPLACED"));
 }
 
+TEST_F(T_Util, ProcessExists) {
+  EXPECT_TRUE(ProcessExists(getpid()));
+  EXPECT_TRUE(ProcessExists(1));
+  EXPECT_FALSE(ProcessExists(999999999));
+  EXPECT_DEATH(ProcessExists(0), ".*");
+}
+
 TEST_F(T_Util, BlockSignal) {
   EXPECT_DEATH(kill(getpid(), SIGUSR1), ".*");
   BlockSignal(SIGUSR1);
@@ -1417,7 +1652,7 @@ TEST_F(T_Util, WaitForSignal) {
 
 TEST_F(T_Util, WaitForChild) {
   ASSERT_DEATH(WaitForChild(0), ".*");
-  ASSERT_DEATH(WaitForChild(getpid()), ".*");
+  EXPECT_THROW(WaitForChild(getpid()), std::runtime_error);
 
   pid_t pid = fork();
   switch (pid) {
@@ -1552,17 +1787,54 @@ TEST_F(T_Util, ManagedExecCommandLine) {
   map<int, int> fd_map;
   fd_map[fd_stdout[1]] = 1;
 
-  success = ManagedExec(command_line, preserve_filedes, fd_map, true, true,
-      &pid);
+  success = ManagedExec(command_line, preserve_filedes, fd_map,
+                        true /* drop_credentials */, false /* clear_env */,
+                        true /* double_fork */,
+                        &pid);
   ASSERT_TRUE(success);
   close(fd_stdout[1]);
-  ssize_t bytes_read = read(fd_stdout[0], buffer, message.length());
+  ssize_t bytes_read = read(fd_stdout[0], buffer.weak_ref(), message.length());
   EXPECT_EQ(static_cast<size_t>(bytes_read), message.length());
   string result(reinterpret_cast<char *>(buffer.weak_ref()));
   ASSERT_EQ(message, result);
   close(fd_stdout[0]);
 }
 
+
+TEST_F(T_Util, ManagedExecClearEnv) {
+  bool success;
+  pid_t pid;
+  int fd_stdout[2];
+  int fd_stdin[2];
+  UniquePtr<unsigned char> buffer(static_cast<unsigned char*>(scalloc(100, 1)));
+  MakePipe(fd_stdout);
+  MakePipe(fd_stdin);
+  vector<string> command_line;
+  command_line.push_back("/usr/bin/env");
+
+  set<int> preserve_filedes;
+  preserve_filedes.insert(1);
+
+  map<int, int> fd_map;
+  fd_map[fd_stdout[1]] = 1;
+
+  success = ManagedExec(command_line, preserve_filedes, fd_map,
+                        true /* drop_credentials */, true /* clear_env */,
+                        true /* double_fork */,
+                        &pid);
+  close(fd_stdout[1]);
+  ASSERT_TRUE(success);
+  ssize_t bytes_read = read(fd_stdout[0], buffer.weak_ref(), 64);
+
+  // env will be cleared (normally it is way larger than 64 bytes)
+  // debug mode: there will be a log message in similar to 25-byte long:
+  // (cvmfs) execve'd /usr/bin/env (PID: 335527)    [12-15-2022 14:09:11 CET]
+  // thats way checking if bytes_read < 30 bytes
+  EXPECT_LE(bytes_read, 30);
+  close(fd_stdout[0]);
+}
+
+
 TEST_F(T_Util, ManagedExecRunShell) {
   int fd_stdin;
   int fd_stdout;
@@ -1570,7 +1842,7 @@ TEST_F(T_Util, ManagedExecRunShell) {
   bool retval = Shell(&fd_stdin, &fd_stdout, &fd_stderr);
   EXPECT_TRUE(retval);
 
-  Pipe shell_pipe(fd_stdout, fd_stdin);
+  Pipe<kPipeTest> shell_pipe(fd_stdout, fd_stdin);
   const char *command = "echo \"Hello World\"\n";
   retval = shell_pipe.Write(command, strlen(command));
   EXPECT_TRUE(retval);
@@ -1605,12 +1877,24 @@ TEST_F(T_Util, ManagedExecExecuteBinaryDoubleFork) {
   ASSERT_TRUE(retval);
   EXPECT_EQ(0, kill(child_pid, 0));
 
+  // Orphaned process are attached to some system process if their parent
+  // process dies before them.
+  // Traditionally this is "init" with pid 1.
+  // However, POSIX leaves this implementation-defined and e.g.
+  // Ubuntu uses "systemd" - which pid changes per session
+
   // check that the PPID of the process is 1 (belongs to init)
   pid_t child_parent_pid = GetParentPid(child_pid);
-  EXPECT_EQ(1, child_parent_pid);
+
+  if (child_parent_pid == 1) {
+    EXPECT_EQ(1, child_parent_pid);
+  } else {
+    std::string name = GetProcessname(child_parent_pid);
+    EXPECT_STREQ(name.c_str(), "systemd");
+  }
 
   // tell the process to terminate
-  Pipe shell_pipe(fd_stdout, fd_stdin);
+  Pipe<kPipeTest> shell_pipe(fd_stdout, fd_stdin);
   const char *quit = "quit\n";
   retval = shell_pipe.Write(quit, strlen(quit));
   EXPECT_TRUE(retval);
@@ -1656,7 +1940,7 @@ TEST_F(T_Util, ManagedExecExecuteBinaryAsChild) {
   EXPECT_EQ(my_pid, child_parent_pid);
 
   // tell the process to terminate
-  Pipe shell_pipe(fd_stdout, fd_stdin);
+  Pipe<kPipeTest> shell_pipe(fd_stdout, fd_stdin);
   const char *quit = "quit\n";
   retval = shell_pipe.Write(quit, strlen(quit));
   EXPECT_TRUE(retval);
@@ -1753,7 +2037,7 @@ TEST_F(T_Util, MemoryMappedFile) {
 
 
 TEST_F(T_Util, SetLimitNoFile) {
-  EXPECT_EQ(-1, SetLimitNoFile(100000000));
+  EXPECT_EQ(-1, SetLimitNoFile(3000000000U));
 
   struct rlimit rpl;
   memset(&rpl, 0, sizeof(rpl));
@@ -1768,7 +2052,52 @@ TEST_F(T_Util, GetLimitNoFile) {
   GetLimitNoFile(&soft_limit, &hard_limit);
   EXPECT_LT(0U, soft_limit);
   EXPECT_LE(soft_limit, hard_limit);
-  EXPECT_LT(hard_limit, 10000000U);
+  EXPECT_LT(hard_limit, 3000000000U);
+}
+
+
+TEST_F(T_Util, Lsof) {
+  std::vector<LsofEntry> list;
+  CreateFile("cvmfs_test_lsof", 0600, false /* ignore_failure */);
+  int fd = open("cvmfs_test_lsof", O_RDWR);
+  EXPECT_GE(fd, 0);
+  list = Lsof(GetCurrentWorkingDirectory());
+  close(fd);
+#ifndef __APPLE__
+  EXPECT_GT(list.size(), 0U);
+
+  bool found = false;
+  for (unsigned i = 0; i < list.size(); ++i) {
+    if (list[i].pid != getpid())
+      continue;
+    if (list[i].path != GetCurrentWorkingDirectory())
+      continue;
+
+    found = true;
+    EXPECT_EQ(geteuid(), list[i].owner);
+    EXPECT_EQ(ReadSymlink("/proc/self/exe"), list[i].executable);
+    EXPECT_TRUE(list[i].read_only);
+    break;
+  }
+  EXPECT_TRUE(found);
+
+  found = false;
+  for (unsigned i = 0; i < list.size(); ++i) {
+    if (list[i].pid != getpid())
+      continue;
+    if (list[i].path != GetCurrentWorkingDirectory() + "/cvmfs_test_lsof")
+      continue;
+
+    found = true;
+    EXPECT_EQ(geteuid(), list[i].owner);
+    EXPECT_EQ(ReadSymlink("/proc/self/exe"), list[i].executable);
+    EXPECT_FALSE(list[i].read_only);
+    break;
+  }
+  EXPECT_TRUE(found);
+#else
+  EXPECT_TRUE(list.empty());
+#endif
 }
 
 
@@ -1791,6 +2120,35 @@ TEST_F(T_Util, DiffTree) {
   EXPECT_FALSE(DiffTree(".", "/"));
 }
 
+TEST_F(T_Util, Bitfields) {
+  int field = 0;
+
+  SetBit(0, &field);
+  EXPECT_EQ(1, field);
+  EXPECT_TRUE(TestBit(0, field));
+  ClearBit(0, &field);
+  EXPECT_EQ(0, field);
+  EXPECT_FALSE(TestBit(0, field));
+
+  SetBit(1, &field);
+  SetBit(2, &field);
+  EXPECT_EQ(6, field);
+  EXPECT_TRUE(TestBit(2, field));
+  EXPECT_FALSE(TestBit(3, field));
+  ClearBit(3, &field);
+  EXPECT_EQ(6, field);
+  ClearBit(1, &field);
+  EXPECT_EQ(4, field);
+
+  int64_t long_field = 0;
+  SetBit(63, &long_field);
+  EXPECT_LT(long_field, static_cast<int64_t>(0));
+  EXPECT_TRUE(TestBit(63, long_field));
+  EXPECT_FALSE(TestBit(62, long_field));
+  ClearBit(63, &long_field);
+  EXPECT_EQ(0, long_field);
+}
+
 TEST(Log2Histogram, 2BinEmpty) {
   Log2Histogram log2hist(2);
   log2hist.Add(10);
@@ -1804,6 +2162,9 @@ TEST(Log2Histogram, 2BinEmpty) {
   for (int i = 0; i < 3; i++) {
     EXPECT_EQ(res[i], atomic_read32(&bins[i]));
   }
+
+  unsigned int q = log2hist.GetQuantile(0.5);
+  EXPECT_EQ(q, 4U);
 }
 
 TEST(Log2Histogram, 2Bins) {
